#include <stdio.h>
#include <bppvm.h>
#include <bppvec.h>
#include <dictio.h>
#include <bppstrng.h>
#include <bppufunc.h>
#include <bppprint.h>
#ifdef NOCONIO
 #undef NOCONIO
 #ifndef _CONSOLE
   #define NOCONIO
 #endif
#endif
#ifndef NOCONIO
  #include <conio.h>
#endif

#ifndef OUTNAME
  #define OUTNAME "bp2"
#endif

#ifndef RTNAME
  #define RTNAME "bp2r.exe"
#endif


static const BppChar* BANNER = _T("B++ v1.1 - Copyright (c) 2005-2018, by Ralf-Erik Ebert\n")
_T("usage:\n")
_T(OUTNAME)
_T(" [-i][-d][-l][-v][-t][-c][-e][-E] [sourcefile [..]] [-r objfile [..]] \n")
_T("   [-o outfile] [# userarg [..]]\n")
_T("options:\n")
_T("    -i ... info (shows syntax and names of compiled functions)\n")
_T("    -d ... shows debug information during compilation\n")
_T("    -l ... includes source file and line information into bytecode\n")
_T("    -v ... includes local variables information into bytecode\n")
_T("    -t ... trace bytecodes during execution\n")
_T("    -c ... compile only\n")
_T("    -e ... builds standalone exe file (compiles and links against runtime)\n")
#ifdef WIN32
_T("    -E ... builds standalone exe file without console (compiles and links against runtime)\n")
#endif
_T("    -r ... specifies precompiled files to be loaded\n")
_T("    -o ... name of output file (used only with -c or -e)\n")
_T("     # ... marks begin of user argument list\n"); 

enum ArgType
{
   AT_ANY = 0,
   AT_SOURCE,
   AT_LIB,
   AT_OUT
};

static bool BPPAPI dbgWait()
{
#ifdef NOCONIO
#ifdef WINCE
#define MB_SYSTEMMODAL MB_TOPMOST|MB_SETFOREGROUND
#endif
   return (MessageBox(GetActiveWindow(),_T("Press OK for next step\nor Cancel to leave step mode."),
      _T("Single Step Mode"),MB_OKCANCEL|MB_SYSTEMMODAL) == IDOK);
#else
   return (getch() != 27);
#endif
}


unsigned long cpy_file(const BppChar* src, const BppChar* dest, bool append)
{
   FILE *fps, *fpd;
    int c;
   unsigned long flen = 0;
   fps = _tfopen(src,_T("rb"));
   if (fps == NULL)
   {
       _tprintf(_T("Unable to open file ('%s')\n"),src);
       return 0;
   }
   fpd = _tfopen(dest,append ? _T("ab") : _T("wb"));
   if (fpd == NULL)
   {
       fclose(fps);
       _tprintf(_T("Unable to open file ('%s')\n"),dest);
       return 0;
   }
   while(!feof(fps))
   {
      c = fgetc(fps);
      if (c == EOF) break;
      if (fputc(c,fpd) == EOF)
      {
         _tprintf(_T("Error while writing '%s'.\n"),dest);
         break;
      }
      ++flen;
   }
   fclose(fps);
   fclose(fpd);
   return flen;
}

/* build exe from compiled file */
static bool build_exe(const BppChar* outname, bool noConsole = false)
{
   FILE* fp = 0;
   unsigned long flen;
   const BppChar* rtname = _T(RTNAME);
#ifdef WIN32
    BppString rtnameStr = rtname;
    if (noConsole) {
       const BppChar* pr = _tcsrchr(rtname, _T('r'));
       long pos = rtnameStr.find(_T('r'));
       if (pos > 0) {
          rtnameStr.insert(pos + 1, _T('w'));
          rtname = rtnameStr.c_str();
       }
    }
#endif
    const BppChar* rtpath = searchPath(rtname);
    if (rtpath == NULL)
          _tprintf(_T("Unable to locate runtime file ('%s')\n"),rtname);
    flen = cpy_file(rtpath,outname,false);
    cpy_file(_T("out.tmp"),outname,true);
    fp = _tfopen(outname,_T("r+b"));
    if (fp == 0)
    {
       _tprintf(_T("Open output file ('%s') failed.\n"),outname);
       return false;
    }
    fseek(fp,0,SEEK_END);
       fwrite(&flen,1,sizeof(unsigned long),fp);
       fwrite("BPR",1,4,fp);
       fclose(fp);
    return true;
}

#ifdef STKLEN
extern unsigned _stklen = STKLEN;
#endif

int
#ifdef _WIN32_WCE_EMULATION
__cdecl
#endif
 _tmain(int argc, BppChar* argv[])
{
   BppFileStreamPrinter fprt(stdout);
   BppVirtualMachine vm(&fprt);
   vm.setDbgWaitProc(dbgWait);
   vm.cmdargc() = argc;
   vm.cmdargv() = argv;
   bool showInfo = false;
   bool showDebug = false;
   bool doTrace = false;
   bool compileOnly = false;
   bool linkExe = false;
#ifdef WIN32
   bool noConsole = false;
#endif
   BppString outfile = _T("out");
   BppVector moduleVec;
   BppVector srcVec;
   ArgType argtype = AT_SOURCE;
   if (argc < 2)
   {
      _tprintf(_T("%s"),BANNER);
      return -1;
   }
   for (int i=1; i<argc; i++)
   {
      const BppChar* arg = argv[i];
      BppChar a0 = arg[0];
      if (a0 == _T('#')) break; // rest of cmdline contains user args
      if (a0 == _T('-')) // option
      {
         switch (arg[1])
         {
         case _T('d'):
            showDebug = true;
            break;
         case _T('l'):
            vm.setGenerateLineInfo(true);
            break;
         case _T('v'):
            vm.setGenerateLocalVarInfo(true);
            break;
         case _T('c'):
            compileOnly = true;
            break;
         case _T('E'):
#ifdef WIN32
            noConsole = true;
#endif
         case _T('e'):
            compileOnly = true;
			linkExe = true;
            break;
         case _T('t'):
            doTrace = true;
            break;
         case _T('i'):
            showInfo = true;
            break;
         case _T('r'):
            argtype = AT_LIB;
            break;
         case _T('o'):
            argtype = AT_OUT;
            break;
         default:
            _tprintf(_T("invalid option %s\n%s"),arg,BANNER);
            return -1;
         }
         continue;
      }
      if (argtype == AT_OUT)
      {
          outfile = arg;
      }
      else if (argtype == AT_LIB)
      {
         BppString* s = new BppString(arg);
         if (_tcschr(arg,_T('.')) == 0)
              (*s) += _T(".bpm");
         moduleVec.push_back(s);
      }
      else // assume source file
      {
         BppString* s = new BppString(arg);
         if (_tcschr(arg,_T('.')) == 0)
              (*s) += _T(".bpp");
         srcVec.push_back(s);
      }
   }
   if (outfile.find(_T('.')) < 0)
	  outfile += linkExe ? _T(".exe") : _T(".bpm");
   unsigned cnt = moduleVec.size();
   unsigned j;
   for (j=0; j<cnt; j++)
   {
      const BppChar* modulename = ((BppString*)moduleVec[j])->c_str();
      const BppChar* fname = searchEnv(_T("BPPLIB"),modulename);
      if (fname == 0)
         fname = modulename;

      if (!vm.readModule(fname))
         return -1;
   }
   if (showInfo) _tprintf(_T("%s"),BANNER);
   cnt = srcVec.size();
   for (j=0; j<cnt; j++)
      if (!vm.compileFile(((BppString*)srcVec[j])->c_str(),showInfo,showDebug))
         return -1;
   if (compileOnly)
   {
	   if (!vm.writeModule(linkExe ? _T("out.tmp") : outfile.c_str()))
         return -1;
	   if (linkExe && (!build_exe(outfile.c_str()
#ifdef WIN32
         ,noConsole
#endif
      )))
		   return -1;
      return 0;
   }
   if (!vm.execute(_T("main"),argc,(const BppChar**)argv,doTrace))
   {
      return -1;
   }
   return (int) (BppInt) vm.getResult();
}

#ifdef WINCE
int WINAPI WinMain(HINSTANCE instance, HINSTANCE prev_instance, BppChar* command_line, int show_command)
{    
    int    argc;
    BppChar** argv;    
    BppChar*  arg;
    int    index;
    int    result;    // count the arguments
    BppChar strchar = 0;
    argc = 1;
    arg  = command_line;
    
    while (arg[0] != 0) 
	{        
		while (arg[0] == _T(' ')) 
		{
            arg++;
      }
		if ((arg[0] == _T('\'')) || (arg[0] == _T('"')))
		{
			strchar = arg[0];
			++arg;
		}
        if (arg[0] != 0) 
		{
            argc++;
            while ((arg[0] != 0) && (arg[0] !=  ((strchar == 0) ? _T(' ') : strchar))) 
			{
                arg++;
            }
			if (strchar != 0)
			{
				strchar = 0;
				++arg;
			}
        }
    }    
    
    // tokenize the arguments

    argv = (BppChar**)allocmem(argc * sizeof(BppChar**)); 
	arg = command_line;
    index = 1;    
	strchar = 0;
	while (arg[0] != 0) 
	{        
		while (arg[0] == _T(' ')) 
		{
            arg++;
        }        
		if ((arg[0] == _T('\'')) || (arg[0] == _T('"')))
		{
			strchar = arg[0];
			++arg;
		}
		if (arg[0] != 0) 
		{
            argv[index] = arg;
            index++;
            while ((arg[0] != 0) && (arg[0] !=  ((strchar == 0) ? _T(' ') : strchar))) 
			{
                arg++;
            }
			if ((arg[0] != 0) && (strchar != 0))
			{
				strchar = 0;
				arg[0] = 0;
				++arg;
			}
            if (arg[0] != 0) 
			{
                arg[0] = 0;    
                arg++;
            }
        }
    
    }        // put the program name into argv[0]

    BppChar filename[_MAX_PATH];
    
    GetModuleFileName(NULL, filename, _MAX_PATH);
    argv[0] = filename;    // call the user specified main function    
    
    result = _tmain(argc, argv);
    
    free(argv);
    return result;
}
#endif
