#ifndef BPPDICT_H
#define BPPDICT_H

#include <bppstrng.h>
#include <bppvalue.h>

#ifdef __TURBOC__
  #define __NOSTL__
#endif

#ifdef __WATCOMC__
  #define __NOSTL__
#endif

#ifdef __SC__
  #define __NOSTL__
  #ifdef DEBUG
    #ifndef _DEBUG
      #define _DEBUG
    #endif
  #endif
  #ifdef _MSDOS
     #define __MSDOS__
  #endif
#endif


/** Possible types of symbols .
Dictionaries are used not only by B++ language but also by its implementation.
The B++ compiler and runtime environment use dictionaries for internal symbol
tables, i.e. for defined classes and functions too. Therefore each entry in a 
dictionary may have a symbol type attribute. This enumeration defines
possible symbol types.
*/
enum BppSymbolType
{
   ST_NONE = 0,  ///< no symbol
   ST_CLASS,     ///< class definition
   ST_DATA,      ///< data member
   ST_SDATA,     ///< static data member
   ST_FUNCTION,  ///< function member
   ST_SFUNCTION  ///< static function member
};

class BppDictionary;
class BppDictionaryIO;

/** Class type of a single dictionary entry.
BppDictionaryEntry is mainly for internal use. It represents a single entry
of a dictionary. Furthermore it is used for internal representation of a variable.
@see BppDictionary
*/
class BPPLIB_DLLEXP BppDictionaryEntry
{
   friend class BppDictionary;
   friend class BppDictionaryIO;
public:
   /// initialized constructor (key is pointer)
   BppDictionaryEntry(BppString* key, const BppValue& val, BppSymbolType stype = ST_NONE);
   /// initialized constructor (key is const reference)
   BppDictionaryEntry(const BppString& key, const BppValue& val, BppSymbolType stype = ST_NONE);
   /// copy constructor
   BppDictionaryEntry(const BppDictionaryEntry& cpy);
   // dtor
   virtual ~BppDictionaryEntry();
   /// assignment operator
   BppDictionaryEntry& operator=(const BppDictionaryEntry& assignment);
   /// value assignment operator
   BppDictionaryEntry& operator=(const BppValue& assignment);
   /**  Gets type of symbol.
   @return type of symbol represented by dictionary entry
   */
   inline BppSymbolType getSymbolType() const
   {
      return _stype;
   }

   /** Sets type of symbol.
   @param stype type of symbol represented by dictionary entry
   */
   inline void setSymbolType(BppSymbolType stype)
   {
      _stype = stype;
   }


   /** Get key of dictionary entry.
   @return key of dictionary entry.
   @note The key is reference counted. Client code must not delete returned value.
   */
   inline BppString* getKey() const
   {
      return _key;
   }

   /** Gets value of dictionary entry.
   @return value of dictionary entry
   */
   inline const BppValue& getValue() const
   {
      return _value;
   }

   /// set value of dictionary entry
   bool setValue(const BppValue& val, bool typeSafe = false, bool copyFlags = false);
 
   /** Gets dictionary owning the entry.
   @return pointer to dixtionary owning this value (or NULL if none)
   */
   inline BppDictionary* getDictionary() const
   {
      return _dict;
   }

   /** Gets next entry.
   @return pointer to next entry in dictionary (or NULL if none)
   */
   inline BppDictionaryEntry* getNext() const
   {
      return _next;
   }
   /// equality operator
   bool operator==(const BppDictionaryEntry& cmp) const;
   /// unequality operator
   bool operator!=(const BppDictionaryEntry& cmp) const;
private:
   BppSymbolType _stype; ///< symbol type
   BppString* _key;      ///< key of entry
   BppValue _value;      ///< value of entry
   BppDictionary* _dict; ///< dictionary owning the entry 
   BppDictionaryEntry* _next;  ///< pointer to next dictionary entry
};


#ifndef __NOSTL__
#include <map>

/// internal map type for BppDictionary class
typedef const BppString* const CPBPPSTR;
template <> struct std::less<CPBPPSTR>
{
   bool operator()(const CPBPPSTR& left, const CPBPPSTR& right) const
   {
      return _tcscmp(left->c_str(), right->c_str()) < 0;
   }
};

//bool operator()(const BppDictionaryEntry *& left, const BppDictionaryEntry *& right) const;

typedef std::map<CPBPPSTR,BppDictionaryEntry*> DictMapType;
#endif

/** Dictionary class.
<p>Dictionaries are used in B++ 2 for two purposes. First they represent internal symbol tables, i.e. tables for 
globally defined functions and variables table as soon as table of classes. Furthermore each class 
itselfs has dictionaries for tables of its data and function members.<br>
On the other hand a dictionary is a common container type for use in a B++ program.</p>
<p>A Dictionary in B++ can be used in two modes - a "sorted" and an "unsorted" mode. In sorted
mode dictionary entries are ordered in alphanumeric order of keys where unsorted mode
preserves order of insertion.<br>
Therefore this class implements a dictionary by a single linked list of BppDictionaryEntry 
type elements. In order to allow fast element access for compiler environments supporting
STL an additional index, based on std::map class, is used.</p>
*/
class BPPLIB_DLLEXP BppDictionary : public BppRefObject
{
   friend class BppDictionaryEntry;
   friend class BppDictionaryIO;
   friend class BppClass;
public:
   /// constructor
   BppDictionary(bool sorted = false);
   /// destructor
   virtual ~BppDictionary();
    /// create clone of current object
	BppDictionary* clone() const;

   /// get first entry
   BppDictionaryEntry* getFirst() const;
   /// find entry by key
   BppDictionaryEntry* find(const BppString& key) const;
   /// element access by key
   BppDictionaryEntry& operator[](const BppString& key);
   /// element access by key pointer
   BppDictionaryEntry* operator[](BppString* key);
   /// clear dictionary
   void clear();
   /// erase specified entry
   bool erase(BppDictionaryEntry* entry);
   /// erase entry by key
   bool erase(const BppString& key);
   /// get number of entries in dictionary
   unsigned count() const;
   /// equality operator
   bool operator==(const BppDictionary& cmp) const;
   /// unequality operator
   bool operator!=(const BppDictionary& cmp) const;

private:
   /// pointer to owner class (for tables in class definitions)
   BppClass* _ownerClass;
   /// copy ctor forbidden
   BppDictionary(const BppDictionary&);
   /// assignment forbidden
   BppDictionary& operator=(const BppDictionary&);
   /// pointer to first entry
   BppDictionaryEntry* _firstElement;
   /// flag for sorted mode
   bool _sorted;
#ifndef __NOSTL__
   /// map of dictionary entries
   DictMapType& _entryMap;
#endif
};

#endif // BPPDICT_H
