#ifndef BPPINT_H
#define BPPINT_H

#include <bppvec.h>
#include <bppstrng.h>
#include <bppvalue.h>

class BppVirtualMachine;
class BppDictionary;
class IBppPrinter;
class BppDebugger;


class BppInterpreter;

struct BppErrorHandler;

/** Helper class containing information about a function currently running in VM.
Instances of this class are created by BppInterpreter for support of debugging and exception handling.
*/
class BPPLIB_DLLEXP BppProcInfo
{
public:
	/// Ctor.
   BppProcInfo(BppInterpreter& interpreter, BppVector* bytecode,int nargs,ValPtr bp, ValPtr fp,bool nocaller = false);
   /// DTor.
   ~BppProcInfo();
   /** Gets access to part of Bytecode vector by index.
   @param idx Index of part to get, where IDX_CODE (0) is the code buffer, IDX_CLASS (1) the class,
          IDX_NAME (2) the function name and IDX_DEFAULTARGS (4) vector of default argument values.
   @return Constant reference to part.
   @see BppByteCodeIndex
   */
   inline const BppValue& operator[](unsigned idx) const
   {
      return (*_bytecode)[idx];
   }

   /// Local variable types.
   enum LocalVarType {
      LVT_ARG = 0, ///< Argument
      LVT_VAR, ///< Function level (temporary) variable
      LVT_STATIC ///< Function level static variable
   };

   /// Gets access to local variables info vector (if any).
   const BppVector* getLocalVarInfo() const;

   /// Gets name of a local variable, specified by index and variable type.
   const BppChar* getLocalVarName(unsigned idx, LocalVarType lvt) const;

   /// Gets declared name of <i>idx</i>th argument.
   inline const BppChar* getArgumentName(unsigned idx) const { return getLocalVarName(idx, LVT_ARG); }
   /// Gets declared name of <i>idx</i>th local (temporary) variable.
   inline const BppChar* getLocalVarName(unsigned idx) const { return getLocalVarName(idx, LVT_VAR); }
   /// Gets declared name of <i>idx</i>th local static variable.
   inline const BppChar* getLocalStaticVarName(unsigned idx) const { return getLocalVarName(idx, LVT_STATIC); }

   /// Gets pointer to code buffer contents.
   operator BppByte*() const;
   /// Gets number of arguments.
  inline int getNargs() const
   {
      return _nargs;
   }
   /// Gets function name.
   const BppChar* getName() const;


   /// Gets class defining function.
   BppClass* getClass() const;

   /// bp points before 1st argument (i.e. to object in method calls).
   inline ValPtr& bp()
   {
      return _bp;
   }

   /// fp points before 1st local (temporary) variable).
   inline ValPtr& fp()
   {
      return _fp;
   }

   /** Gets Descriptor of function calling the current one.
   @return Pointer to caller function descriptor or 0, if current function has no caller.
   @note Function behaves similary to getParent function. The only difference is checking a flag signaling
   that there is no caller.
   */
   inline BppProcInfo* getCaller() const
   {
      return _nocaller ? 0 : _parent;
   }

   /** Gets parent item of function in call stack.
   @return Pointer to parent item or 0, if there is no parent.
   */
   inline BppProcInfo* getParent() const
   {
      return _parent;
   }

   /** Explicitely sets parent of current function.
   @note Function is intended for internal use and called by BppInterpreter.
   Client applications never should call it, however, debuggers may use it to implement
   manipulations of call stack.
   */
   inline void setParent(BppProcInfo* parent)
   {
      _parent = parent;
   }

   /** Pushes error handler to error errorhandler stack
   @param pc Program counter (index in bytecode buffer) of start of handler code (catch block).
   @param pStack Stack pointer value at begin of code section protected by handler (start of try block)
   @see BppErrorHandler
   */
   void pushErrorHandler(BppNWord pc, const ValPtr& pStack);
   /// Removes top of errorhandler stack.
   bool popErrorHandler();
   /// Gets top of errorhandler stack.
   inline BppErrorHandler* getEHStack() const
   {
      return _errHandlerStack;
   }

private:
#ifndef DOXY
   BppProcInfo(const BppProcInfo&);
   BppProcInfo& operator=(const BppProcInfo&);
#endif
   BppInterpreter& _interpreter; ///< Reference to interpreter.
   BppErrorHandler* _errHandlerStack; ///< Top of exception handler stack.
   BppVector* _bytecode; ///< ByteCode vector of function
   ValPtr _bp; ///< Base pointer, points before 1st argument (i.e. to object in method calls)
   ValPtr _fp; ///< Frame pointer, points before 1st local (temporary) variable)
   int _nargs; ///< Number of arguments
   BppProcInfo* _parent; ///< Pointer to parent entry on call stack.
   bool _nocaller; ///< Flag for extern called function (e.g. from inside a native function)
};

struct __BppProcHelper;

/** B++ Bytecode Interpreter.
@note Bytecode interpreter is used by VM of B++.
Client code normally should prefer calling methods of VM instead of using interpreter directly if possible.
*/
class BPPLIB_DLLEXP BppInterpreter
{
   friend class BppVirtualMachine;
   friend struct __BppProcHelper;
public:
   BppInterpreter(BppVirtualMachine& vm);
   ~BppInterpreter();

   /// Gets reference to VM owning the interpreter
   inline BppVirtualMachine& getVirtualMachine()
   {
      return _vm;
   }


   /// execute entry point function
   bool execute(const BppChar* name, int argc, const BppChar* argv[]);
   /// call a function
   bool call(int nargs,BppVector* byteCode);
   /// call object method
   bool send(int nargs, const BppChar* sel = 0, const BppValue* funcVal = 0);
   // reset interpreter
   void reset();
   /// access error message buffer
   inline BppString& errors()
   {
      return _errBuffer;
   }   
   /// set debugger for this interpreter
   void setDebugger(BppDebugger* debugger);
   /// Gets information about current procedure
   inline BppProcInfo* getProcInfo() const
   {
      return _procStack;
   }
   /// Helper function for getting string from object value
   BppString* getStringFromValue(BppClass* ownClass, BppValue& val);

   /// Helper function for getting string from object value without class context
   BppString* getStringFromValueExtern(BppValue& val)
   {
      return getStringFromValue(0, val);
   }
   /// Helper function for conversion to numeric value
   bool doNumCast(BppValue& dest, const BppValue& src, BppValueType dstType, const BppChar* fname=_T(""));
   /// Helper function for conversion to numeric value
   bool doStrCast(BppValue& dest, const BppValue& src);
   /// Invokes any function specified by funcval.
   bool invokeFunc(BppValue& result, const BppValue& funcval, BppVector* args = 0);
   /// Invoke a global function.
   bool invoke(BppValue& result, const BppChar* name, BppVector* args = 0);
   /// Invoke member function of an object.
   bool invokeMember(BppValue& result, BppObject* pobj, const BppChar* name, BppVector* args = 0);
   /// Invoke static method or function assigned to static member variable of a class.
   bool invokeStatic(BppValue& result, BppClass* pclass, const BppChar* name, BppVector* args = 0);
   /// Create object instance of class specified by name.
   bool createObject(BppValue& result, const BppString& className, BppVector* args = 0);
   /// Create object instance of class.
   bool createObject(BppValue& result, BppClass* pclass, BppVector* args = 0);
   /// build error value and return from current procedure
   bool returnError(const BppChar* msg, const BppChar* context = 0);
   /// return error value from current procedure
   bool returnError(const BppValue& errValue);
   /// return method not found error
   bool noMethod(const BppChar* name);
   /// check value having specified data type
   bool checkType(const BppValue& val, BppValueType vtype);
   /// return invalid type error
   bool badType(BppValueType type, BppValueType expected);
   /// return invalid type error
   bool badType(const BppValue& val, const BppValue& expected);
   /// check if break requests are allowed
   inline bool getAllowBreakRequest() const
   {
      return _allowBreakRequest;
   }

   /// enable/disable break requests
   inline void setAllowBreakRequest(bool val)
   {
      _allowBreakRequest = val;
   }

   /// set cancel request flag
   inline void cancel()
   {
      _requestCancel = true;
   }
private:
   /// procedure entry helper function
   void enterProc(BppVector* bytecode,int nargs,ValPtr bp, ValPtr fp);
   /// procedure leave helper function
   void leaveProc();
   /// helper for typesafe assignment
   bool safeAssign(BppValue& dest, const BppValue& src);
/** @name opcode handler functions */
   //@[
   /// handle + operator
   bool doOP_ADD();
   /// handle - operator
   bool doOP_SUB();
   /// handle * operator
   bool doOP_MUL();
   /// handle / operator
   bool doOP_DIV();
   /// handle % operator
   bool doOP_REM();
   /// handle & operator
   bool doOP_BAND(); 
   /// handle | operator
   bool doOP_BOR(); 
   /// handle ^ operator
   bool doOP_XOR(); 
   /// handle ~ operator
   bool doOP_BNOT(); 
   /// handle << operator
   bool doOP_SHL();  
   /// handle >> operator
   bool doOP_SHR();
   /// handle => operator
   bool doOP_MKREF();
   /// handle =* operator
   bool doOP_UNREF(BppClass* ownClass);
   /// handle [] operator (get)
   bool doOP_VREF();
   /// handle [] operator (set)
   bool doOP_VSET();
   /// handle . operator (get)
   bool doOP_PREF();
   /// handle . operator (set)
   bool doOP_PSET();
   /// handle compare operators
   bool doCompare(BppClass* ownClass);
   /// handle equality operators
   bool doEquals(BppClass* ownClass);
   //@}
   /// copy ctor (forbidden)
   BppInterpreter(const BppInterpreter&);
   /// assignment operator (forbidden)
   BppInterpreter& operator=(const BppInterpreter&);
   /// interpreter stack
   BppVector _stack;
   /// interpreter stack pointer
   ValPtr _sp;
   /// stack of pending exceptions
   BppVector _errStack;
   /// dictionary of defined classes
   BppDictionary& _classes;
   /// dictionary of globally defined symbols
   BppDictionary& _symbols;
   /// reference to default printer
   IBppPrinter& _printer;
   /// reference to virtual machine owning this interpreter
   BppVirtualMachine& _vm;
   /// error message buffer
   BppString _errBuffer;
   /// procedure call stack
   BppProcInfo* _procStack;
   /// debugger currently used
   BppDebugger* _debugger;
   /// flag for trace execution
   bool _trace;
   /// flag for single step mode
   bool _step;
   /// flag for allowing break requests
   bool _allowBreakRequest;
   /// flag for cancel request
   bool _requestCancel;
};

#endif // BPPINT_H
