#ifndef BPPVALUE_H
#define BPPVALUE_H

#include <stdio.h>
#include "bppdef.h"

/// Type catagory classification.
enum BppVTypeCategory {
   BTC_PSEUDO = 0, ///< Pseuto types (null, void, auto and internal pseudo types)
   BTC_NUMBER = 0x40, ///< Number types 
   BTC_WEAKREFFLAG = 0x40, ///< Marks weak references for reference types.
   BTC_REFTYPE = 0x80, ///< Reference types
   BTC_WEAKREF = BTC_REFTYPE | BTC_WEAKREFFLAG ///< Weak reference types
};


/// Sub-category of number type.
enum BppNumberCategory {
   BNC_INTEGRAL = 0, ///< Integral type
   BNC_REAL = 0x20 ///< Real number type
};

/// Signed property of integral types.
enum BppSignedType {
   BST_SIGNED = 0, ///< Signed number
   BST_UNSIGNED = 1 ///< Unsigned number
};

/// Size type flags of integral numbers.
enum NumSizeShift {
   BNS_BYTE = 0, ///< Single byte
   BNS_WORD = 4, ///< Word (16 bits)
   BNS_DWORD = 8, ///< DWord (32 bits)
   BNS_INTPTR = 0x0A, ///< Address (32 or 64 bits)
   BNS_QWORD = 0x0C ///< QWord (64 bits)
};

/// Subtype of reference types.
enum BppRefTypeCategory {
   BRTC_PTRTYPE = 0, ///< Simple pointer type
   BRTC_REFCOUNTED = 0x20 ///< Reference counted type
};

/// Subtype of reference counted type.
enum RefCountedCategory {
   BRCC_SINGLE = 0, ///< Single element
   BRCC_SEQUENCE = 0x10 ///< Sequence of elements
};

/// Subtype of single reference counted type.
enum BppRefSingleCategory {
   BRSC_OBJECTTYPE = 0, ///< Object
   BRSC_CLASS = 0x08 ///< Class
};

/// Subtype of reference counted sequence type.
enum BppRefSequenceCategory {
   BRSQC_SIMPLE = 0, ///< Simple sequence (strings, buffers)
   BRSQC_VECTOR = 0x04, ///< Vector
   BRSQC_DICT = 0x08 ///< Dictionary
};



/** Enumeration of B++ internal data types.
*/
enum BppValueType
{
 _BVTMIN	= 0,
 BVT_NULL = BTC_PSEUDO, ///< null type (not initialized value)
 BVT_VOID, ///< defines no value (for return type of functions only)
 BVT_AUTO, ///< defines strong typed value, where type is determined on first assignment
/** @name Pseudo types (for internal typechecks only.
@see BppInterpreter::checkType
*/
//@{
BVT_NUMBER,
BVT_INTNUMBER,
BVT_UINTNUMBER,
BVT_INTEGRALTYPE,
BVT_PTRTYPE,
BVT_REALNUMBER,
BVT_REFCNTTYPE,
BVT_REFERENCE,
//@}
   _BVTNUMBERBASE = BTC_NUMBER,
 _BVTINTEGRALBASE = _BVTNUMBERBASE | BNC_INTEGRAL,
 BVT_SBYTE = _BVTINTEGRALBASE | BNS_BYTE | BST_SIGNED, ///< 8 bits integer
 BVT_INTEGRAL_MIN = BVT_SBYTE,
 BVT_BYTE = _BVTINTEGRALBASE | BNS_BYTE | BST_UNSIGNED, ///< 8 bits unsigned integer
 BVT_SHORT = _BVTINTEGRALBASE | BNS_WORD | BST_SIGNED, ///< 16 bits signed integer
 BVT_USHORT = _BVTINTEGRALBASE | BNS_WORD | BST_UNSIGNED, ///<16 bits unsigned integer
 BVT_INT = _BVTINTEGRALBASE | BNS_DWORD | BST_SIGNED, ///< 32 bits signed integer
 BVT_UINT = _BVTINTEGRALBASE | BNS_DWORD | BST_UNSIGNED, ///< 32 bits unsigned integer
 BVT_INTPTR = _BVTINTEGRALBASE | BNS_INTPTR | BST_SIGNED, ///< signed integer of same size as pointer
 BVT_LONG = _BVTINTEGRALBASE | BNS_QWORD | BST_SIGNED, ///< 64 bits signed integer
 BVT_ULONG = _BVTINTEGRALBASE | BNS_QWORD | BST_UNSIGNED, ///< 64 bits unsigned integer
 BVT_INTEGRAL_MAX = BVT_ULONG,
 _BVTREALNUMBASE = _BVTNUMBERBASE | BNC_REAL,
 BVT_FLOAT = _BVTREALNUMBASE | BNS_DWORD, ///< single precision real type (32 bits)
 BVT_REAL_MIN = BVT_FLOAT,
 BVT_DOUBLE = _BVTREALNUMBASE | BNS_QWORD, ///< double precision real type (64 bits)
 BVT_REAL_MAX = BVT_DOUBLE,
 _BVTVALMAX = BVT_DOUBLE,
 _BVTREFTYPEBASE = BTC_REFTYPE,
 _BVTREFPTRBASE = _BVTREFTYPEBASE | BRTC_PTRTYPE,
 BVT_FILE = _BVTREFPTRBASE, ///< file descriptor pointer
 BVT_CODE, ///< native code (function) pointer
 BVT_VAR, ///< variable (contained in dictionaries)
 _BVTREFCNTBASE = _BVTREFTYPEBASE | BRTC_REFCOUNTED,
 _BVTREFCNTSINGLEBASE = _BVTREFCNTBASE | BRCC_SINGLE,
 _BVTOBJECTBASE = _BVTREFCNTSINGLEBASE | BRSC_OBJECTTYPE,
 BVT_OBJECT = _BVTOBJECTBASE, ///< B++ object (instance of class)
 BVT_RCNT_MIN = BVT_OBJECT,
 BVT_STRONGOBJECT, ///< strong typed holder of B++ object (instance of class)
 _BVTCLASSBASE = _BVTREFCNTSINGLEBASE | BRSC_CLASS,
 BVT_CLASS = _BVTCLASSBASE,  ///< B++ class descriptor object
 _BVTSEQUENCEBASE = _BVTREFCNTBASE | BRCC_SEQUENCE,
 _BVTSEQSIMPLEBASE = _BVTSEQUENCEBASE | BRSQC_SIMPLE,
 BVT_STRING = _BVTSEQSIMPLEBASE, ///< string value
 BVT_BUFFER, ///< byte buffer
 BVT_CHARBUFFER, ///< character buffer
 _BVTVECTORBASE = _BVTSEQUENCEBASE | BRSQC_VECTOR,
 BVT_VECTOR = _BVTVECTORBASE, ///< vector (dynamic array) of values
 BVT_BYTECODE, ///< bytecode vector
 BVT_LOCALVARINFO, ///< vector of local variable descriptions
 _BVTDICTBASE = _BVTSEQUENCEBASE | BRSQC_DICT,
 BVT_DICTIONARY = _BVTDICTBASE, ///< dictionary (maps key strings to values)
 BVT_RCNT_MAX = BVT_DICTIONARY,
 _BVTMAX = BVT_RCNT_MAX,
 _BVTWEAKREFBASE = _BVTREFCNTBASE | BTC_WEAKREFFLAG, 
  BVT_OBJECTREF = BVT_OBJECT | BTC_WEAKREFFLAG, ///< reference to an object
 _BVTREFMIN = BVT_OBJECTREF,
 BVT_STRONGOBJECTREF = BVT_STRONGOBJECT | BTC_WEAKREFFLAG,
 BVT_CLASSREF = BVT_CLASS | BTC_WEAKREFFLAG, ///< reference to class descriptor
 BVT_STRINGREF = BVT_STRING | BTC_WEAKREFFLAG, ///< reference to string
 BVT_BUFFERREF = BVT_BUFFER | BTC_WEAKREFFLAG, ///< reference to byte buffer
 BVT_CHARBUFFERREF = BVT_CHARBUFFER | BTC_WEAKREFFLAG, ///< reference to character buffer
 BVT_VECTORREF = BVT_VECTOR | BTC_WEAKREFFLAG, ///< reference to vector
 BVT_BYTECODEREF = BVT_BYTECODE | BTC_WEAKREFFLAG, ///< reference to bytecode vector
 BVT_VARINFOREF = BVT_LOCALVARINFO | BTC_WEAKREFFLAG,
 BVT_DICTIONARYREF = BVT_DICTIONARY | BTC_WEAKREFFLAG, ///< dictionary reference
 _BVTREFMAX = BVT_DICTIONARYREF
};

inline BppValueType maskWeakRef(BppValueType t) { return (t & BTC_REFTYPE) ? BppValueType(t & ~BTC_WEAKREFFLAG) : t; }

/** Additionally attribute flags for B++ values.
@note Not all flags are used in current implementation - some of them are reserved
for further versions.
*/
enum BppValueFlags
{
   BVF_NONE = 0,
   BVF_CONST = 1, ///< constant (reserved)
   BVF_STRONGTYPED = 2,///< strong typed value
   BVF_BYREF = 4, ///< byref argument (reserved)
   BVF_ABSTRACT = 8, ///< abstract class or method (reserved)
   BVF_SEALED = 0x10, ///< sealed class or method (reserved)
   BVF_INVALID = 0x18, ///< marks value invalid
   BVF_PRIVATE = 0x20, ///< private member
   BVF_PROTECTED = 0x40, ///< protected member
   BVF_PUBLIC = BVF_PRIVATE | BVF_PROTECTED, ///< public member
   BVF_STATIC = 0x80 /// < marks value static

};

/// Helper function for getting readable type name from type id.
const BPPLIB_DLLEXP BppChar* getValueTypeName(BppValueType type, bool getPrefix = false);

class BppClass;
class BppObject;
class BppVector;
class BppBytecode;
class BppBuffer;
class BppCharBuffer;
class BppRefObject;
class BppString;
class BppDictionaryEntry;
class BppDictionary;
class BppInterpreter;
class BppDefaultFunctions;
#ifndef DOXY
   #define v_type v_exttype._type
   #define v_flags v_exttype._flags
#endif

/** B++ value object class.
All values in B++ are represented by instances of this class. Furthermore
such value objects are used by compiler and virtual machine itself.<br>
Class provides a set of constructors to create object from different data objects
and a set of conversion operators to get these data objects bach from value object.
*/
class BPPLIB_DLLEXP BppValue
{
public:
  /** @name Constructors. */
  //@{
  /// Default ctor.
  BppValue();
  /// Constructs special typed empty value.
  BppValue(BppValueType vtype, BppValueFlags flags, bool createDefaultObjects);
  /// Creation from any reference counted value by explicit type specification.
  BppValue(BppRefObject* val,  BppValueType vtype);
  /// Creation from class.
  BppValue(BppClass*, bool isRef = false);
  /// Creation from object.
  BppValue(BppObject*, bool isRef = false);
  /// Creation from Vector
  BppValue(BppVector*,bool isBytecode = false, bool isRef = false);
  /// Creation from string pointer.
  BppValue(BppString*, bool isRef = false);
  /// Creation from string.
  BppValue(const BppString&);
  /// Creation from byte buffer.
  BppValue(BppBuffer*, bool isRef = false);
  /// Creation from character buffer.
  BppValue(BppCharBuffer*, bool isRef = false);
  /// Creation from Ditionary.
  BppValue(BppDictionary*, bool isRef = false);
  /// Creation from dictionary entry (variable descriptor).
  BppValue(BppDictionaryEntry*);
  /// Creation from native (user defined) function.
  BppValue(BPPUSERFUNC);
  /// Creation from BppSByte (8 bits integer) value.
  BppValue(BppSByte);
  /// Creation from BppByte (8 bits unsigned integer) value.
  BppValue(BppByte);
  /// Creation from BppShort (16 bits integer) value.
  BppValue(BppShort);
  /// Creation from BppUShort (16 bits unsigned integer) value.
  BppValue(BppUShort);
  /// Creation from BppInt (32 bits integer) value.
  BppValue(BppInt, bool isptr = false);
  /// Creation from BppUInt (32 bits unsigned integer) value.
  BppValue(BppUInt);
  /// Creation from BppLong (64 bits integer) value.
  BppValue(BppLong, bool isptr = false);
  /// Creation from BppULong (64 bits unsigned integer) value.
  BppValue(BppULong);
  /// Creation from float (single precision real) value.
  BppValue(float);
  /// Creation from double (double precision real) value.
  BppValue(double);
  /// Creation from file descriptor.
  BppValue(FILE*);
  /// Copy constructor.
  BppValue(const BppValue& cpy, bool clone=false);
  //@}
  /// Destructor.
  ~BppValue();
  /// Assignment operator.
  BppValue& operator=(const BppValue&);
  /// Assignment operator.
  bool assign(const BppValue&);

  /** Gets value type.
  @return value type attribute
  */
  inline BppValueType getType() const
  {
     return (BppValueType) v_type;
  }

  /** Gets value flags.
  @return value flags attribute.
  @see BppValueFlags
  */
  inline unsigned char getFlags() const
  {
     return v_flags;
  }

  /** Sets value flags.
  @param flags new value flags attribute.
  @see BppValueFlags
  */
  inline void setFlags(unsigned char flags)
  {
     v_flags = flags;
  }
  
  /** @name Type conversion operators. */
  //@{
  /// Conversion to class.
  operator BppClass*() const;
  /// Conversion to object.
  operator BppObject*() const;
  /// Conversion to vector.
  operator BppVector*() const;
  /// Conversion to string.
  operator BppString*() const;
  /// Conversion to byte buffer.
  operator BppBuffer*() const;
  /// Conversion to character buffer.
  operator BppCharBuffer*() const;
  /// Conversion to dictionary.
  operator BppDictionary*() const;
  /// Conversion to variable descriptor.
  operator BppDictionaryEntry*() const;
  /// Conversion to native function pointer.
  operator BPPUSERFUNC() const;
  /// Conversion to file descriptor.
  operator FILE*() const;
  /// Conversion to 8 bits integer.
  operator BppSByte() const;
  /// Conversion to 8 bits unsigned integer.
  operator BppByte() const;
  /// Conversion to 16 bits integer.
  operator BppShort() const;
  /// Conversion to 16 bits unsigned integer.
  operator BppUShort() const;
  /// Conversion to 32 bits integer.
  operator BppInt() const;
  /// Conversion to 32 bits unsigned integer.
  operator BppUInt() const;
  /// Conversion to 64 bits integer.
  operator BppLong() const;
  /// Conversion to 64 bits unsigned integer.
  operator BppULong() const;
  /// Conversion to single precision real.
  operator float() const;
  /// Conversion to double precision real.
  operator double() const;
  //@}
  /// Generic compare function
  int compare(const BppValue& cmp) const;
  /// Equality operator
  bool operator==(const BppValue& cmp) const;
  /// Non-equality operator
  bool operator!=(const BppValue& cmp) const;

  /** Get (re-interpret) value as untyped pointer.
  @return Stored value as raw pointer.
  @note Function simply re-interprets value as a native untyped pointer, i.e. in case
  of any reference type it returns pointer to data object. In case of a value type
  it behaves like a reinterpret-cast in C++.
  @see addr
  */
  void* ptr() const
  {
     return _v.v_addr;
  }

 
  /// Get address of data (depending on value type).
  void* addr();
  /// Dispose value.
  void dispose(); 
  /// Assign reference type data vvalue.
  void assignRef(BppRefObject*);

  /** Gets reference type data value.
  @return Pointer to value interpreted as reference(counted) type.
  @note Function doesn't do any type check, i.e. it behaves like C++ reinterpret cast.
  */
  inline BppRefObject* getRef() const
  {
     return _v.v_refObj;
  }
  /// Get descriptor of strong typed object.
  BppVector* getStrongObjDesc() const;
  /// Set protection attribute.
  void setProtection(short protflags);
  
  /**
  @result Protection flag mask of value, i.e. BVF_PRIVATE or BVF_PROTECTED or BVF_PUBLIC or BVF_NONE.
  @see setProtection
  */
  inline unsigned short getProtection() const
  {
     return v_flags & BVF_PUBLIC;
  }

  /// Checks if value type is in specified range.
  inline bool inTypeRange(BppValueType from, BppValueType to) const
  {
     return v_type >= from && v_type <= to;
  }
  /// Checks if value type is integral number.
  inline bool isIntegralType() const
  {
     return v_type >= BVT_INTEGRAL_MIN && v_type <= BVT_INTEGRAL_MAX;
  }

  /// Checks if value type is real number.
 inline bool isRealType() const
  {
     return v_type >= BVT_REAL_MIN && v_type <= BVT_REAL_MAX;
  }

 /// Checks if value is reference counted.
 inline bool isRefCounted() const
 {
    return v_type >= BVT_RCNT_MIN && v_type <= BVT_RCNT_MAX;
 }

  /** Does raw copy of value members.
  @param dst Destination value
  @param src Source value
  @note Function copies data without any typechecking and without calling dispose()
  on destination before. It is intended for internal use or implementation of native
  extensions. 
  */
  static inline void rawcopy(BppValue& dst, const BppValue& src)
  {
	  dst._v = src._v;
	  dst.v_exttype = src.v_exttype;
  }
private:
   friend class BppInterpreter;
   friend class BppDefaultFunctions;
   /// Type descriptor structure.
   struct BPPLIB_DLLEXP ExtType
   {
      /// Value type.
      unsigned short _type : 8;
      /// Additional attribute flags for value.
      unsigned short _flags : 8;
      /// Default ctor.
      ExtType() : _type(BVT_NULL),_flags(0) {}
      /// Initialized ctor.
      ExtType(BppValueType type,unsigned short flags) : _type(type),_flags(flags) {}
   } v_exttype; ///< Type specification.
  /// Data area.
  union
  {			/* value */
    BppRefObject* v_refObj;
    BppClass* v_class;
    BppObject* v_object;
    BppVector* v_vector;
    BppString* v_string;
    BppBuffer* v_buffer;
    BppCharBuffer* v_charbuffer;
    BppDictionary* v_dict;
    BppDictionaryEntry* v_var;
    BPPUSERFUNC v_code;
	BppSByte v_sbyte;
	BppByte v_byte;
	BppShort v_short;
	BppUShort v_ushort;
    BppInt v_integer;
	BppUInt v_uint;
	_Int64 v_long;
	_UInt64 v_ulong;
	BppIntPtr v_intptr;
    float v_float;
    double v_double;
    void* v_addr;
    FILE* v_fp;
  } _v;
};

#endif // BPPVALUE_H
