#include "bppcomp.h"
#include <limits.h>
#include <stdlib.h>
#include <math.h>
#include "bppvec.h"
#include "bppclass.h"
#include "bppobjct.h"
#include "bppscan.h"
#include "bppvalue.h"
#include "bppbuf.h"
#include "bppufunc.h"
#include "bppprint.h"
#include "dictio.h"
#include "bppvm.h"
#include "bppdbg.h"
#include "bppconst.h"

/// maximum number of errors before abort compiling
#define MAXERRS 10


/** Internal helper type for code generation.
This type is used for temporary store of information about a variable
to be coded.
*/
struct BppVarInfo
{
   /// variable type
   BppVarType type;
   /// index inside a container
   int index;
   /// value type
   BppValueType varvtype;
   /// requested class (for strong typed objects)
   BppClass* reqClass;
   /// ctor
   BppVarInfo() :type(VAR_NONE),index(-1),varvtype(BVT_NULL),reqClass(0) {}
};




/** Helper for compiling a function.
This helper struct contains data needed during compile a function.
It's used for argument type by some of the production functions.
*/
struct BppFuncContext
{
   /// ctor - does initialization
   BppFuncContext() : codeVec(0), codeBuffer(0), temporaries(1,true), catchLevel(0),
	   returnType(BVT_NULL),prClass(0),contextClass(0) {}
   /// vector of function arguments
   BppVector arguments;
   /// vector of functions temporary (i.e. local) variables
   BppVector temporaries;
   /// break stack
   BppBuffer breakStack; 
   /// continue stack
   BppBuffer contStack;
   /// destination vector for result of compilation (contains codeBuffer and variables)
   BppVector* codeVec;
   /// buffer for bytecode produced on compiling (part of codeVec)
   BppBuffer* codeBuffer;
   /// Dictionary of static variables - value is index of variable in codeVec
   BppDictionary statics;
   /// level of catch blocks - must be > 0 for throw statement without following expression
   unsigned catchLevel;

   /// return type of function (BVT_NULL if not defined) 
   BppValueType returnType;
   /// optionally class if return type is of object type
   BppClass* prClass;
   /// optionally context class for check member protection
   BppClass* contextClass;


   /** Push entry to break stack.
   @param cbIdx index in codeBuffer where to store jump destination index
   @return old size of stack (i.e. index of new entry)
   */
   BppNWord pushBreak(BppNWord cbIdx)
   {
      return breakStack.push_backNW(cbIdx);
   }
   /** Push entry to continue stack.
   @param cbIdx index in codeBuffer where to store jump destination index
   @return old size of stack (i.e. index of new entry)
   */
   BppNWord pushContinue(BppNWord cbIdx)
   {
      return contStack.push_backNW(cbIdx);
   }

   /** Removes break from stack.
   @param oldIdx old index from stack (result of previous pop-call)
   @param savedVal value to be returned if no entry is removed (index in codeBuffer)
   @return If <i>oldIdx</i> is lower than index of TOS, value on TOS is
   returned (and removed), otherwise <i>savedVal</i>
   */
   BppNWord popBreak(BppNWord oldIdx,BppNWord savedVal)
   {
      return (oldIdx < breakStack.size()) ? breakStack.pop_backNW() : savedVal;
   }

   /** Removes continue from stack until <i>oldIdx</i> reached.
   @param oldIdx old index from stack (result of previous pop-call)
   @return value of last removed entry (index in codeBuffer) or 0
   */
   BppNWord popContinue(BppNWord oldIdx)
   {
      BppNWord result = 0;
      BppNWord curr = contStack.size();
      while (curr > oldIdx)
      {
         result = contStack.pop_backNW();
         curr = contStack.size();
      }
      return result;
   }

   /** Peek top of break stack.
   @return value of item on TOS if any, otherwise 0
   */
   BppNWord peekBreak() 
   {
      BppNWord s = breakStack.size();
      if (s > 0) return breakStack.getNW(s-sizeof(BppNWord));
      return 0;
   }

   /** Peek top of continue stack.
   @return value of item on TOS if any, otherwise 0
   */
   BppNWord peekContinue() 
   {
      BppNWord s = contStack.size();
      if (s > 0) return contStack.getNW(s-sizeof(BppNWord));
      return 0;
   }
};



/**
Constructor assigns reference to virtual machine, dictionaies of classes
and globally symbols as soon as printer interface provided by virtual machine.
@param vm Virtual machine owning the compiler.
@param errPrinter Pointer to printer interface for error message reporting.
@note If <i>errPrinter</i> not specified (or NULL), default output of virtual 
machine is used for reporting errors.
@see BppVirtualMachine::printer
*/
BppCompiler::BppCompiler(BppVirtualMachine& vm, IBppPrinter* errPrinter)
:_scanner(0),_vm(vm),_symbols(vm.symbols()),_classes(vm.classes()),_namedLiterals(true),
_print((errPrinter) ? *errPrinter : vm.printer()),
_showInfo(false),_debugger(0),_genLineInfo(false),_genLocalVarInfo(false),_currentLine(-1)
{
   _currentSource.addRef();
#ifndef WINCE
   errno = 0;
#endif
}

/**
does nothing
*/
BppCompiler::~BppCompiler()
{
}

/**
@return value of _showInfo flag
@see setShowInfo
*/
bool BppCompiler::getShowInfo() const
{
   return _showInfo;
}

/**
If flag set, names of classes and functions are written to output during compilation.
@param showInfo new value of flag
@see _doClass,_doRegularFunction,_doMemberFunction, hetShowInfo
*/
void BppCompiler::setShowInfo(bool showInfo)
{
   _showInfo = showInfo;
}

/**
@return current value of flag
@see setGenerateLineInfo
*/
bool BppCompiler::getGenerateLineInfo() const
{
   return _genLineInfo;
}

/**
If flag is set to <i>true</i> (should be done before call to compile-method), additional information
about source files and line numers are inserted into bytecode for debugging purposes.
@param genInfo new value of flag
@note Insertion of line information means additional opcodes and data entries in compile result,
i.e. it makes code bigger and slower. Therefore line imformation should be generated
for debug builds only.
@see getGenerateLineInfo
*/
void BppCompiler::setGenerateLineInfo(bool genInfo)
{
   _genLineInfo = genInfo;
}

/**
@return current value of flag
@see setGenerateLocalVarInfo
*/
bool BppCompiler::getGenerateLocalVarInfo() const
{
   return _genLocalVarInfo;
}

/**
If <i>genInfo</i> is <tt>true</tt> name information is appended to end of
bytecode vector of a compiled function.
Name information consists of a two-elements vector, where both elements are vectors too.
First element contains names of declared arguments, second names of local (temporary)
variables. Indexes of names in element vectors correspond to relative position of
variables inside stack frame of function while execution.<br>
If Function contains static variables, an additional dictionary is appended to second
element. Keys in that dictionary are names of local static variables, values are their 
indexes in bytecode vector of compiled function.
@param genInfo Value of flag for include information about local variable names in generated code.
@note Creation of information about local variables is useful for debugging purposes.
It increases code size but doesn't affect execution speed.
@see getGenerateLocalVarInfo
*/
void BppCompiler::setGenerateLocalVarInfo(bool genInfo)
{
   _genLocalVarInfo = genInfo;
}

void BppCompiler::setDebugger(BppDebugger* debugger)
{
   _debugger = debugger;
}

void BppCompiler::onScannerDelete(BppScanner * scanner)
{
   if (!scanner) return;
    int i = _srcDirStack.getStringIndex(getFileDir(scanner->_srcName.c_str(), true));
   if (i >= 0) _srcDirStack.erase(i, 1);
}

bool BppCompiler::compile(FILE* fp, const BppChar* srcName)
{
   if (fp == 0)
   {
     _print(_T("BppCompiler::compile(FILE* fp) invalid file handle\n"));
      return false;
   }
#ifndef WINCE
   errno = 0;
#endif
   BppScanner scanner(fp,srcName);
   _srcDirStack.clear();
   _srcDirStack.push_back(new BppString(getFileDir(srcName, true)));
   _scanner = &scanner;
   BppScanner::_notifyCompiler = this;
   bool ok = compile();
   BppScanner::_notifyCompiler = 0;
   return ok;
}

bool BppCompiler::compile(const BppChar* src, const BppChar* srcName)
{
   if (src == 0)
   {
      _print(_T("BppCompiler::compile(const BppChar* src) invalid source\n"));
      return false;
   }
#ifndef WINCE
   errno = 0;
#endif
   BppScanner scanner(src,srcName);
   _scanner = &scanner;
   BppScanner::_notifyCompiler = this;
   bool ok = compile();
   BppScanner::_notifyCompiler = 0;
   return ok;
}

BppVector * BppCompiler::compileFunction(const BppChar * src, const BppChar * args, const BppChar* rtype, const BppChar * srcName, BppClass * pclass)
{
	if (!src)
	{
		compile(src, srcName);
		return 0;
	}
   C_LOCALE;
#ifndef WINCE
   errno = 0;
#endif
	BppString strSrc = BppString::format(_T("%s){\n%s\n}"), args, src);
	BppValueType rvaltype = BVT_NULL;
	BppClass* resultClass = 0;
	int errcnt = 0;
	BppScanner* psc = _scanner;
   BppCompiler* pscomp = BppScanner::_notifyCompiler;
   BppScanner scanner(strSrc.c_str(), srcName);
	_scanner = &scanner;
   BppScanner::_notifyCompiler = this;
	if (rtype && *rtype)
	{
		rvaltype = _vm.findVTypeFromName(rtype, true, &resultClass);
		if (rvaltype == BVT_NULL)
		{
			if (_tcscmp(rtype, _T("var")) && _tcscmp(rtype, _T("null")))
			{
				_error(BppString::format(_T("invalid typename %s"), rtype).c_str(),&errcnt);
			}
		}
	}
	BppVector* result = _codeFunctionBody(_T("<anonymous>"), errcnt, pclass, rvaltype, resultClass);
	if (result && (errcnt > 0))
	{
		delete result;
		result = 0;
	}
   BppScanner::_notifyCompiler = pscomp;
	_scanner = psc;
	return result;
}


void BppCompiler::initCompiler()
{
	_namedLiterals.clear();
	for (unsigned i = _BVTMIN; i <= _BVTREFMAX; ++i)
	{
		_namedLiterals[getValueTypeName((BppValueType)i, true)] = (BppInt)i;
	}
   _namedLiterals[_T("EDOM")] = (BppInt)EDOM;
   _namedLiterals[_T("ERANGE")] = (BppInt)ERANGE;
   _namedLiterals[_T("LC_ALL")] = (BppInt)LC_ALL;
   _namedLiterals[_T("LC_COLLATE")] = (BppInt)LC_COLLATE;
   _namedLiterals[_T("LC_CTYPE")] = (BppInt)LC_CTYPE;
   _namedLiterals[_T("LC_MONETARY")] = (BppInt)LC_MONETARY;
   _namedLiterals[_T("LC_NUMERIC")] = (BppInt)LC_NUMERIC;
   _namedLiterals[_T("LC_TIME")] = (BppInt)LC_TIME;
   _namedLiterals[_T("true")] = (BppInt)1;
	_namedLiterals[_T("false")] = (BppInt)0;
	_namedLiterals[_T("UNICODE")] =
#ifdef _UNICODE
	(BppInt) 1;
#else
		(BppInt)0;
#endif
	_namedLiterals[_T("WIN32")] =
#ifdef WIN32
	(BppInt) 1;
#else
		(BppInt)0;
#endif
	_namedLiterals[_T("WINCE")] =
#ifdef WINCE
	(BppInt) 1;
#else
		(BppInt)0;
#endif
	_namedLiterals[_T("MSDOS")] =
#ifdef __MSDOS__
	(BppInt) 1;
#else
		(BppInt)0;
#endif
	_namedLiterals[_T("BPPI64")] =
#ifdef BPP_I64
	(BppInt) 1;
#else
		(BppInt)0;
#endif
	_currentLine = -1;

}

// start symbol
// bppprog ::= { <function> | <classdeclaration> | <preprocinstruction> } .
bool BppCompiler::compile()
{
   C_LOCALE;
	initCompiler();   /* process statements until end of file */
   int errCnt = 0;
   BppToken tkn;
   while ((errCnt < MAXERRS) && ((tkn = _scanner->getToken()) != T_EOF)) 
   {
	  BppValueType rtype = BVT_NULL;
      switch (tkn) 
      {
      case T_ERROR:
         errCnt++;
         _print(_scanner->getTokenString());
         break;
	  case T_VAR:
      case T_NIL:
         if (!_fetchRequire(T_IDENTIFIER))
         {
            errCnt++;
            _scanner->back();
            break;
         }
	  case T_VOID:
	  case T_AUTO:
	  case T_FUNCTION:
      case T_IDENTIFIER:
		  {
			  BppClass* prClass = 0;
			  rtype = _vm.findVTypeFromName(_scanner->getTokenString(),true,&prClass);
			  if (rtype != BVT_NULL)
			  {
				 BppString typeName = _scanner->getTokenString();
				 BppToken t = tkn;
				 tkn = _scanner->getToken();
				 if (tkn != T_IDENTIFIER) 
				 {
					_scanner->back();
					_scanner->_token = t;
					_scanner->_stringBuffer = typeName.c_str();
					rtype = BVT_NULL;
				 }
			  }
              if (!_doFunctionOrVariable(errCnt,rtype,prClass)) return false;
		  }
         break;
      case T_CLASS:
         if (!_doClass(errCnt)) return false;
         break;
      case T_ANY:
         if (_scanner->getChar() == _T('#'))
         {
            if (!_doPreproc(errCnt)) return (errCnt == 0);
            break;
         }
      default:
         _error(_T("declaration expected"),&errCnt);
      }
   }
   return (errCnt == 0);
}


// func ::= <regularfunc> | <globalvardef> | <memberfunction>
bool BppCompiler::_doFunctionOrVariable(int& errCnt, BppValueType rtype, BppClass* prClass)
{
   BppString name(_scanner->getTokenString());
   BppToken tkn = _scanner->getToken();
   switch (tkn) 
   {
   case T_ANY:
	   {
		   BppChar c = _scanner->getChar();
		   switch (c)
		   {
		   case _T('('):
			   if (rtype == BVT_AUTO)
			   {
               _error(_T("return type 'auto' not allowed"), &errCnt);
			   }
				return _doRegularFunction(name,errCnt,rtype,prClass);
		   case _T(','):
		   case _T(';'):
			   if (rtype == BVT_AUTO)
			   {
				   _error(_T("'auto' variables must be initialized"),&errCnt);
			   }
		   case _T('='):
			   _scanner->back();
			   return _doGlobalVariable(name,errCnt,rtype,prClass);
		   }
	   }
      break;
   case T_CC:
	   if (rtype == BVT_AUTO)
	   {
		   _error(_T("return type 'auto' not allowed"),&errCnt);
	   }
      return _doMemberFunction(name,errCnt,rtype,prClass);
   case T_ERROR:
      _print(_scanner->getTokenString());
	  break;
   default:
      
	   

      break;
    }
   _error(_T("declaration of function or variable expected"),&errCnt);
   return true;
}

const BppChar* BppCompiler::_getOperatorFunc(int& errCnt)
{
   BppToken tkn = _scanner->getToken(); // 'L', 'R' or operator symbol
   bool isRop = false;
   bool hasPrefix = false;
   bool incdec = false;
   bool subop = false;
   bool addop = false;
   bool mulop = false;
   const BppChar* result = 0;
   if (tkn == T_IDENTIFIER)
   {
      const BppChar* ps = _scanner->getTokenString();
	  BppClass* pclass = 0;
	  BppValueType casttype = _vm.findVTypeFromName(ps,true,&pclass);
      if (casttype == BVT_STRONGOBJECT)
		  result = pclass->getClassName()->c_str();
	  else if ((casttype != BVT_NULL) && (casttype != BVT_VOID))
		  result = _vm.getTypeNameFromVType(casttype);
	  switch (casttype)
	  {
	  case BVT_INT:
	  case BVT_UINT:
	  case BVT_SHORT:
	  case BVT_USHORT:
	  case BVT_LONG:
	  case BVT_ULONG:
	  case BVT_SBYTE:
	  case BVT_BYTE:
	  case BVT_INTPTR:
	  case BVT_FLOAT:
	  case BVT_DOUBLE:
	  case BVT_BUFFER:
	  case BVT_CHARBUFFER:
	  case BVT_VECTOR:
	  case BVT_STRING:
	  case BVT_DICTIONARY:
	  case BVT_FILE:
	  case BVT_STRONGOBJECT:
//    case BVT_OBJECT: 
		  break;
	  default:
		  result = 0;
		  {
			  hasPrefix = true;
			  if (_tcscmp(ps,_T("R")) == 0)
				  isRop = true;
			  else if (_tcscmp(ps,_T("L")) != 0)
			  {
				  _error(_T("invalid operator prefix"),&errCnt);
			  }
			  tkn = _scanner->getToken();
		  }
	  }
   }
   if (hasPrefix)
   {
      switch (tkn)
      {
      case T_SHL:
      case T_SHR:
         break;
      case T_ANY:
         if (_tcschr(_T("+-*/%&|^"),_scanner->getChar()) != 0)
            break;
      default:
         _error(_T("invalid operator token behind prefix"),&errCnt);
      }
   }
   switch (tkn)
   {
   case T_IDENTIFIER:
	   if (result != 0)
	   break;
   case T_SHL:
      result = isRop ? _T("OP_SHL_R") : _T("OP_SHL");
      break;
   case T_SHR:
      result = isRop ? _T("OP_SHR_R") : _T("OP_SHR");
      break;
   case T_INC:
      result = _T("OP_INC");
      incdec = true;
      break;
   case T_DEC:
      result = _T("OP_DEC");
      incdec = true;
      break;
   case T_ANY:
      {
         BppChar c = _scanner->getChar();
         switch (c)
         {
         case _T('+'):
            result = isRop ? _T("OP_ADD_R") : _T("OP_ADD");
            if (!hasPrefix) addop = true;
            break;
         case _T('-'):
            result = isRop ? _T("OP_SUB_R") : _T("OP_SUB");
            if (!hasPrefix) subop = true;
            break;
         case _T('*'):
            result = isRop ? _T("OP_MUL_R") : _T("OP_MUL");
            if (!hasPrefix) mulop = true;
            break;
         case _T('/'):
            result = isRop ? _T("OP_DIV_R") : _T("OP_DIV");
            break;
         case _T('%'):
            result = isRop ? _T("OP_REM_R") : _T("OP_REM");
            break;
         case _T('&'):
            result = isRop ? _T("OP_BAND_R") : _T("OP_BAND");
            break;
         case _T('|'):
            result = isRop ? _T("OP_BOR_R") : _T("OP_BOR");
            break;
         case _T('^'):
            result = isRop ? _T("OP_XOR_R") : _T("OP_XOR");
            break;
         case _T('~'):
            result = _T("OP_BNOT");
            break;
         case _T('!'):
            result = _T("OP_NOT");
            break;
         case _T('('):
            result = _T("OP_CALL");
            if (!_fetchRequire(T_ANY,_T(')')))
               ++errCnt;
            break;
         case _T('['):
            {
               BppToken t = _scanner->getToken();
			   if ((t == T_VAR) || (t == T_IDENTIFIER && _tcscmp(_scanner->getTokenString(),_T("int"))==0))
               {
                  result = _T("OP_VSET");
                  t = _scanner->getToken();
               }
               else
                  result = _T("OP_VREF");
               if (!_require(T_ANY,t,_T(']')))
                  ++errCnt;
               t = _scanner->getToken();
               if ((t == T_ANY) && (_scanner->getChar() == _T('=')))
               {
                  result = _T("OP_VSET");
               }
               else
			   {
				  _scanner->back();
			   }
           }
			break;
		 case _T('.'):
			 {
               BppToken t = _scanner->getToken();
               if ((t == T_ANY) && (_scanner->getChar() == _T('=')))
               {
                  result = _T("OP_PSET");
               }
               else
			   {
                  result = _T("OP_PREF");
				  _scanner->back();
			   }
			 }
            break;
         default:
            _error(_T("invalid operator definition"),&errCnt);
         }
      }
      break;
   default:
      _error(_T("invalid operator definition"),&errCnt);
   }
   if (!_fetchRequire(T_ANY,_T('(')))
      ++errCnt;
   if (addop || subop || mulop)
   {
      tkn = _scanner->getToken();
      if ((tkn == T_ANY) && (_scanner->getChar() == _T(')')))
		  result = subop ? _T("OP_NEG") : addop ?  _T("OP_PLUS") : _T("OP_UNREF");
      _scanner->back();
   } 
   else if (incdec)
   {
      BppToken t = _scanner->getToken();
      if ((t == T_VAR) || (t == T_IDENTIFIER && _tcscmp(_scanner->getTokenString(),_T("int"))==0))
      {
         if (tkn == T_INC)
            result = _T("OP_PINC");
         else
            result = _T("OP_PDEC");
      }
      else
         _scanner->back();
   }
   return result;
};

// classdeclaration ::= "class" <identifier> (";" | ( [ <baseclassdecl> ] "{" { <preprocinstruction> | <memberdeclaration> } "}" [";"]) )
// baseclassdecl ::= ":" <identifier>
// memberdeclaration ::= <staticvarmemberdecl> | <varmemberdecl> | <memberfuncdecl>
// staticvarmemberdecl ::= "static" ["var"] <identifier> ["=" <literal>] { ","<identifier> ["=" <literal>] } ";"
// varmemberdecl ::= ["var"] <identifier> { "," <identifier> } ";"
// memberfuncdecl ::= [ "static" ] ["var" | "null"] <identifier> "(" [ <argumentlist ] [ ";" <localvarlist> ] ")" ";"
// Rules:
//   - Constructor is a member function of same name as class.
//   - Destructor is a member function without any parameters and defined as ~<classname>() or dtor().
//   - Constructor and destructor can't be static.
//   - Redefinition of base class is forbidden.
//   - In redeclaration of an already existing class adding new non-static data members is forbidden.
//   - Member variables already defined as non-static can't be redefined to be static.
//   - Declaration of non-static member functions is optional and for better readability only. Any arguments
//     specified in their declaration are ignored.
//   - Static data members may be initialized in declaration by any literal value,
//   - If static modifier is specified on start of a variable list, all elements of list are static
bool BppCompiler::_doClass(int& errCnt)
{
   /* get the class name */
   if (!_fetchRequire(T_IDENTIFIER))
   {
      errCnt++;
      return false;
   }
   BppString className = _scanner->getTokenString();
   BppDictionaryEntry* clsEntry = _classes.find(className);
   BppClass* pclass = (clsEntry != 0) ? (BppClass*) clsEntry->getValue() : 0;


   BppToken tkn = _scanner->getToken();
   if (tkn == T_ERROR)
      _print(_scanner->getTokenString());
   // check for forward declaration
   bool isForward = (tkn == T_ANY) && (_scanner->getChar() == _T(';'));

   BppClass* pBaseClass = 0;
   if (!isForward)
   {
      if ((tkn == T_ANY) && (_scanner->getChar() == _T(':'))) // base class
      {
         if (!_fetchRequire(T_IDENTIFIER)) errCnt++;
         else
         {
            pBaseClass = _findClass(_scanner->getTokenString(),errCnt);
            if (pBaseClass != 0)
            {
               if (pBaseClass == pclass)
               {
                  _error(_T("class can't be its own base"),&errCnt);
                  return false;
               }
               if (_showInfo)
                  _print(_T("class %s : %s\n"),className.c_str(),pBaseClass->getClassName()->c_str());
            }
         }
      } // end if baseclass
      else
      {
         _scanner->back();
         if (_showInfo)
            _print(_T("class %s\n"),className.c_str());
      }
   } // end !isForward
   else
   {
      if (_showInfo)
         _print(_T("class %s forward declaration\n"),className.c_str());
   }

   // Flag if class is already used as base of another one. If so, additional functions but not
   // any new member variables may be defined.
   bool isBaseClass = false;
   // If class is already defined we skip its registration in _classes dictionary.
   if (clsEntry == 0) // class was not already defined
   {
      BppString* pname = new BppString(className);
      clsEntry = _classes[pname];
      clsEntry->setSymbolType(ST_CLASS);
      pclass = new BppClass(_vm,pname);
      clsEntry->setValue(pclass);
   }
   else // check if class is already base of another one
   {
      for (BppDictionaryEntry* pe = _classes.getFirst(); pe != 0; pe = pe->getNext())
      {
         BppClass* pc = pe->getValue();
         if ((pc != 0) && (pc->getBaseClass() == pclass))
         {
            isBaseClass = true;
            break;
         }
      }
   }

   if (pBaseClass != 0) // set base class if specified
   {
      if (pclass->getBaseClass() == 0)
         pclass->setBaseClass(pBaseClass);
      else if (pclass->getBaseClass() != pBaseClass)
      {
         // Note: It's not an error to define same (but only same) base class once more!
          _error(_T("illegal redefinition of base class"),&errCnt);
       }
   }
   if (isForward)
   {
      pclass->computeDataSize();
      return true;
   }

   if (!_fetchRequire(T_ANY,_T('{')))
   {
      errCnt++;
      return false;
   }
   unsigned short protlevel = 0;
   /* handle each variable declaration */
   BppChar tknChar = 0;
   while (!(((tkn = _scanner->getToken()) == T_ANY) && ((tknChar = _scanner->getChar()) == _T('}')))) 
   {
      if (tkn == T_EOF)
      {
         _error(_T("unexpected EOF"),&errCnt);
         return false;
      }
      if ((tkn == T_ANY) && (_scanner->getChar() == _T('#')))
      {
         if (!_doPreproc(errCnt))
            return false;
         if (((tkn = _scanner->getToken()) == T_ANY) && ((tknChar = _scanner->getChar()) == _T('}')))
            break;
      }
      if (tkn == T_PUBLIC)
      {
         protlevel = BVF_PUBLIC;
         if (!_fetchRequire(T_ANY,_T(':')))
         {
            ++errCnt;
         }
         continue;
      }
      if (tkn == T_PROTECTED)
      {
         protlevel = BVF_PROTECTED;
         if (!_fetchRequire(T_ANY,_T(':')))
         {
            ++errCnt;
         }
         continue;
      }
      if (tkn == T_PRIVATE)
      {
         protlevel = BVF_PRIVATE;
         if (!_fetchRequire(T_ANY,_T(':')))
         {
            ++errCnt;
         }
         continue;
      }
      if (tkn == T_ERROR)
         _print(_scanner->getTokenString());
      BppToken type = tkn;
      bool isDtor = (type == T_ANY) && (tknChar == _T('~'));
      /* check for static members */
      if (type == T_STATIC)
         tkn = _scanner->getToken();

      bool varPrefix = (tkn == T_VAR);
      bool nullPrefix = (tkn == T_NIL);
	  BppValueType vtype = BVT_NULL;
	  BppClass* pvClass = 0;
	  // skip next token for var, null and dtor, otherwise check for type modifier
      if (isDtor || varPrefix || nullPrefix)
	     tkn = _scanner->getToken();
	  else
	  {
		  if (tkn == T_IDENTIFIER || tkn == T_VOID || tkn == T_FUNCTION || tkn == T_AUTO)
		  {
			  vtype = _vm.findVTypeFromName(_scanner->getTokenString(),true,&pvClass);
			  if (vtype != BVT_NULL)
			  {
				  BppString typeName = _scanner->getTokenString();
				  BppToken t = tkn;
				  tkn = _scanner->getToken();
				  if (tkn == T_IDENTIFIER || tkn == T_OPERATOR) 
				      varPrefix = true;
				  else
				  {
					  _scanner->back();
					  _scanner->_token = t;
					  tkn = t;
					  _scanner->_stringBuffer = typeName.c_str();
					  vtype = BVT_NULL;
					  //typeName.erase(0);
					  varPrefix = false;
				  }
			  }
		  }
	  }

      /* check for destructor or optionally var or null keyword */
      /* get the first identifier */
      bool opdef = false;
	  bool funcdef = false;
      if (tkn == T_OPERATOR)
      {
         if (type == T_STATIC)
         {
            _error(_T("static operators not allowed"),&errCnt);
         }
         opdef = true;
      }
      else if (tkn != T_IDENTIFIER)
      {

         if (tkn == T_ERROR)
            _print(_scanner->getTokenString());
         _error(_T("member declaration expected"),&errCnt);
         continue;
      }
      BppString id = opdef ? _getOperatorFunc(errCnt) : _scanner->getTokenString();
	  if (opdef && errCnt == 0)
	  {
		  BppClass* prClass = 0;
		  BppValueType rtype = _vm.findVTypeFromName(id,true,&prClass);
		  if (vtype != BVT_NULL && rtype != BVT_NULL)
		  {
			  _error(_T("illegal type before conversion operator"),&errCnt);
		  }
		  else if (rtype != BVT_NULL){ vtype = rtype; pvClass = prClass; }
	  }
      bool isCtor = false;
      if (isDtor)
      {
         if (id.compare(className) != 0)
         {
            _error(_T("destructor declaration expected"),&errCnt);
         }
         else
            id = _T("dtor");
      }
      else
      {
         if (id.compare(className) == 0)
            isCtor = true;
         else if (id.compare(_T("dtor")) == 0)
            isDtor = true;
      }
      if (type == T_STATIC)
      {
         if (isDtor)
         {
            _error(_T("destructor can't be static"),&errCnt);
         }
         else if (isCtor)
         {
            _error(_T("constructor can't be static"), &errCnt);
         }
      }
      else if (protlevel && isDtor && protlevel != BVF_PUBLIC)
      {
         _error(_T("destructor must be public"), &errCnt);
      }
      if (varPrefix || nullPrefix)
      {
         if (isDtor)
         {
            _error(_T("destructor can't return a value"), &errCnt);
         }
         else if (isCtor)
         {
            _error(_T("constructor can't return a value"), &errCnt);
         }
      }
      /* check for a member function declaration */
      if (!opdef) // open bracket already read by _getOperatorFunc
      {
         tkn = _scanner->getToken();
         if (tkn == T_ERROR)
            _print(_scanner->getTokenString());
      }
      if (opdef || ((tkn == T_ANY) && (_scanner->getChar() == _T('('))))
      {
		  BppDictionaryEntry& entry = pclass->funcMembers()[id];
		  entry.setSymbolType((type == T_STATIC) ? ST_SFUNCTION : ST_FUNCTION);
		  ((BppValue&) entry.getValue()).setProtection(protlevel ? protlevel : BVF_PUBLIC);
		  if (_showInfo)
			  _print(_T("member function %s::%s\n"),pclass->getClassName()->c_str(),id.c_str());
		  _scanner->_funcName = BppString::format(_T("%s::%s"),pclass->getClassName()->c_str(),id.c_str());
		  /* compile the body of the function */
		  funcdef = true;
		  BppVector* code = _codeFunctionBody(id,errCnt,pclass,vtype,pvClass);
		  _scanner->_funcName.clear();
		  if (code)
		  {
			  /* store the new method */
			  entry.setValue(BppValue(code,true));
			  ((BppValue&) entry.getValue()).setProtection(protlevel ? protlevel : BVF_PUBLIC);
			  /* build OP_PINC / OP_PDEC methods if needed */
			  const BppString& key = *(entry.getKey());
			  BppString* selector = 0;
			  if (key.compare(_T("OP_INC")) == 0)
				  selector = new BppString(_T("OP_PINC"));
			  else if (key.compare(_T("OP_DEC")) == 0)
				  selector = new BppString(_T("OP_PDEC"));
			  if (selector != 0)
			  {
				  selector->addRef();
				  BppDictionaryEntry* pentry = pclass->findMemberPL(BVF_PROTECTED,*selector,true);
				  if (pentry == 0) // add post increment/decrement method if it not exists
				  {
					  pentry = pclass->funcMembers()[selector];
					  pentry->setSymbolType(ST_FUNCTION);
					  pentry->setValue(BppValue(code,true));
					  ((BppValue&) pentry->getValue()).setProtection(protlevel ? protlevel :BVF_PUBLIC);
				  }
				  selector->release();
			  }
		  }
      }
      /* handle data members */
      else if (type != T_STATIC)
      {
         if (nullPrefix)
         {
            _error(_T("unexpected null-keyword before member variable declaration"), &errCnt);
         }
         if (isBaseClass)
         {
            _error(_T("classes already used as base must not define new member variables"), &errCnt);
         }
         while(1)
         {
			 BppDictionaryEntry tmpEntry(id,BppValue());
             bool initialized = false;
            if ((tkn == T_ANY) && (_scanner->getChar() == _T('=')))
            {
               tkn = _scanner->getToken();
               if (tkn == T_ERROR)
                  _print(_scanner->getTokenString());
               initialized = _doLitExpr(tkn,tmpEntry,errCnt);
               tkn = _scanner->getToken();
               if (tkn == T_ERROR)
                  _print(_scanner->getTokenString());
            }

			BppValue v(vtype,vtype != BVT_NULL ? BVF_STRONGTYPED : BVF_NONE,initialized ? false : true);
			if (vtype == BVT_STRONGOBJECT && pvClass)
			{
				BppVector* pv  = new BppVector(2,true);
				(*pv)[IDX_CLASS] = pvClass;
				BppValue vdesc((BppRefObject*)pv, vtype);
				v = vdesc;
			}
			if (initialized && (!v.assign(tmpEntry.getValue())))
			{
			 _error(BppString::format(_T("Type mismatch: %s(%s) = (%s)."),
			     id.c_str(),
				 getValueTypeName(v.getType()),
				 getValueTypeName(tmpEntry.getValue().getType())).c_str(), &errCnt);
			}
			if (vtype == BVT_AUTO && (!initialized))
			{
				 _error(_T("Variables of 'auto' type must be initialized."), &errCnt);
			}
            v.setProtection(protlevel ? protlevel : BVF_PROTECTED);
            
			if (!pclass->addDataMember(id.c_str(),v))
			{
				
				if (!pclass->defineDataMember(id.c_str(),v))
				{
					  _error(_T("illegal redefinition to data member"), &errCnt);
					  errCnt++;
				}
            else _error(_T("warning: redefinition of data member"));
			}
            if ((tkn == T_ANY) && (_scanner->getChar() == _T(',')))
            {
               if (!_fetchRequire(T_IDENTIFIER)) errCnt++;
               else 
                  id = _scanner->getTokenString();
               tkn = _scanner->getToken();
            }
            else
            {
               _scanner->back();
               break;
            }
         } // end while
		 pclass->fixupDataDefs();
      }
      else // static datamember(s)
      {
         if (nullPrefix)
         {
            _error(_T("unexpected null-keyword before class variable declaration"), &errCnt);
         }
         while(1)
         {
			 BppDictionaryEntry tmpEntry(id,BppValue());
             bool initialized = false;
            if ((tkn == T_ANY) && (_scanner->getChar() == _T('=')))
            {
               tkn = _scanner->getToken();
               if (tkn == T_ERROR)
                  _print(_scanner->getTokenString());
               initialized = _doLitExpr(tkn,tmpEntry,errCnt);
               tkn = _scanner->getToken();
               if (tkn == T_ERROR)
                  _print(_scanner->getTokenString());
            }
			BppValue v(vtype,vtype != BVT_NULL ? BVF_STRONGTYPED : BVF_NONE,initialized ? false : true);
			if (vtype == BVT_STRONGOBJECT && pvClass)
			{
				BppVector* pv  = new BppVector(2,true);
				(*pv)[IDX_CLASS] = pvClass;
				BppValue vdesc((BppRefObject*)pv, vtype);
				v = vdesc;
			}
			if (initialized && (!v.assign(tmpEntry.getValue())))
			{
			 _error(BppString::format(_T("Type mismatch: %s(%s) = (%s)."),
			     id.c_str(),
				 getValueTypeName(v.getType()),
				 getValueTypeName(tmpEntry.getValue().getType())).c_str(), &errCnt);
			}
			if (vtype == BVT_AUTO && (!initialized))
			{
				 _error(_T("Variables of 'auto' type must be initialized."), &errCnt);
			}
            v.setProtection(protlevel ? protlevel : BVF_PUBLIC);
            
			if (!pclass->addStaticDataMember(id.c_str(),v))
			{
				
				if (!pclass->defineStaticDataMember(id.c_str(),v))
				{
					  _error(_T("illegal redefinition to static data member"), &errCnt);
					  errCnt++;
				}
            else _error(_T("warning: redefinition of static data member"), &errCnt);
			}
            if ((tkn == T_ANY) && (_scanner->getChar() == _T(',')))
            {
               if (!_fetchRequire(T_IDENTIFIER)) errCnt++;
               else 
                  id = _scanner->getTokenString();
               tkn = _scanner->getToken();
            }
            else
            {
               _scanner->back();
               break;
            }
         } // end while
      }
	  if (!funcdef)
	  {
         if (!_fetchRequire(T_ANY,_T(';'))) errCnt++;
	  }
   } // end while member
   tkn = _scanner->getToken();
   if ((tkn != T_ANY) || (_scanner->getChar() != _T(';')))
      _scanner->back();

   pclass->fixupDataDefs();
   return true;
}

// preprocinstruction ::= "#" ( <use> | <import> | <endsrc> | <defvar> | <deflit> | 
//                              <literaldef> | <undeflit> | <include> )
// use ::= "use" """ <modulenameorpath> """
// import ::= "import" """ <dllnameorpath> """
// endsrc ::= "endsrc"
// defvar ::= "defvar" <identifier> <literal>
// deflit ::= "deflit" <identifier> <literal>
// literaldef ::= "literal" <identifier> <literal>
// undeflit ::= "undeflit" <identifier>
// include ::= "include" ( "<" <filenameorpath> ">") | (""" <filenameorpath> """ )
// In use instruction module is searched in current directory first and then in 
// directories of BPPLIB-environment variable.
bool BppCompiler::_doPreproc(int& errCnt)
{
   /* get the preprocessor symbol */
   if (!_fetchRequire(T_IDENTIFIER))
   {
      errCnt++;
      return true;
   }
   const BppChar* tkstr = _scanner->getTokenString();
   if (_tcscmp(tkstr,_T("use")) == 0)
   {
      if (!_fetchRequire(T_STRING))
      {
         errCnt++;
         return true;
      }
      const BppChar* modulename = _scanner->getTokenString();
      const BppChar* fname = searchEnv(_T("BPPLIB"),modulename);
      if (fname == 0)
         fname = modulename;
      BppDictionaryIO reader(_vm);
      if (!reader.readModule(fname))
      {
         _error(_T("library import failed"), &errCnt);
      }
   }
   else if (_tcscmp(tkstr,_T("import")) == 0)
   {
      if (!_fetchRequire(T_STRING))
      {
         errCnt++;
         return true;
      }
      const BppChar* modulename = _scanner->getTokenString();
      if (!_vm.loadNativeModule(modulename))
      {
         _error(_T("native module import failed"), &errCnt);
      }
   }
   else if (_tcscmp(tkstr,_T("endsrc")) == 0)
   {
      return false;
   }
   else if (_tcscmp(tkstr,_T("defvar")) == 0)
   {
      if (!_fetchRequire(T_IDENTIFIER))
      {
         errCnt++;
         return true;
      }
      tkstr = _scanner->getTokenString();
      BppString* s = new BppString(tkstr);
      BppDictionaryEntry* entry = _symbols[s];
      entry->setSymbolType(ST_SDATA);
      return _doLitExpr(_scanner->getToken(),*entry,errCnt);
   }
   else if ((_tcscmp(tkstr,_T("deflit")) == 0) || (_tcscmp(tkstr,_T("define")) == 0))
   {
      if (!_fetchRequire(T_IDENTIFIER))
      {
         errCnt++;
         return true;
      }
      tkstr = _scanner->getTokenString();
      BppString* s = new BppString(tkstr);
      BppDictionaryEntry* entry = _namedLiterals[s];
      return _doLitExpr(_scanner->getToken(),*entry,errCnt);
   }
   else if (_tcscmp(tkstr,_T("literal")) == 0)
   {
      if (!_fetchRequire(T_IDENTIFIER))
      {
         errCnt++;
         return true;
      }
      tkstr = _scanner->getTokenString();
      BppString* s = new BppString(tkstr);
      BppDictionaryEntry de(s,BppValue());
      if (!_doLitExpr(_scanner->getToken(),de,errCnt)) return false;
      if (_namedLiterals.find(*s) == 0)
      {
          BppDictionaryEntry* entry = _namedLiterals[s];
          entry->setValue(de.getValue());
      }
      return true;
   }
   else if ((_tcscmp(tkstr,_T("undeflit")) == 0) || (_tcscmp(tkstr,_T("undef")) == 0))
   {
      if (!_fetchRequire(T_IDENTIFIER))
      {
         errCnt++;
         return true;
      }
      tkstr = _scanner->getTokenString();
      BppDictionaryEntry* entry = _namedLiterals.find(tkstr);
      if (entry != 0)
         _namedLiterals.erase(entry);
      return true;
   }
   else if (_tcscmp(tkstr,_T("include")) == 0)
      return _doInclude(errCnt);
   else
   {
      _error(_T("undefined preprocessor symbol"), &errCnt);
   }
   return true;
}


// include ::= "include" ( "<" <filenameorpath> ">") | (""" <filenameorpath> """ )
// First notation searches in BPPINC environment variable path for specified file.
bool BppCompiler::_doInclude(int& errCnt)
{
   BppString path;
   BppToken tkn = _scanner->getToken();
   bool isLibPath = false;
   if (tkn == T_STRING)
      path = _scanner->getTokenString();
   else if ((tkn == T_ANY) && (_scanner->getChar() == _T('<')))
   {
      isLibPath = true;
      if ((tkn = _scanner->getToken())  == T_STRING)
         path = _scanner->getTokenString();
      else
      {
         _scanner->back();
         while (((tkn = _scanner->getToken()) >= T_IDENTIFIER) && (tkn <= T_NIL))
         {
            path += _scanner->getTokenString();
            tkn = _scanner->getToken();
            if (tkn == T_ANY)
            {
               BppChar c[2];
               c[1] = 0;
               c[0] = _scanner->getChar();
               if ((c[0] == _T('/')) || (c[0] == _T('.')))
               {
                  path += (const BppChar*) c;
                  continue;
               }
            }
            _scanner->back();
            break;
         } // end while
      } // end else
      if (!_fetchRequire(T_ANY,_T('>')))
      {
         errCnt++;
         return false;
      }
   } // end !T_STRING
   if (path.length() == 0)
   {
      _error(_T("include file name expected"), &errCnt);
      return false;
   }
   BppString fullpath = path;
   
   FILE* fp = _tfopen(path.c_str(),_T("rt"));
   if (fp == 0 && (!isLibPath))
   {
      size_t n = _srcDirStack.size();
      for (size_t i = 0; (!fp) && i < n; ++i)
      {
         BppString* s = _srcDirStack[i];
         const BppChar* fpath = searchDirList(s->c_str(), path.c_str());
         if (fpath)
         {
            fullpath = fpath;
            fp = _tfopen(fullpath.c_str(), _T("rt"));
         }
      }
   }
   if (fp == 0)
   {
      const BppChar* inc = searchEnv(_T("BPPINC"),path.c_str());
      if (inc != 0)
      {
         fullpath = (const BppChar*) inc;
         fp = _tfopen(fullpath.c_str(),_T("rt"));
      }
   }
   if (fp == 0)
   {
      _scanner->parseError(_T("file not found"));
      _print(_T("%s %s"),fullpath.c_str(),_scanner->getTokenString());
      errCnt++;
      return false;
   }
   _scanner->push(fp,fullpath.c_str());
   _srcDirStack.push_front(new BppString(getFileDir(fullpath.c_str(), true)));
   return true;
}

 // globalvardef ::= <singlevardef> {"," <singlevardef> } ";". (vtype)
 // singlevardef = <identifier> [ "=" <litexpr> ]
 bool BppCompiler::_doGlobalVariable(const BppString& name, int& errCnt, BppValueType vtype, BppClass* pvClass)
 {
      BppDictionaryEntry* entry = _symbols.find(name);
	  if (entry)
			  _error(BppString::format(_T("Illegal redefinition of global symbol '%s'."),name.c_str()).c_str(), &errCnt);
	  else
	  {
		  entry = &_symbols[name];
          entry->setSymbolType(ST_SDATA);
	  }
	  if (vtype == BVT_STRONGOBJECT && pvClass)
	  {
          BppVector* desc = new BppVector(2,true);
		  (*desc)[IDX_CLASS] = pvClass;
		  entry->setValue(BppValue((BppRefObject*)desc,vtype));
	  }
	  else entry->setValue(BppValue(vtype,BVF_NONE,true));
	  if (vtype != BVT_NULL) ((BppValue&)entry->getValue()).setFlags(BVF_STRONGTYPED);
	  while (_fetchRequire(T_ANY))
	  {
		  BppChar c = _scanner->getChar();
		  switch (c)
		  {
		  case _T(';'): 
			  return (errCnt == 0);
		  case _T(','):
			  if (!_fetchRequire(T_IDENTIFIER))
			  {
				  errCnt++;
				  return false;
			  }
			  return _doGlobalVariable(_scanner->getTokenString(),errCnt,vtype,pvClass);
		  case _T('='):
		  case _T('('):
			  {
                   BppDictionaryEntry dummy(0,BppValue());
				   if (_doLitExpr(_scanner->getToken(),dummy,errCnt) &&
					   (!entry->setValue(dummy.getValue(),true)))
				   {
					 _error(BppString::format(_T("Type mismatch: %s(%s) = (%s)."),
						 name.c_str(),
						 getValueTypeName(entry->getValue().getType()),
						 getValueTypeName(dummy.getValue().getType())).c_str(), &errCnt);
				   }
                if (c == _T('(') && (!_fetchRequire(T_ANY,_T(')'))))
				   {
					  ++errCnt;
					  return false;
				   }
			  }
			  break;
		  default:
			  _require(T_ANY,T_ANY,_T(';'));
			  ++errCnt;
			  return false;
		  }
	  }
	  errCnt++;
      return false;
 }

// regularfunc ::= <identifier> <functionbody>
bool BppCompiler::_doRegularFunction(const BppString& name, int& errCnt, BppValueType rtype, BppClass* prClass)
{
	BppDictionaryEntry* sym = _symbols.find(name);
	if (sym)
	{
		if (sym->getSymbolType() != ST_SFUNCTION)
		{
			_error(_T("illegal redefinition to function"), &errCnt);
			return false;
		}
	}
	else 
	{
		sym = &_symbols[name];
        sym->setSymbolType(ST_SFUNCTION);
	}
    _scanner->_funcName = name;

   if (_showInfo)
      _print(_T("function %s\n"),name.c_str());

    /* compile the body of the function */
    BppVector* code = _codeFunctionBody(name,errCnt,0,rtype,prClass);
    _scanner->_funcName.clear();
    if (code != 0)
    /* store the function definition */
       *sym = BppValue(code,true);
   return true;
}

// memberfunc ::= <classname> "::" <identifier> <functionbody>
bool BppCompiler::_doMemberFunction(const BppString& name, int& errCnt, BppValueType rtype, BppClass* prClass)
{
    BppString* selector = 0;
   BppClass* pclass = _findClass(name,errCnt);
   if (pclass == 0) return false;
   bool opdef = false;
   /* get the selector */
   BppToken tkn = _scanner->getToken();
   if ((tkn == T_ANY) && (_scanner->getChar()==_T('~')))
   {
      if (!_fetchRequire(T_IDENTIFIER)) 
         errCnt++;
      else if (_tcscmp(_scanner->getTokenString(),name.c_str()) != 0)
      {
         _error(_T("illegal destructor definition"), &errCnt);
      }
      selector = new BppString(_T("dtor"));
      selector->addRef();
   }
   else
   {
      if (tkn == T_OPERATOR)
         opdef = true;
      if (!(opdef || _require(T_IDENTIFIER,tkn)))
         errCnt++;
      else
      {
         selector = new BppString(opdef ? _getOperatorFunc(errCnt) : _scanner->getTokenString());
         selector->addRef();
	     if (opdef && errCnt == 0)
		 {
			 BppClass* pvClass = 0;
		    BppValueType vtype = _vm.findVTypeFromName(*selector,true,&pvClass);
			if (vtype != BVT_NULL && rtype != BVT_NULL)
			{
			  _error(_T("illegal type before conversion operator"), &errCnt);
			}
			else if (vtype != BVT_NULL) { rtype = vtype, prClass = pvClass; }
		 }
      }
   }
   if (selector == 0)
      return false;
   if (!(opdef || _fetchRequire(T_ANY,_T('(')))) ++errCnt;
   if (_showInfo)
      _print(_T("member function %s::%s\n"),pclass->getClassName()->c_str(),selector->c_str());
 
   BppDictionaryEntry* entry = pclass->findMemberPL(BVF_PRIVATE,*selector,false);
    /* make sure the type matches the declaration */
    if (entry != NULL)
    {
       BppSymbolType st = entry->getSymbolType();
       if ((st != ST_FUNCTION) && (st != ST_SFUNCTION))
       {
          _error(_T("illegal redefinition"), &errCnt);
       }
    }
    else // new entry
    {
       entry = pclass->funcMembers()[selector];
       entry->setSymbolType(ST_FUNCTION);
    }
    _scanner->_funcName = BppString::format(_T("%s::%s"),pclass->getClassName()->c_str(),selector->c_str());
    /* compile the body of the function */
    BppVector* code = _codeFunctionBody(*selector,errCnt,pclass,rtype,prClass);
    selector->release();
    _scanner->_funcName.clear();
    //if (code == 0) return false;


	if (code)
	{
		/* store the new method */
		unsigned short protlevel = entry->getValue().getProtection();
		entry->setValue(BppValue(code,true));
		((BppValue&) entry->getValue()).setProtection(protlevel ? protlevel : BVF_PUBLIC);
		/* build OP_PINC / OP_PDEC methods if needed */
		const BppString& key = *(entry->getKey());
		selector = 0;
		if (key.compare(_T("OP_INC")) == 0)
		   selector = new BppString(_T("OP_PINC"));
		else if (key.compare(_T("OP_DEC")) == 0)
		   selector = new BppString(_T("OP_PDEC"));
		if (selector != 0)
		{
         protlevel = entry->getValue().getProtection();
			selector->addRef();
			BppDictionaryEntry* pentry = pclass->findMemberPL(BVF_PRIVATE,*selector,false);
			if (pentry == 0) // add post increment/decrement method if it not exists
			{
			   pentry = pclass->funcMembers()[selector];
			   pentry->setSymbolType(ST_FUNCTION);
			   pentry->setValue(BppValue(code,true));
			   ((BppValue&) pentry->getValue()).setProtection(protlevel ? protlevel : BVF_PUBLIC);
			}
			selector->release();
		}
	}
   return true;
}




// functionbody ::= "(" [<argumentlist>][ ";" <localvarlist> ] ")" <blockstatement>
// argumentlist ::= <idlist>
// localvarlist ::= <idlist>
BppVector* BppCompiler::_codeFunctionBody(const BppString& name, int& errCnt, BppClass* ownerClass, BppValueType rtype, BppClass* prClass)
{
   BppFuncContext context;
   context.returnType = rtype;
   context.prClass = prClass;
   BppVector* defargs = new BppVector();
   bool isdtor = false;
    /* add the implicit 'this' argument for member functions */
    if (ownerClass != NULL)
    {
	   BppDictionaryEntry* pe = ownerClass->findMemberPL(BVF_PRIVATE,name,false);
	   if (!(pe && (pe->getSymbolType() == ST_SFUNCTION)))
	   {
		   isdtor = (_tcscmp(name.c_str(),_T("dtor")) == 0);
		   context.arguments.push_back(new BppString(_T("this")));
		   defargs->push_back(BppValue());
	   }
    }
    /* get the argument list */
    BppValue dummy = defargs;
    _getIdList(context.arguments,_T(";)"),errCnt,defargs);
	if (isdtor && context.arguments.size() > 1)
	{
        _error(_T("Destructor can't have arguments."), &errCnt);
	}
    /* get temporary variables, at least 1 for exception value needed */
    int tcnt = 1;
    BppToken tkn = _scanner->getToken();
    if ((tkn == T_ANY) && (_scanner->getChar() == _T(';')))
    {
            // add 1 to temporary count for exception value)
            tcnt = _getIdList(context.temporaries,_T(")"),errCnt) +1;
       tkn = _scanner->getToken();
    }
    if (!_require(T_ANY,tkn,_T(')')))
    {
       errCnt++;
       return 0;
    }
	tkn = _scanner->getToken();
	if (tkn == T_ANY && _scanner->getChar() == _T(';')) // forward deklaration
	{
		return 0;
	}
	_scanner->back();
    context.codeBuffer = new BppBuffer();
    context.codeVec = new BppVector();
    BppVector* result = context.codeVec;
    // store the bytecodes, class and function name as the first literals 
    result->push_back(BppValue(context.codeBuffer)); // buffer for bytecodes
    if (ownerClass != 0)
        result->push_back(BppValue(ownerClass));        // class (may be NULL!)
    else
       result->push_back(BppValue());
    context.contextClass = ownerClass;
    result->push_back(new BppString(name)); // function name
    result->push_back(defargs); // vector of default argument values
    result->push_back(BppValue(result,true,true)); // self reference
    if (tcnt > 0) // reserve space for temporaries
    {
       _codeIndexedOpcode(*result,OP_TSPACE,(BppNWord) tcnt, true);
    }


    context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));

    if (!_fetchRequire(T_ANY,_T('{'))) errCnt++;
    // compile statement block
    _codeBlock(context,errCnt);

    if (ownerClass != 0)
    {
       if (_tcscmp(ownerClass->getClassName()->c_str(),name.c_str()) == 0)
                {
    //       putcbyte(OP_PUSH);
         BppVarInfo varInfo = _findVariable(_T("this"),context,false,errCnt);
         // return 'this' pointer on TOS in constructor
         _codeRValue(*result,varInfo);
            }
       else if ((ownerClass->getBaseClass() != 0) && (_tcscmp(_T("dtor"),name.c_str()) == 0))
            {
          // push 'this' pointer
          context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));
         BppVarInfo varInfo = _findVariable(_T("this"),context,false,errCnt);
         _codeRValue(*result,varInfo);
         // push call to 'BC_dtor' (magic name of inherited dtor)
          context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));
          result->push_back(new BppString(_T("BC_dtor")));
          _codeLiteral(*result,result->size() - 1);
          context.codeBuffer->push_back(BppOpCodeByte(OP_SEND,ODT_BYTE));
          context.codeBuffer->push_back(IDX_CLASS);
                }
        }
    // insert return statement if code doesn't already end with return
    size_t bsize = context.codeBuffer->size();
    if ((bsize < 2) ||
       ((BppOpCodeByte((*context.codeBuffer)[bsize - 1]).opCode() != OP_RETURN) &&
       (BppOpCodeByte((*context.codeBuffer)[bsize - 2]).opCode() != OP_RETURN)))
    {
       if (context.returnType != BVT_NULL)
       {
          if (context.returnType == BVT_VOID)
          {
             context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH, ODT_BYTE));
             context.codeBuffer->push_back(BVT_VOID);
          }
          context.codeBuffer->push_back(BppOpCodeByte(OP_RETURN, ODT_BYTE));
          context.codeBuffer->push_back(context.returnType);
       }
       else
          context.codeBuffer->push_back(BppOpCodeByte(OP_RETURN));
    }
    if (_genLocalVarInfo)
    {
       /*
          Erzeugt wird ein Vector mit zwei Elementen.
          Das erste Element ist ein Vektor mit den Namen aller
          formalen Parameter, das zweite ein Vektor mit den Namen
          der lokalen (temporren) Variablen. Die Reihenfolge der
          Vektorelemente entspricht jeweils der spteren Ablage
          der lokalen Variablen auf dem Stack.
          Existieren zustzlich statische lokale Variablen, so wird an den
          zweiten Vektor ein zustzliches Element vom Dictionary-Typ
          angehngt, das die Zuordnungen der Namen der statischen lokalen
          Variablen zu den Indizes im zurckgelieferten Bytecode-Vektor
          enthlt.
       */
       BppVector* info = new BppVector(2,true);
       (*info)[0] = new BppVector(context.arguments);
       BppVector* tmpVec = new BppVector(context.temporaries);

       (*info)[1] = tmpVec;
       if (context.statics.count() > 0)
       {
          BppDictionary* pstat = new BppDictionary(true);
          for (BppDictionaryEntry* p = context.statics.getFirst();
             p != 0; p=p->getNext())
          {
             BppDictionaryEntry* pnew = (*pstat)[p->getKey()];
             pnew->setValue(p->getValue());
          }
          tmpVec->push_back(pstat);
       }

       result->push_back(BppValue(info,true));
    }
     // free unused buffer space
    result->shrink();
    context.codeBuffer->shrink();
	if (context.statics.count() > 0)
		(*result)[IDX_CODE].setFlags(BVF_STATIC);
    /* show the generated code */
    if (_debugger != 0)
       _debugger->decodeProcedure(result);
    /* return the code object */
    return (result);
}

// statementblock ::= "{" { <preprocinstruction | <statement> } "}"
void BppCompiler::_codeBlock(BppFuncContext& context, int& errCnt)
{
   BppToken tkn = _scanner->getToken();
   if ((tkn != T_ANY) || (_scanner->getChar() != _T('}')))
   {
      do
      {
         while ((tkn == T_ANY) && (_scanner->getChar() == _T('#')))
         {
            if (!_doPreproc(errCnt))
               break;
			else
				tkn = _scanner->getToken();
			
         }
           _scanner->back();
         _codeStatement(context,errCnt);
         tkn = _scanner->getToken();
         if (tkn == T_EOF)
         {
            _error(_T("unexpected EOF in block"), &errCnt);
            break;
         }
      } while ((tkn != T_ANY) || (_scanner->getChar() != _T('}')));
   }
   else
   {
      context.codeBuffer->push_back(BppOpCodeByte(OP_NIL));
   }
}

void BppCompiler::_codeSourceInfo(BppFuncContext& context, int&)
{
   bool newline = false;
   const BppChar* sname = _scanner->getSourceName();
   long lnum = _scanner->getLineNumber();
   if (_tcscmp(_currentSource.c_str(),sname) != 0)
   {
      newline = true;
      _currentSource = sname;
   }
   if (lnum != _currentLine)
   {
      newline = true;
      _currentLine = lnum;
   }
   if (!newline) return;
   int idx = context.codeVec->getStringIndex(sname);
   if (idx < 0)
   {
      context.codeVec->push_back(new BppString(sname));
      idx = (int) context.codeVec->size() - 1;
   }
   context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));
   _codeLiteral(*(context.codeVec),(BppNWord)idx);
   _codeIndexedOpcode(*(context.codeVec),OP_NEWLINE,(BppNWord)lnum);
}



// statement ::= <blockstatement> | <if> | <while> | <dowhile> | <for> | 
// <break> | <continue> | <switch> | <return> | 
// ("TRON" ";") | ("TROFF" ";") | ("TRSTEP" ";") | 
// <try> | <throw> | (<staticvardefinition> | <varstatement> | [<expr>]   ";" ) .
// <varstatement> ::= ("var" | "function" | <typename> | <classname> ) <vardefinition> .
// typename = "int" | "float" | "double" | "FILE" | "buffer" | "charbuffer" |
// "object" | "vector" | "string" | "dictionary" .
// classname::= (identifier of an already defined class)
void BppCompiler::_codeStatement(BppFuncContext& context, int& errCnt)
{
	if (_genLineInfo)
		_codeSourceInfo(context,errCnt);
	BppToken tkn = _scanner->getToken();
	//BppBuffer& buf = *context.codeBuffer;
	switch (tkn) 
	{
	case T_IF:           _codeIf(context,errCnt);        break;
	case T_WHILE:        _codeWhile(context,errCnt);     break;
	case T_DO:           _codeDoWhile(context,errCnt);   break;
	case T_FOR:          _codeFor(context,errCnt);       break;
	case T_BREAK:        _codeBreak(context,errCnt);     break;
	case T_CONTINUE:     _codeContinue(context,errCnt);  break;
	case T_SWITCH:       _codeSwitch(context,errCnt);    break;
	case T_RETURN:       _codeReturn(context,errCnt);    break;
	case T_TRY:          _codeTry(context,errCnt);       break;
	case T_THROW:        _codeThrow(context,errCnt);     break;
	case T_VAR:          
		_codeVarDefinition(context,errCnt);
		if (!_fetchRequire(T_ANY,_T(';'))) errCnt++;
		break;   
	case T_STATIC:
		{
			BppValueType vtype = BVT_NULL;
            BppToken tnext = _scanner->getToken();
			BppString str = _scanner->getTokenString();
			if (tnext == T_IDENTIFIER || tnext == T_FUNCTION)
			{
			    BppClass* pclass = 0;
				vtype = _vm.findVTypeFromName(str,true,&pclass);
                if (vtype == BVT_NULL)
					_scanner->back();
			}
			else
			{
				_scanner->back();
			}
			_codeStaticVarDefinition(context,errCnt,vtype);
			if (!_fetchRequire(T_ANY,_T(';'))) errCnt++;
		}
		break;   
	case T_TRON: 
		_codeOpcode(*context.codeVec,OP_TRON);
		            //buf.push_back(BppOpCodeByte(OP_TRON));
					if (!_fetchRequire(T_ANY,_T(';'))) errCnt++; 
					break;
	case T_TROFF:
		_codeOpcode(*context.codeVec,OP_TROFF);
		//buf.push_back(BppOpCodeByte(OP_TROFF))
		if (!_fetchRequire(T_ANY,_T(';'))) errCnt++; 
		break;
	case T_TRSTEP:  
		_codeOpcode(*context.codeVec,OP_TRSTEP);
		//buf.push_back(BppOpCodeByte(OP_TRSTEP));
		if (!_fetchRequire(T_ANY,_T(';'))) errCnt++; 
		break;
	case T_ANY:
		{
			BppChar c = _scanner->getChar();
			if (c == _T('{'))
			{
				_codeBlock(context,errCnt);
				return;
			}
			if (c == _T(';'))
				return;
		}
	case T_IDENTIFIER:
	case T_FUNCTION:
	case T_AUTO:
		{
			BppString str = _scanner->getTokenString();
			BppClass* pclass = 0;
			BppValueType vtype = _vm.findVTypeFromName(str,true,&pclass);
			if (vtype != BVT_NULL)
			{
				_scanner->back(); 
				BppToken tnext = _scanner->getToken(true);
				if (tnext == T_IDENTIFIER)
				{
					_scanner->back(true); // save tnext
					_scanner->getToken(); // pop type name token
					_codeVarDefinition(context,errCnt,vtype,pclass);
					if (!_fetchRequire(T_ANY,_T(';'))) errCnt++;
					break;   
				}
				else if (tkn == T_FUNCTION)
				{
					_require(T_IDENTIFIER,tnext);
					++errCnt;
					break;
				}
			}
		}
	default:   
		{
			_scanner->back(true);
			//size_t bsize = buf.size();
			_codeExpr(context,errCnt);
			//_adjustAssignment(buf,bsize);
			if (!_fetchRequire(T_ANY,_T(';'))) errCnt++;
			break;
		}
	}
}

// <varstatement> := "var" <identifier> { "," <identifier> } [ "=" <expr2> ] .
// <vardef> ::= <identifier> [ "=" ]
void BppCompiler::_codeVarDefinition(BppFuncContext& context, int& errCnt, BppValueType defType, BppClass* reqClass)
{
	BppString tname;
	if (defType != BVT_NULL) 
		tname = _scanner->getTokenString(); 
   static const size_t sizeidx = 1;
   BppBuffer* buf = context.codeBuffer;
   BppOpCodeByte firstOpCode = (*buf)[0];
   if (firstOpCode.opCode() != OP_TSPACE) // sollte nie passieren (vgl. _codeFunctionBody)
   {
      _error(_T("No local variables space in current function"), &errCnt);
   }
   BppToken t = _scanner->getToken();
   // walk through definition list
   while (1)
   {
      if (!_require(T_IDENTIFIER,t))
      {
         ++errCnt;
         break;
      }
      BppString name = _scanner->getTokenString();
      if (context.statics.find(name))
      {
         _error(BppString::format(_T("Static variable of name '%s' already defined"),name.c_str()).c_str(), &errCnt);
      }
      if (context.arguments.getStringIndex(name.c_str()) >= 0)
      {
         _error(BppString::format(_T("Argument of name '%s' already defined"),name.c_str()).c_str(), &errCnt);
      }
      if (context.temporaries.getStringIndex(name.c_str()) >= 0)
      {
         _error(BppString::format(_T("Local variable of name '%s' already defined"),name.c_str()).c_str(), &errCnt);
      }
      context.temporaries.push_back(new BppString(name));

      t = _scanner->getToken();
      if ((t == T_ANY) && (_scanner->getChar() == _T('=')))
      {
         BppVarInfo varInfo;
         varInfo.index = context.temporaries.size()-1;
         varInfo.type = VAR_LOCAL;
		 varInfo.varvtype = defType;
		 varInfo.reqClass = reqClass;
       if (defType == BVT_STRONGOBJECT && reqClass)
       {
          //BppVector* desc = new BppVector(2, true);
          //(*desc)[IDX_CLASS] = reqClass;
          context.codeVec->push_back(BppValue((BppRefObject*)new BppObject(reqClass), BVT_OBJECT));
          BppVarInfo vinfo;
          vinfo.type = VAR_SLOCAL;
          vinfo.varvtype = BVT_OBJECT;
          vinfo.reqClass = reqClass;
          vinfo.index = context.codeVec->size() - 1;
          //context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));

          //_codeStaticLocal(*context.codeVec, VA_LOAD, vinfo.index);
          _codeVarInfo(*context.codeVec, VA_LOAD, vinfo); // load empty strong object
          vinfo = varInfo;
          //_codeTemporary(*context.codeVec, VA_STORE, vinfo.index);
          //context.codeBuffer->push_back(BppOpCodeByte(OP_POP));
           _codeVarInfo(*context.codeVec, VA_STORE, vinfo); // store result
       }
       BppVarInfo rInfo;
         _codeExpr2_assign(context,rInfo,errCnt);
         _codeRValue(*context.codeVec,rInfo);

         _codeVarInfo(*context.codeVec,VA_STORE,varInfo); // store result
      }
	  else if ((t == T_ANY) && (_scanner->getChar() == _T('(')))
	  {
		  _scanner->back();
         BppVarInfo varInfo;
         varInfo.index = context.temporaries.size()-1;
         varInfo.type = VAR_LOCAL;
		 varInfo.varvtype = defType;
		 BppVarInfo rInfo;
			  _scanner->_token = T_IDENTIFIER;
		  switch (defType)
		  { 
		  case BVT_OBJECT:
		  case BVT_STRONGOBJECT:
			  if (_tcscmp(tname.c_str(),_T("object")) != 0)
			  {
				  _scanner->_stringBuffer = tname.c_str();
				  _scanner->back();
				  _codeNew(context,rInfo,errCnt); 
			  }
			  else
			  {
				 _scanner->getToken(); // eat '('
				 _codeExpr1_comma(context,rInfo,errCnt);
				 if (!_fetchRequire(T_ANY,_T(')'))) errCnt++;
			  }
			  break;
		  case BVT_SBYTE:
		  case BVT_BYTE:
		  case BVT_SHORT:
		  case BVT_USHORT:
		  case BVT_INT:
		  case BVT_UINT:
		  case BVT_LONG:
		  case BVT_ULONG:
		  case BVT_INTPTR:
		  case BVT_FLOAT:
		  case BVT_DOUBLE:
		  case BVT_STRING:
		  case BVT_BUFFER:
		  case BVT_CHARBUFFER:
		  case BVT_VECTOR:
		  case BVT_DICTIONARY:
			  _scanner->_stringBuffer = tname.c_str();
			  _scanner->back();
              _codeExpr15_post(context,rInfo,errCnt); 
			  break;
		  case BVT_FILE:
			  _scanner->_stringBuffer = _T("fopen");
			  _scanner->back();
              _codeExpr15_post(context,rInfo,errCnt); 
			  break;
		  default:
			 _scanner->getToken(); // eat '('
			 _codeExpr1_comma(context,rInfo,errCnt);
			 if (!_fetchRequire(T_ANY,_T(')'))) errCnt++;
		  }
		  _codeRValue(*context.codeVec,rInfo);
		  _codeVarInfo(*context.codeVec,VA_STORE,varInfo); // store result
	  }
      else
	  {
		  _scanner->back();
		  if (defType != BVT_NULL)
		  {
			  BppVarInfo varInfo;
			  varInfo.index = context.temporaries.size()-1;
			  varInfo.type = VAR_LOCAL;
		      varInfo.varvtype = defType;
			  BppVarInfo rInfo;
			  switch (defType)
			  {
			  case BVT_SBYTE:
			  case BVT_BYTE:
			  case BVT_SHORT:
			  case BVT_USHORT:
			  case BVT_INT:
			  case BVT_UINT:
			  case BVT_LONG:
			  case BVT_ULONG:
			  case BVT_INTPTR:
			  case BVT_FLOAT:
			  case BVT_DOUBLE:
				  context.codeVec->push_back(BppValue(defType,BVF_NONE,false));
				  _codeLiteral(*context.codeVec,context.codeVec->size()-1);
				  rInfo.type = VAR_DONE;
				  break;
			  case BVT_BUFFER:
				  context.codeVec->push_back(new BppBuffer());
				  _codeLiteral(*context.codeVec,context.codeVec->size()-1);
				  rInfo.type = VAR_DONE;
				  break;
			  case BVT_CHARBUFFER:
				  context.codeVec->push_back(new BppCharBuffer());
				  _codeLiteral(*context.codeVec,context.codeVec->size()-1);
				  rInfo.type = VAR_DONE;
				  break;
			  case BVT_VECTOR:
				  context.codeVec->push_back(new BppVector());
				  _codeLiteral(*context.codeVec,context.codeVec->size()-1);
				  rInfo.type = VAR_DONE;
				  break;
			  case BVT_DICTIONARY:
				  context.codeVec->push_back(new BppDictionary());
				  _codeLiteral(*context.codeVec,context.codeVec->size()-1);
				  rInfo.type = VAR_DONE;
				  break;
			  case BVT_FILE:
				  context.codeVec->push_back(BppValue((FILE*)0));
				  _codeLiteral(*context.codeVec,context.codeVec->size()-1);
				  rInfo.type = VAR_DONE;
				  break;
			  case BVT_OBJECT:
			  case BVT_STRONGOBJECT:
				  if (_tcscmp(tname.c_str(),_T("object")) != 0)
				  {
					  _scanner->_token = T_ANY;
					  _scanner->_charBuffer[0] = _T(')');
					  _scanner->back();
					  _scanner->_charBuffer[0] = _T('(');
					  _scanner->back();
					  _scanner->_stringBuffer = tname.c_str();
					  _scanner->_token = T_IDENTIFIER;
					  _scanner->back();
					  _codeNew(context,rInfo,errCnt); 
				  }
				  else
				  {
					  context.codeVec->push_back(BppValue());
					  _codeLiteral(*context.codeVec,context.codeVec->size()-1);
					  rInfo.type = VAR_DONE;
				  }
				  break;
			  case BVT_STRING:
				  _scanner->_stringBuffer = _T("");
				  _scanner->_token = T_STRING;
				  _scanner->back();
				  _codeExpr15_post(context,rInfo,errCnt); 
				  break;
			  case BVT_AUTO:
				 _error(_T("Variables of 'auto' type must be initialized."),&errCnt);
			  default:
				  context.codeVec->push_back(BppValue());
				  _codeLiteral(*context.codeVec,context.codeVec->size()-1);
				  rInfo.type = VAR_DONE;
				  break;
			  }
			  _codeRValue(*context.codeVec,rInfo);
			  _codeVarInfo(*context.codeVec,VA_STORE,varInfo); // store result
		  }
	  }
      t = _scanner->getToken();
      if ((t == T_ANY) && (_scanner->getChar() == _T(',')))
      {
         t = _scanner->getToken();
      }
      else
      {
         _scanner->back();
         break;
      }
   }
   unsigned varcnt = context.temporaries.size();

      buf->setNW((size_t) sizeidx,(BppNWord) varcnt);
}

// <staticvarstatement> := "static" ["var" | <typename>] <identifier> [ "=" literal ] "," <identifier> [ "=" literal ]}  .
void BppCompiler::_codeStaticVarDefinition(BppFuncContext& context, int& errCnt, BppValueType defType, BppClass* defClass)
{
	BppString tname;
	if (defType != BVT_NULL) 
		tname = _scanner->getTokenString(); 
   BppToken t = _scanner->getToken();
   if (defType == BVT_NULL && t == T_VAR)
      t = _scanner->getToken();
   while (1)
   {
      if (!_require(T_IDENTIFIER,t))
      {
         ++errCnt;
         break;
      }
      BppString name = _scanner->getTokenString();
      if (context.statics.find(name))
      {
         _error(BppString::format(_T("Static variable of name '%s' already defined"),name.c_str()).c_str(), &errCnt);
      }
      if (context.arguments.getStringIndex(name.c_str()) >= 0)
      {
         _error(BppString::format(_T("Argument of name '%s' already defined"),name.c_str()).c_str(), &errCnt);
      }
      if (context.temporaries.getStringIndex(name.c_str()) >= 0)
      {
         _error(BppString::format(_T("Local variable of name '%s' already defined"),name.c_str()).c_str(), &errCnt);
      }
      unsigned varcnt = context.codeVec->size();
	  if (defType == BVT_STRONGOBJECT && defClass)
	  {
		  BppVector* desc = new BppVector(2,true);
		  (*desc)[IDX_CLASS] = defClass;
		  context.codeVec->push_back(BppValue((BppRefObject*)desc,defType));
	  }
	  else 
		  context.codeVec->push_back(BppValue(defType,BVF_NONE,true));
      context.statics[name].setValue((BppInt)(varcnt));
	  (*context.codeVec)[varcnt].setFlags(defType == BVT_NULL ? BVF_STATIC : BVF_STATIC | BVF_STRONGTYPED);
      t = _scanner->getToken();
	  BppChar chr = _scanner->getChar();
      if ((t == T_ANY) && ( (chr == _T('=')) || (chr == _T('('))  ))
      {
         BppDictionaryEntry dummy(0,BppValue());
         _doLitExpr(_scanner->getToken(),dummy,errCnt);
         varcnt = context.codeVec->size();
		 if (!(*(context.codeVec))[varcnt-1].assign(dummy.getValue()))
		 {
			 _error(BppString::format(_T("Type mismatch: %s(%s) = (%s)."),
			     name.c_str(),
				 getValueTypeName((*(context.codeVec))[varcnt-1].getType()),
				 getValueTypeName(dummy.getValue().getType())).c_str(), &errCnt);
		 }
		 if (chr == _T('(') && (!_fetchRequire(T_ANY,_T(')')))) ++errCnt;
      }
      else
	  {
		  if (defType == BVT_AUTO)
		  {
		     _error(_T("Variables of 'auto' type must be initialized."), &errCnt);
		  }
         _scanner->back();
	  }
      t = _scanner->getToken();
      if ((t == T_ANY) && (_scanner->getChar() == _T(',')))
      {
         t = _scanner->getToken();
      }
      else
      {
         _scanner->back();
         break;
      }
   }
}
// throw ::= "throw" <expr> ";"
void BppCompiler::_codeThrow( BppFuncContext& context, int& errCnt )
{
   BppBuffer& buf = *context.codeBuffer;
   if (context.catchLevel)
   {
      // inside of catch block throw may be called without expression
      BppToken t = _scanner->getToken();
      if ((t == T_ANY) && (_scanner->getChar() == _T(';')))
      {
         // throw last exception again
         _codeTemporary(*context.codeVec,VA_LOAD,IDX_EXCEPTION);
         buf.push_back(BppOpCodeByte(OP_THROW));
         return;
      }
      _scanner->back();
   }
   _codeExpr(context,errCnt);
   buf.push_back(BppOpCodeByte(OP_THROW));
   if (!_fetchRequire(T_ANY,_T(';'))) errCnt++;
}

// try ::= "try" <blockstatement> "catch" "(" <identifier> ")" <blockstatement> 
void BppCompiler::_codeTry( BppFuncContext& context, int& errCnt )
{
   BppBuffer& buf = *context.codeBuffer;
   buf.push_back(BppOpCodeByte(OP_TRY,ODT_NWORD));
   BppNWord nxt = (BppNWord) buf.push_backNW(0); // place holder for jump offset

   if (!_fetchRequire(T_ANY,_T('{'))) errCnt++;
   BppValue savedExeptionName = context.temporaries[IDX_EXCEPTION];
   
   // compile protected block   
   _codeBlock(context,errCnt);
   buf.push_back(BppOpCodeByte(OP_ENDTRY,ODT_NWORD));
   BppNWord end = (BppNWord) buf.push_backNW(0); // place holder for jump offset
   _fixupRefChain(context,nxt,buf.size());
   if (!_fetchRequire(T_CATCH)) errCnt++;
   if (!_fetchRequire(T_ANY,_T('('))) errCnt++;
   if (!_fetchRequire(T_IDENTIFIER)) errCnt++;
   else
     context.temporaries[IDX_EXCEPTION] = new BppString(_scanner->getTokenString());
   if (!_fetchRequire(T_ANY,_T(')'))) errCnt++;
   if (!_fetchRequire(T_ANY,_T('{'))) errCnt++;
   context.catchLevel++;
   // compile exception handler block
   _codeBlock(context,errCnt); 
   buf.push_back(BppOpCodeByte(OP_ENDCATCH));
   context.catchLevel--; 
   _fixupRefChain(context,end,buf.size());
   context.temporaries[IDX_EXCEPTION]=savedExeptionName;
}

// if ::= "if" <condition> <statement> [ "else" <statement> ]
void BppCompiler::_codeIf(BppFuncContext& context, int& errCnt)
{
   /* compile the test expression */
   _codeCondition(context,errCnt);

   BppBuffer& buf = *context.codeBuffer;
   /* skip around the 'then' clause if the expression is false */
   buf.push_back(BppOpCodeByte(OP_BRF,ODT_NWORD));
   BppNWord nxt = (BppNWord) buf.push_backNW(0); // place holder for jump offset

   /* compile the 'then' clause */
   _codeStatement(context,errCnt);

   /* compile the 'else' clause */
   BppToken tkn = _scanner->getToken();
   if (tkn == T_ELSE) 
   {
      buf.push_back(BppOpCodeByte(OP_BR,ODT_NWORD));
      BppNWord end = (BppNWord) buf.push_backNW(0); // place holder for jump offset
      _fixupRefChain(context,nxt,buf.size());
      _codeStatement(context,errCnt);
      nxt = end;
   }
   else
      _scanner->back();

   /* handle the end of the statement */
   _fixupRefChain(context,nxt,buf.size());
}

// while ::= "while" <condition> <statement>
void BppCompiler::_codeWhile(BppFuncContext& context, int& errCnt)
{
   BppBuffer& buf = *context.codeBuffer;

   /* compile the test expression */
   BppNWord nxt = buf.size();
   _codeCondition(context,errCnt);

   /* skip around the loop body if the expression is false */
   buf.push_back(BppOpCodeByte(OP_BRF,ODT_NWORD));
   BppNWord end = (BppNWord) buf.push_backNW(0); // place holder for jump offset

   /* compile the loop body */
   BppNWord oldBreak = context.pushBreak(end);
   BppNWord oldCont = context.pushContinue(nxt);

   _codeStatement(context,errCnt);
   end  = context.popBreak(oldBreak,end);
   context.popContinue(oldCont);

   /* branch back to the start of the loop */
   buf.push_back(BppOpCodeByte(OP_POPC));
   buf.push_back(BppOpCodeByte(OP_BR,ODT_NWORD));
   buf.push_backNW(nxt);

      /* handle the end of the statement */
      _fixupRefChain(context,end,buf.size());
}

// dowhile ::= "do" <statement> "while" <condition> ";"
void BppCompiler::_codeDoWhile(BppFuncContext& context, int& errCnt)
{
   BppBuffer& buf = *context.codeBuffer;

   /* remember the start of the loop */
   BppNWord nxt = buf.size();

   /* compile the loop body */
   BppNWord oldBreak = context.pushBreak(0);
   BppNWord oldCont = context.pushContinue(nxt);

   _codeStatement(context,errCnt);
   BppNWord end = context.popBreak(oldBreak,0);
   context.popContinue(oldCont);

   /* compile the test expression */
   if (!_fetchRequire(T_WHILE)) errCnt++;
   _codeCondition(context,errCnt);
   if (!_fetchRequire(T_ANY,_T(';'))) errCnt++;

   /* branch to the top if the expression is true */
   buf.push_back(BppOpCodeByte(OP_POPC));
   buf.push_back(BppOpCodeByte(OP_BRT,ODT_NWORD));
   buf.push_backNW(nxt);

   /* handle the end of the statement */
   _fixupRefChain(context,end,buf.size());
}

// for ::= "for" "(" <varstatement> | <expr> ";" <expr> ";" <expr> ")" <statement> .
void BppCompiler::_codeFor(BppFuncContext& context, int& errCnt)
{
   BppBuffer& buf = *context.codeBuffer;

   /* compile the initialization expression */
   if (!_fetchRequire(T_ANY,_T('('))) errCnt++;
   BppToken tkn = _scanner->getToken();
   switch (tkn)
   {
   case T_VAR:
      _codeVarDefinition(context,errCnt);
	  break;
   case T_IDENTIFIER:
   case T_AUTO:
		{
			BppString str = _scanner->getTokenString();
			BppClass* pclass = 0;
			BppValueType vtype = _vm.findVTypeFromName(str,true,&pclass);
			if (vtype != BVT_NULL)
			{

				_scanner->back(); // save type name token
				BppToken tnext = _scanner->getToken(true);
				if (tnext == T_IDENTIFIER)
				{
   				    _scanner->back(true); // save tnext
					_scanner->getToken(); // pop type name token
					_codeVarDefinition(context,errCnt,vtype,pclass);
					break;   
				}
			}
		}
   case T_ANY:
	   if (_scanner->getChar() == _T(';')) {
		   _scanner->back();
		   break;
	   }
   default:
      _scanner->back(true);
      _codeExpr(context,errCnt);
   }

   /* compile the test expression */
   if (!_fetchRequire(T_ANY,_T(';'))) errCnt++;
   BppNWord nxt = buf.size();
   tkn = _scanner->getToken();
   if (tkn == T_ANY && _scanner->getChar() == _T(';'))
   {
	   _scanner->back();
	   buf.push_back(BppOpCodeByte(OP_BR, ODT_NWORD));
   }
   else
   {
	   _scanner->back();
	   _codeExpr(context, errCnt);
	   /* branch to the loop body if the expression is true */
	   buf.push_back(BppOpCodeByte(OP_BRT, ODT_NWORD));
   }
   BppNWord body = (BppNWord) buf.push_backNW(0);

   /* branch to the end if the expression is false */
   buf.push_back(BppOpCodeByte(OP_BR,ODT_NWORD));
   BppNWord end = (BppNWord) buf.push_backNW(0);

   /* compile the update expression */
   if (!_fetchRequire(T_ANY,_T(';'))) errCnt++;
   BppNWord update = buf.size();
   tkn = _scanner->getToken();
   if (tkn != T_ANY|| _scanner->getChar() != _T(')'))
   {
	   _scanner->back();
	   _codeExpr(context, errCnt);
	   if (!_fetchRequire(T_ANY, _T(')'))) errCnt++;
   }

   /* branch back to the test code */
   buf.push_back(BppOpCodeByte(OP_BR,ODT_NWORD));
   buf.push_backNW(nxt);

   /* compile the loop body */
   _fixupRefChain(context,body,buf.size());
   BppNWord oldBreak = context.pushBreak(end);
   BppNWord oldCont = context.pushContinue(update);
   _codeStatement(context,errCnt);
   end = context.popBreak(oldBreak,end);
   context.popContinue(oldCont);

   /* branch back to the update code */
   buf.push_back(BppOpCodeByte(OP_POPC)); 
   buf.push_back(BppOpCodeByte(OP_BR,ODT_NWORD));
   buf.push_backNW(update);
   /* handle the end of the statement */
   _fixupRefChain(context,end,buf.size());
}

void BppCompiler::_codeSwitch(BppFuncContext& context, int& errCnt)
{

   BppBuffer& buf = *context.codeBuffer;

   /* compile the test expression */
   _codeCondition(context,errCnt);
   if (!_fetchRequire(T_ANY,_T('{'))) errCnt++;

   buf.push_back(BppOpCodeByte(OP_SWITCH,ODT_NWORD));
   BppNWord end = (BppNWord) buf.push_backNW(0); // place holder for jump offset
   buf.push_back(BppOpCodeByte(OP_PUSH)); // reserve space for second value to compare
   BppNWord nxt = 0;
   BppNWord nxtBlock = 0;
   BppNWord cdefault = 0;
   BppNWord oldBreak = context.pushBreak(end);
   bool statementAllowed = false;
   BppToken tkn;
   while ((tkn = _scanner->getToken()) != T_EOF)
   {
      if ((tkn == T_ANY) && (_scanner->getChar() == _T('}')))
      {
         break;
      }
      if (tkn == T_CASE)
      {
         statementAllowed = true;
         if (nxt != 0)
         {
            buf.push_back(BppOpCodeByte(OP_BR,ODT_NWORD));
            nxtBlock = (BppNWord) buf.push_backNW(0);
            _fixupRefChain(context,nxt,buf.size());
         }
         else 
            nxtBlock = 0;
         buf.push_back(BppOpCodeByte(OP_POP)); // reserve space for second value to compare
         buf.push_back(BppOpCodeByte(OP_DUP)); // reserve space for second value to compare
         buf.push_back(BppOpCodeByte(OP_DUP)); // reserve space for second value to compare
         _codeExpr(context,errCnt);
         if (!_fetchRequire(T_ANY,_T(':'))) errCnt++;
         buf.push_back(BppOpCodeByte(OP_EQ));
         buf.push_back(BppOpCodeByte(OP_BRF,ODT_NWORD));
         nxt = (BppNWord) buf.push_backNW(0); 
         if (nxtBlock != 0)
         {
            _fixupRefChain(context,nxtBlock,buf.size());
            nxtBlock = 0;
         }
      }
      else if (tkn == T_DEFAULT)
      {
         if (cdefault != 0)
         {
            _error(_T("multiple 'default' branches not allowed"), &errCnt);
         }
         statementAllowed = true;
         if (!_fetchRequire(T_ANY,_T(':'))) errCnt++;

		 cdefault = (BppNWord) buf.size();
         if (nxtBlock != 0)
         {
            _fixupRefChain(context,nxtBlock,buf.size());
            nxtBlock = 0;
         }
      }
      else
      {
         if (!statementAllowed)
         {
            _error(_T("'case' or 'default' expected"), &errCnt);
         }
         _scanner->back();
         _codeStatement(context,errCnt);
      }
   }
   BppNWord cptr = (BppNWord) buf.size();
   end = context.popBreak(oldBreak,end);
   _fixupRefChain(context,end,cptr);
   if (nxt != 0)
   {
      if (cdefault != 0)
         _fixupRefChain(context,nxt,cdefault);
      else
         _fixupRefChain(context,nxt,cptr);
   }
   if (nxtBlock != 0)
      _fixupRefChain(context,nxtBlock,cptr);

   buf.push_back(BppOpCodeByte(OP_POPC));
}

// break ::= "break" ";"
void BppCompiler::_codeBreak(BppFuncContext& context, int& errCnt)
{
   BppBuffer& buf = *context.codeBuffer;

   if (context.breakStack.size() > 0) 
   {
      buf.push_back(BppOpCodeByte(OP_BR,ODT_NWORD));
      int idx = buf.push_backNW(context.popBreak(0,0));
      context.pushBreak((BppNWord) idx);
    }
    else
    {
       _error(_T("break outside of loop found"), &errCnt);
   }
   if (!_fetchRequire(T_ANY,_T(';'))) errCnt++;
}

// continue ::= "continue" ";"
void BppCompiler::_codeContinue(BppFuncContext& context, int& errCnt)
{
   BppBuffer& buf = *context.codeBuffer;

   if (context.breakStack.size() > 0) 
   {
      buf.push_back(BppOpCodeByte(OP_BR,ODT_NWORD));
      buf.push_backNW(context.peekContinue());
    }
    else
    {
       _error(_T("continue outside of loop found"), &errCnt);
   }
   if (!_fetchRequire(T_ANY,_T(';'))) errCnt++;
}

// return ::= "return" [ <expr> ] ";"
void BppCompiler::_codeReturn(BppFuncContext& context, int& errCnt)
{
   BppToken tkn = _scanner->getToken();
   if ((tkn != T_ANY) || (_scanner->getChar() != _T(';')))
   {
	   if (context.returnType == BVT_VOID)
	   {
		   _error(_T("return value not allowed"), &errCnt);
	   }
	   _scanner->back();
	   _codeExpr(context,errCnt);
	   if (!_fetchRequire(T_ANY,_T(';'))) errCnt++;
   }
   else if (context.returnType != BVT_NULL && context.returnType != BVT_VOID)
   {
       _error(_T("return value expected"), &errCnt);
   }
   if (context.returnType != BVT_NULL)
   {
	   if (context.returnType == BVT_VOID)
	   {
		  context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH,ODT_BYTE));
		  context.codeBuffer->push_back(BVT_VOID);
	   }
      context.codeBuffer->push_back(BppOpCodeByte(OP_RETURN,ODT_BYTE));
	  context.codeBuffer->push_back(context.returnType);
   }
   else
      context.codeBuffer->push_back(BppOpCodeByte(OP_RETURN));
}

void BppCompiler::_fixupRefChain(BppFuncContext& context, BppNWord startIdx, BppNWord destIdx)
{
   BppBuffer& buf = *context.codeBuffer;
   BppNWord idx = startIdx;
   BppNWord bsize = (BppNWord) buf.size();
   while ((idx != 0) && (idx < bsize))
   {
#ifdef WINCE
	   BppByte* pi =  &(buf.data()[idx]);
	   memcpy(&idx,pi,sizeof(BppNWord));
	   memcpy(pi,&destIdx,sizeof(BppNWord));
#else
  #ifdef GCC
	   BppByte* pi =  &(buf.data()[idx]);
	   memcpy(&idx,pi,sizeof(BppNWord));
	   memcpy(pi,&destIdx,sizeof(BppNWord));
  #else
      BppNWord* pi = (BppNWord*) &(buf.data()[idx]);
      idx = *pi;
      *pi = destIdx;
  #endif    
#endif
   }
}

// condition ::= "(" <expr> ")"
void BppCompiler::_codeCondition(BppFuncContext& context, int& errCnt)
{
    if (!_fetchRequire(T_ANY,_T('('))) errCnt++;
    _codeExpr(context,errCnt);
    if (!_fetchRequire(T_ANY,_T(')'))) errCnt++;
}

void BppCompiler::_codeExpr(BppFuncContext& context, int& errCnt)
{
    BppVarInfo varInfo;
    _codeExpr1_comma(context,varInfo,errCnt);
    _codeRValue(*context.codeVec,varInfo);
}

// expr ::= <expr1>
// expr1 ::= <expr2> { "," <expr> }
/**
handles operator ','
*/
void BppCompiler::_codeExpr1_comma(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
   _codeExpr2_assign(context,varInfo,errCnt);
   BppToken tkn = _scanner->getToken();
   while ((tkn == T_ANY) && (_scanner->getChar() == _T(',')))
   {
      _codeRValue(*context.codeVec,varInfo);
      _codeExpr1_comma(context,varInfo,errCnt);
      _codeRValue(*context.codeVec,varInfo);
      tkn = _scanner->getToken();
   }
   _scanner->back();
}

// expr2 ::= <expr3> <assignop> <expr>
// assignop ::= "=" | "+=" | "-=" | "*=" | "/=" | "%=" | "&=" | "|=" | "^=" | "~=" |  "<<=" | ">>=" | "=>"
/**
handles assignment operators
*/
void BppCompiler::_codeExpr2_assign(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{

   _codeExpr3_ternop(context,varInfo,errCnt); // handle the '?:' operator
   BppToken tkn;
   while ( (((tkn = _scanner->getToken()) == T_ANY) && (_scanner->getChar() == _T('='))) 
      ||     tkn == T_ADDEQ
      ||     tkn == T_SUBEQ
      ||     tkn == T_MULEQ
      ||     tkn == T_DIVEQ
      ||     tkn == T_REMEQ 
      ||     tkn == T_BANDEQ
      ||     tkn == T_BOREQ
      ||     tkn == T_XOREQ
      ||     tkn == T_BNOTEQ
      ||     tkn == T_SHLEQ
      ||     tkn == T_SHREQ
      //||     tkn == T_REFEQ
      //||     tkn == T_UNREFEQ
      )
   {
      if (!_checkLValue(varInfo)) errCnt++;
      switch (tkn) 
      {
      case T_ANY: // simple assignment
      //case T_REFEQ: // reference  assignment
      //case T_UNREFEQ: // unreference eaasignment
         {
            if (varInfo.type == VAR_INDEX || varInfo.type == VAR_PROP)
              context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH)); // push right expression result
            // evaluate right expression
            BppVarInfo rInfo;
            _codeExpr1_comma(context,rInfo,errCnt);
            _codeRValue(*context.codeVec,rInfo);
            //if (tkn == T_REFEQ)
            //   context.codeBuffer->push_back(BppOpCodeByte(OP_MKREF)); // make reference
            //if (tkn == T_UNREFEQ)
            //   context.codeBuffer->push_back(BppOpCodeByte(OP_UNREF)); // make value from reference

            _codeVarInfo(*context.codeVec,VA_STORE,varInfo); // store result
         }
         break;
      case T_ADDEQ:
         _codeAssign(context,OP_ADD,varInfo,errCnt);
         break;
      case T_SUBEQ:
         _codeAssign(context,OP_SUB,varInfo,errCnt);
         break;
      case T_MULEQ:
         _codeAssign(context,OP_MUL,varInfo,errCnt);
         break;
      case T_DIVEQ:
         _codeAssign(context,OP_DIV,varInfo,errCnt);
         break;
      case T_REMEQ:
         _codeAssign(context,OP_REM,varInfo,errCnt);
         break;
      case T_BANDEQ:
         _codeAssign(context,OP_BAND,varInfo,errCnt);
         break;
      case T_BOREQ:
         _codeAssign(context,OP_BOR,varInfo,errCnt);
         break;
      case T_XOREQ:
         _codeAssign(context,OP_XOR,varInfo,errCnt);
         break;
      case T_BNOTEQ:
         _codeAssign(context,OP_BNOT,varInfo,errCnt);
         break;
      case T_SHLEQ:
         _codeAssign(context,OP_SHL,varInfo,errCnt);
         break;
      case T_SHREQ:
         _codeAssign(context,OP_SHR,varInfo,errCnt);
         break;
      } // end switch
      varInfo.type = VAR_DONE;
   } // end while
   _scanner->back();
}

void BppCompiler::_codeAssign(BppFuncContext& context, BppOpCode op, BppVarInfo& varInfo, int& errCnt)
{
    _codeVarInfo(*context.codeVec,VA_DUP,varInfo); // duplicate index (??)
    _codeVarInfo(*context.codeVec,VA_LOAD,varInfo); // load variable

    context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH)); // push right expression result
    // evaluate right expression
    BppVarInfo rInfo;
    _codeExpr1_comma(context,rInfo,errCnt);
    _codeRValue(*context.codeVec,rInfo);

    // code operation
    context.codeBuffer->push_back(BppOpCodeByte(op));
    _codeVarInfo(*context.codeVec,VA_STORE,varInfo); // store result
}

// expr3 ::= <expr4> { "?" <expr> ":" <expr> }
/**
handle ?:-operator
*/
void BppCompiler::_codeExpr3_ternop(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
   _codeExpr4_logor(context,varInfo,errCnt);
   BppToken tkn;
   BppBuffer& buf = *context.codeBuffer;
   while (((tkn = _scanner->getToken()) == T_ANY) && (_scanner->getChar() == _T('?'))) 
   {
      _codeRValue(*context.codeVec,varInfo);
      buf.push_back(BppOpCodeByte(OP_BRF,ODT_NWORD));
      BppNWord nxt = buf.push_backNW(0);
      _codeExpr1_comma(context,varInfo,errCnt);
      _codeRValue(*context.codeVec,varInfo);

      if (!_fetchRequire(T_ANY,_T(':'))) errCnt++;

      buf.push_back(BppOpCodeByte(OP_BR,ODT_NWORD));
      BppNWord end = buf.push_backNW(0);
      _fixupRefChain(context,nxt,buf.size());
      _codeExpr1_comma(context,varInfo,errCnt);
      _codeRValue(*context.codeVec,varInfo);
      _fixupRefChain(context,end,buf.size());
   }
   _scanner->back();
}

// expr4 ::= <expr5> { "||" <expr5> }
/**
handle '||' operator
*/
void BppCompiler::_codeExpr4_logor(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
   _codeExpr5_logand(context,varInfo,errCnt);
   BppNWord end=0;
   BppToken tkn;
   while ((tkn = _scanner->getToken()) == T_OR) 
   {
      _codeRValue(*context.codeVec,varInfo);
      context.codeBuffer->push_back(BppOpCodeByte(OP_BRT,ODT_NWORD));
      end = context.codeBuffer->push_backNW(end);
      _codeExpr5_logand(context,varInfo,errCnt);
      _codeRValue(*context.codeVec,varInfo);
   }
   _fixupRefChain(context,end,context.codeBuffer->size());
   _scanner->back();
}


// expr5 ::= <expr6> { "&&" <expr6> }
/**
handle '&&' operator
*/
void BppCompiler::_codeExpr5_logand(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
   _codeExpr6_binor(context,varInfo,errCnt);
   BppNWord end=0;
   BppToken tkn;
   while ((tkn = _scanner->getToken()) == T_AND) 
   {
      _codeRValue(*context.codeVec,varInfo);
      context.codeBuffer->push_back(BppOpCodeByte(OP_BRF,ODT_NWORD));
      end = context.codeBuffer->push_backNW(end);
      _codeExpr6_binor(context,varInfo,errCnt);
      _codeRValue(*context.codeVec,varInfo);
   }
   _fixupRefChain(context,end,context.codeBuffer->size());
   _scanner->back();
}

// expr6 ::= <expr7> { "|" <expr7> }
/**
handle '|' operator
*/
void BppCompiler::_codeExpr6_binor(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
   BppToken tkn;
   _codeExpr7_xor(context,varInfo,errCnt);
   while (((tkn = _scanner->getToken()) == T_ANY) && (_scanner->getChar() == _T('|'))) 
   {
      _codeRValue(*context.codeVec,varInfo);
      context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));

      _codeExpr7_xor(context,varInfo,errCnt);
      _codeRValue(*context.codeVec,varInfo);
      context.codeBuffer->push_back(BppOpCodeByte(OP_BOR));
   }
   _scanner->back();
}


// expr7 ::= <expr8> [ "^" <expr8> }
/**
handle '^' operator
*/
void BppCompiler::_codeExpr7_xor(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
   BppToken tkn;
   _codeExpr8_binand(context,varInfo,errCnt);
   while (((tkn = _scanner->getToken()) == T_ANY) && (_scanner->getChar() == _T('^'))) 
   {
      _codeRValue(*context.codeVec,varInfo);
      context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));

      _codeExpr8_binand(context,varInfo,errCnt);
      _codeRValue(*context.codeVec,varInfo);
      context.codeBuffer->push_back(BppOpCodeByte(OP_XOR));
   }
   _scanner->back();
}

// expr8 ::= <expr9> { "&" <expr9> }
/**
handle '&' operator
*/
void BppCompiler::_codeExpr8_binand(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
   BppToken tkn;
   _codeExpr9_eqops(context,varInfo,errCnt);
   while (((tkn = _scanner->getToken()) == T_ANY) && (_scanner->getChar() == _T('&'))) 
   {
      _codeRValue(*context.codeVec,varInfo);
      context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));

      _codeExpr9_eqops(context,varInfo,errCnt);
      _codeRValue(*context.codeVec,varInfo);
      context.codeBuffer->push_back(BppOpCodeByte(OP_BAND));
   }
   _scanner->back();
}

// expr9 ::= <expr10> { ("==" | "!=") <expr10> }
/**
handle the '==' and '!=' operators
*/
void BppCompiler::_codeExpr9_eqops(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
   _codeExpr10_cmpops(context,varInfo,errCnt);
   BppToken tkn;
   BppOpCode op = OP_NIL;
   while ((tkn = _scanner->getToken()) == T_EQ || tkn == T_NE) 
   {
      switch (tkn) 
      {
      case T_EQ: op = OP_EQ; break;
      case T_NE: op = OP_NE; break;
      }
      _codeRValue(*context.codeVec,varInfo);
      context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));

      _codeExpr10_cmpops(context,varInfo,errCnt);
      _codeRValue(*context.codeVec,varInfo);
      context.codeBuffer->push_back(BppOpCodeByte(op));
   }
   _scanner->back();
}

// expr10 ::= <expr11> { ("<" | ">" | "<=" | ">="} <expr11>
/**
handle the '<', '<=', '>=' and '>' operators
*/
void BppCompiler::_codeExpr10_cmpops(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
   _codeExpr11_shift(context,varInfo,errCnt);
   BppToken tkn;
   BppOpCode op = OP_NIL;
   BppChar c = 0;;
   while ((tkn = _scanner->getToken()) == T_LE 
      || tkn == T_GE 
      || ((tkn == T_ANY) && (((c = _scanner->getChar()) == _T('>')) || (c == _T('<'))))
      ) 
   {
      switch (tkn) 
      {
      case T_LE: op = OP_LE; break;
      case T_GE: op = OP_GE; break;
      case T_ANY: op = (c == _T('<')) ? OP_LT : OP_GT; break;
      }
      _codeRValue(*context.codeVec,varInfo);
      context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));

      _codeExpr11_shift(context,varInfo,errCnt);
      _codeRValue(*context.codeVec,varInfo);
      context.codeBuffer->push_back(BppOpCodeByte(op));
   }
   _scanner->back();
}

// expr11 ::= <expr12> { ("<<" | ">>") <expr12> }
/**
handle the '<<' and '>>' operators
*/
void BppCompiler::_codeExpr11_shift(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
   _codeExpr12_addops(context,varInfo,errCnt);
   BppToken tkn;
   BppOpCode op = OP_NIL;
   while ((tkn = _scanner->getToken()) == T_SHL || tkn == T_SHR) 
   {
      switch (tkn) 
      {
      case T_SHL: op = OP_SHL; break;
      case T_SHR: op = OP_SHR; break;
      }
      _codeRValue(*context.codeVec,varInfo);
      context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));

      _codeExpr12_addops(context,varInfo,errCnt);
      _codeRValue(*context.codeVec,varInfo);
      context.codeBuffer->push_back(BppOpCodeByte(op));
   }
   _scanner->back();

}

// expr12 ::= <expr13> { ("+" | "-") <expr13>}
/**
handle the '+' and '-' operators
*/
void BppCompiler::_codeExpr12_addops(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
   _codeExpr13_mulops(context,varInfo,errCnt);
   BppToken tkn;
   BppOpCode op = OP_NIL;
   BppChar c;
   while ( ((tkn = _scanner->getToken()) == T_ANY) 
      && (((c = _scanner->getChar()) == _T('+')) || (c == _T('-'))) ) 
   {
      switch (c) 
      {
      case _T('+'): op = OP_ADD; break;
      case _T('-'): op = OP_SUB; break;
      }
      _codeRValue(*context.codeVec,varInfo);
      context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));

      _codeExpr13_mulops(context,varInfo,errCnt);
      _codeRValue(*context.codeVec,varInfo);
      context.codeBuffer->push_back(BppOpCodeByte(op));
   }
   _scanner->back();
}

// expr13 ::= <expr14> { ("*" | "/" | "%") <expr14> }
/**
handle the '*','/' and '%' operators
*/
void BppCompiler::_codeExpr13_mulops(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
   _codeExpr14_preops(context,varInfo,errCnt);
   BppToken tkn;
   BppOpCode op = OP_NIL;
   BppChar c;
   while ( ((tkn = _scanner->getToken()) == T_ANY) 
      && (((c = _scanner->getChar()) == _T('*')) || (c == _T('/')) || (c == _T('%'))) ) 
   {
      switch (c) 
      {
      case _T('*'): op = OP_MUL; break;
      case _T('/'): op = OP_DIV; break;
      case _T('%'): op = OP_REM; break;
      }
      _codeRValue(*context.codeVec,varInfo);
      context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));

      _codeExpr14_preops(context,varInfo,errCnt);
      _codeRValue(*context.codeVec,varInfo);
      context.codeBuffer->push_back(BppOpCodeByte(op));
   }
   _scanner->back();
}


 
// expr14 ::=  (("+" | "-" | "!" | "~") <expr14>) | (["++" | "--" | "new" | "delete" | "-" | "!" | "~"] <expr15>)
/**
do_expr14 - handle unary operators
*/
void BppCompiler::_codeExpr14_preops(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
   BppBuffer& buf = *context.codeBuffer;
   BppVector& codeVec = *context.codeVec;
   BppToken tkn =_scanner->getToken();
   switch (tkn) 
   {
   case T_INC:
      _codePreIncDec(context,OP_INC,varInfo,errCnt);
      break;
   case T_DEC:
      _codePreIncDec(context,OP_DEC,varInfo,errCnt);
      break;
   case T_NEW:
      _codeNew(context,varInfo,errCnt);
      break;
   case T_DELETE:
      _codeDelete(context,varInfo,errCnt);
      break;
   case T_ANY:
      {
         BppChar c = _scanner->getChar();
         switch (c)
         {
         case _T('-'):
            _codeExpr14_preops(context,varInfo,errCnt);
            _codeRValue(codeVec,varInfo);
            buf.push_back(BppOpCodeByte(OP_NEG));
            break;
         case _T('!'):
            _codeExpr14_preops(context,varInfo,errCnt);
            _codeRValue(codeVec,varInfo);
            buf.push_back(BppOpCodeByte(OP_NOT));
            break;
         case _T('~'):
            _codeExpr14_preops(context,varInfo,errCnt);
            _codeRValue(codeVec,varInfo);
            buf.push_back(BppOpCodeByte(OP_BNOT));
            break;
         case _T('+'):
            _codeExpr14_preops(context,varInfo,errCnt);
            _codeRValue(codeVec,varInfo);
            buf.push_back(BppOpCodeByte(OP_PLUS));
            //buf.push_back(BppOpCodeByte(OP_NEG));
            break;
         case _T('&'):
            _codeExpr14_preops(context, varInfo, errCnt);
            _codeRValue(codeVec, varInfo);
            buf.push_back(BppOpCodeByte(OP_MKREF)); // make reference
            break;
         case _T('*'):
            _codeExpr14_preops(context, varInfo, errCnt);
            _codeRValue(codeVec, varInfo);
            buf.push_back(BppOpCodeByte(OP_UNREF)); // make value from reference
             break;
         default:
            _scanner->back();
            _codeExpr15_post(context,varInfo,errCnt);
            return;
         }
      }
      break; 
   default:
      _scanner->back();
      _codeExpr15_post(context,varInfo,errCnt);
      return;
   }
}

// expr15 ::= <primary>  <primarysuffix> { <primarysuffix> } .
// primarysuffix ::= <funcop> | <indexop> | <methodcall>  | "--" | "++" .
// methodcall ::= "->" <funcop> .
/**
handle function calls, index operations, method calls, postfix inc/dec
*/
void BppCompiler::_codeExpr15_post(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
   BppToken tkn;
   BppChar c = 0;
   _codePrimary(context,varInfo,errCnt);
   while (
      (((tkn = _scanner->getToken()) == T_ANY) && 
	     (( (c=_scanner->getChar()) == _T('(')) || 
		    (c == _T('[')) || 
			(c == _T('.'))
	  ))
      ||     tkn == T_MEMREF
      ||     tkn == T_INC
      ||     tkn == T_DEC
    )
   {
      switch (tkn) 
      {
      case T_ANY:
         if (c == _T('('))
            _codeFunctionCall(context,varInfo,errCnt);
         else if (c == _T('['))
            _codeIndexedOp(context,varInfo,errCnt);
         else if (c == _T('.'))
            _codePropertyOp(context,varInfo,errCnt);
         break;
      case T_MEMREF:
         if (((tkn = _scanner->getToken()) == T_ANY) && (_scanner->getChar() == _T('('))) 
			 _codeMethodCall(0,varInfo,context,errCnt);
         else
		 {
			 if (!_require(tkn, T_IDENTIFIER)) errCnt++;
            _codeMethodCall(_scanner->getTokenString(),varInfo,context,errCnt);
		 }
         break;
      case T_INC:
         _codePostIncDec(context,OP_PINC,varInfo,errCnt);
         break;
      case T_DEC:
         _codePostIncDec(context,OP_PDEC,varInfo,errCnt);
         break;
      }
   } // end while
   _scanner->back();
}

  
void BppCompiler::_codePreIncDec(BppFuncContext& context, BppOpCode op, BppVarInfo& varInfo, int& errCnt)
{
   _codeExpr15_post(context,varInfo,errCnt);
   if (!_checkLValue(varInfo)) errCnt++;
   _codeVarInfo(*context.codeVec,VA_DUP,varInfo);
   _codeVarInfo(*context.codeVec,VA_LOAD,varInfo);
   context.codeBuffer->push_back(BppOpCodeByte(op));
   _codeVarInfo(*context.codeVec,VA_STORE,varInfo);
   varInfo.type = VAR_DONE;
}

void BppCompiler::_codePostIncDec(BppFuncContext& context, BppOpCode op, BppVarInfo& varInfo, int& errCnt)
{
   if (!_checkLValue(varInfo)) ++errCnt;
   _codeVarInfo(*context.codeVec,VA_DUP,varInfo);
   _codeVarInfo(*context.codeVec,VA_LOAD,varInfo);
  context.codeBuffer->push_back(BppOpCodeByte(OP_DUP));
   context.codeBuffer->push_back(BppOpCodeByte(op));
   _codeVarInfo(*context.codeVec,VA_STORE,varInfo);
  context.codeBuffer->push_back(BppOpCodeByte(OP_POP));
   varInfo.type = VAR_DONE;
}


void BppCompiler::_codeNew(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{

    if (!_fetchRequire(T_IDENTIFIER)) 
    {
       errCnt++;
       return;
    }

    BppString className = _scanner->getTokenString();
    BppClass* pclass = _findClass(className,errCnt);
    BppVector& codeVec = *context.codeVec;
    codeVec.push_back(pclass);
    _codeLiteral(codeVec,codeVec.size() - 1);

    context.codeBuffer->push_back(BppOpCodeByte(OP_NEW));
    varInfo.type = VAR_DONE;
    _codeMethodCall(className.c_str(),varInfo,context,errCnt); 
}

void BppCompiler::_codeDelete(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
   BppBuffer& buf = *context.codeBuffer;
   BppVector& codeVec = *context.codeVec;
   _codePrimary(context,varInfo,errCnt); // parse expr behind delete
   _codeRValue(codeVec,varInfo); 
   buf.push_back(BppOpCodeByte(OP_DELETE));
   varInfo.type = VAR_DONE;
}


static bool containsIdx(const BppDictionary& dict, int idx)
{
	for (BppDictionaryEntry* p = dict.getFirst();p;p=p->getNext())
	{
		if (((BppInt)idx) == (BppInt) p->getValue()) return true;
	}
	return false;
}
// primary ::= <int> | <float> | <string> | "null" | "nil" | <variable> | <static_member> | <globalref> 
// variable ::= <identifier>
// static_member ::= <identifier> :: <identifier>
// globalref ::= "::" <identifier>
void BppCompiler::_codePrimary(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
    BppVector& codeVec = *context.codeVec;
    BppToken tkn = _scanner->getToken();
    switch (tkn) 
    {
    case T_NUMBER:
	case T_FNUMBER:
		{
			BppValue v = _scanner->getNumberValue();
         varInfo.varvtype = v.getType();
			  int idx = codeVec.rfind(v);
			  
			  if (idx < 0 || (containsIdx(context.statics,idx)))
			  {
				 idx = codeVec.size();
				 codeVec.push_back(v);
			  }
		   _codeLiteral(codeVec,idx);
		   varInfo.type = VAR_DONE;
		}
       break;
    case T_STRING:
       {
          // reuse string literal if it already exists
          const BppChar* str = _scanner->getTokenString();
		  BppValue v(new BppString(str));
        varInfo.varvtype = v.getType();
        int idx = codeVec.rfind(v); //.getStringIndex(str);
          if (idx < 0 || (containsIdx(context.statics,idx)))
          {
             idx = codeVec.size();
             codeVec.push_back(v);
          }
          _codeLiteral(codeVec,(BppNWord) idx);
       }
       varInfo.type = VAR_DONE;
       break;
    case T_NIL:
		{
			  BppValue v;
           varInfo.varvtype = v.getType();
           int idx = codeVec.rfind(v);
			  if (idx < 0 || (containsIdx(context.statics,idx)))
			  {
				 idx = codeVec.size();
				 codeVec.push_back(v);
			  }
		   _codeLiteral(codeVec,idx);
		   varInfo.type = VAR_DONE;
		}
       break;
       // putcbyte(OP_NIL);
//       break;
    case T_FUNCTION:
       {
          BppDictionaryEntry dummy(0,BppValue());
          _doLiteralValue(tkn,dummy,errCnt);
		  const BppValue& v = dummy.getValue();
        varInfo.varvtype = v.getType();
        int idx = codeVec.rfind(v);
         if (idx < 0 || (containsIdx(context.statics,idx)))		  
		  {
			 idx = codeVec.size();
			 codeVec.push_back(v);
		  }
		   _codeLiteral(codeVec,idx);
		   varInfo.type = VAR_DONE;
          varInfo.type = VAR_DONE;
          break;
       }
    case T_IDENTIFIER:
       {
          BppString id = _scanner->getTokenString();
          BppDictionaryEntry* entry = _namedLiterals.find(id);
          if (entry != 0)
          {
                codeVec.push_back(entry->getValue());
                varInfo.varvtype = entry->getValue().getType();
                _codeLiteral(codeVec,codeVec.size()-1);
                varInfo.type = VAR_DONE;
                break;
          }
          tkn = _scanner->getToken();
          if (tkn == T_CC) 
          {
             BppValue ownClass = codeVec[IDX_CLASS]; // save own class reference
             BppClass* pown = ownClass;
             BppClass* pother = _findClass(id,errCnt); // temporary set other class
             if (pother == 0) return;
             codeVec[IDX_CLASS] = pother;
             if (_fetchRequire(T_IDENTIFIER))
             {
                BppDictionaryEntry* de = pother->findMember(pown,_scanner->getTokenString());
                if (de != 0)
                {
                   BppSymbolType st = de->getSymbolType();
                   if (st == ST_DATA)
                   {
                      if (!(pown && pown->isSubtypeOf(pother)))
                      {
                         _error(_T("not a static class member"), &errCnt);
                      }
                   }
				      if (st == ST_FUNCTION)
				      {
                         //if (pown && !pown->isSubtypeOf(pother))
                         //{
                         //   _error(_T("not a static class member"), &errCnt);
                         //}
					      if (_scanner->getToken() == T_ANY && _scanner->getChar() == _T('('))
					      {
						      _error(_T("invalid non-virtual member call, use obj->(class::member)() notation"), &errCnt);
					      }
					      _scanner->back();
				      }
                }
                else
                {
                   _error(_T("not a class member"), &errCnt);
                }
                varInfo = _findClassVariable(_scanner->getTokenString(),context,errCnt,pown);
             }
             else
             {
                errCnt++;
                varInfo = BppVarInfo();
             }
             codeVec[IDX_CLASS] = ownClass; // reset class reference
          }
          else 
          {
             _scanner->back();
             varInfo = _findVariable(id.c_str(),context,false,errCnt);
          }
       }
       break;
    case T_CC:                   // erzwungener Globalverweis
       if (_fetchRequire(T_IDENTIFIER))
          varInfo = _findVariable(_scanner->getTokenString(),context,true,errCnt);
       else
       {
          errCnt++;
          varInfo = BppVarInfo();
       }
       break;
    case T_ANY:
       {
          switch (_scanner->getChar())
          {
          case _T('('):
             {
                _codeExpr1_comma(context,varInfo,errCnt);
                if (!_fetchRequire(T_ANY,_T(')'))) errCnt++;
                return;
             }
          case _T('['):
          case _T('{'):
          case _T('#'):
             {
                BppDictionaryEntry dummy(0,BppValue());
                _doLiteralValue(tkn,dummy,errCnt);
                codeVec.push_back(dummy.getValue());
                varInfo.varvtype = dummy.getValue().getType();
                _codeLiteral(codeVec,codeVec.size()-1);
                varInfo.type = VAR_DONE;
                return;
             }
          }
       }
	case T_ERROR:
       _print(_scanner->getTokenString());
       errCnt++;
       break;
    default:
       _error(_T("primary expression expected"), &errCnt);
       break;
    }
}

BppClass* BppCompiler::_findClass(const BppString& name, int& errCnt)
{
   BppDictionaryEntry* classentry = _classes.find(name);
   if (classentry == 0)
   {
      _error(_T("undefined class"), &errCnt);
      return 0;
   }
   return classentry->getValue();
}

BppVarInfo BppCompiler::_findVariable(const BppChar* name, BppFuncContext& context, bool global, int& errCnt)
{
   BppVarInfo result;

   int n = context.temporaries.getStringIndex(name);
   if ((!global) && (n >= 0))
   {
      result.type = VAR_LOCAL;
      result.index = n;
   }
   else if ((!global) &&((n = context.arguments.getStringIndex(name)) >= 0) )
   {
      result.type = VAR_ARG;
      result.index = n;
   }
   else
   {
      BppDictionaryEntry* pstatic = context.statics.find(name);
      if ((!global) && (pstatic != 0))
      {
         result.type = VAR_SLOCAL;
         result.index = (int) (BppInt) pstatic->getValue();
      }
      else 
      {
         if (!global)
         {
            result = _findClassVariable(name,context, errCnt,context.contextClass);
            if ((result.index >= 0) || (result.type == VAR_DONE))
               return result;
         }
         if (_tcscmp(name,_T("self")) == 0)
         {
            _codeLiteral(*(context.codeVec),IDX_SELFREFERENCE);
            result.index = IDX_SELFREFERENCE;
            result.type = VAR_DONE;
            return result;
         }
		 //if (global)
		 //    sym = &_symbols[name] ;
		 //else
         BppDictionaryEntry* sym = _symbols.find(name);
          if (sym)
          {
             result.type = VAR_VAR;

             // search for previous reference to sym first
             result.index = context.codeVec->getVarIndex(sym);
             if (result.index < 0)
             {
                // add reference if not found
                context.codeVec->push_back(sym);
                result.index = context.codeVec->size() - 1;
             }
             return result;
          }
        sym = _classes.find(name);
         if (!sym)
         {
            _error(BppString::format(_T("variable %s not defined"), name).c_str(), &errCnt);
            return result;
         }
         const BppValue& classVal = sym->getValue();
         BppClass* pclass = classVal;
         result.index = -1;
         // unsigned vn = context.codeVec->size();
         int i;
         for (i = 0; i<n; ++i)
         {
            if (((BppClass*)((*context.codeVec)[i])) == pclass)
            {
               result.index = i;
               break;
            }
         }
         if (result.index < 0)
         {
            context.codeVec->push_back(sym->getValue());
            result.index = context.codeVec->size() - 1;
         }
         _codeLiteral(*context.codeVec, result.index);
         result.type = VAR_DONE;

      }
   }
   return result;
}


BppVarInfo BppCompiler::_findClassVariable(const BppChar* name, BppFuncContext& context, int& errCnt, BppClass* refClass)
{
   BppVarInfo result;
   BppClass* ownerClass = (*context.codeVec)[IDX_CLASS];
   if (ownerClass == 0) return result;

   if (_tcscmp(name,_T("self")) == 0)
   {  
      BppVarInfo selfresult;
      selfresult = _findVariable(_T("this"),context,false, errCnt);

      if (selfresult.type == VAR_ARG)
      {
         _codeMethodCall(name,selfresult,context,errCnt);
         return selfresult;
      }
   }

   BppDictionaryEntry* entry = ownerClass->findMember(refClass,name);
   if (entry == 0)
      return result;
   switch (entry->getSymbolType()) 
   {
   case ST_DATA:
	   {
		  result.type = VAR_MEMBER;
		  // index of member variable is stored in value of dictionary entry
		  BppVector* v = entry->getValue();
		  if (!v) 
		  {
			  _error(_T("invalid data member"), &errCnt);
			  break;
		  }
		  result.index = (int) (BppInt) (*v)[0];
	   }
      break;
   case ST_FUNCTION:
	   if (_scanner->getToken() == T_ANY && _scanner->getChar() == _T('('))
	   {
		   _scanner->back();
		  result = _findVariable(_T("this"),context,false, errCnt);
		  _codeMethodCall(name,result,context,errCnt);
		  break;
	   }
	   else
		   _scanner->back();
   case ST_SDATA:
      result.type = VAR_VAR;
      // search for previous reference to entry first
      result.index = context.codeVec->getVarIndex(entry);
      if (result.index < 0)
      {
         // add reference if not found
         context.codeVec->push_back(entry);
         result.index = context.codeVec->size() - 1;
      }
      break;
   case ST_SFUNCTION:
      // search for previous reference to entry first
      result.index = context.codeVec->getVarIndex(entry);
      if (result.index < 0)
      {
         // add reference if not found
         context.codeVec->push_back(entry);
         result.index = context.codeVec->size() - 1;
      }
      _codeVariable(*context.codeVec,VA_LOAD,result.index);
      result.type = VAR_DONE;
      break;
   }
    return result;
}

// code receiver value
void BppCompiler::_codeRValue(BppVector& codeVec, BppVarInfo& varInfo)
{
   switch (varInfo.type)
   {
   case VAR_ARG:
      _codeArgument(codeVec,VA_LOAD,varInfo.index);
      break;
   case VAR_LOCAL:
      _codeTemporary(codeVec,VA_LOAD,varInfo.index);
      break;
   case VAR_SLOCAL:
      _codeStaticLocal(codeVec,VA_LOAD,varInfo.index);
      break;
   case VAR_VAR:
      _codeVariable(codeVec,VA_LOAD,varInfo.index);
      break;
   case VAR_MEMBER:
      _codeMember(codeVec,VA_LOAD,varInfo.index);
      break;
   case VAR_INDEX:
      _codeIndexRef(codeVec,VA_LOAD);
      break;
   case VAR_PROP:
      _codePropertyRef(codeVec,VA_LOAD);
      break;
   default:
      break;
   }
   varInfo.type = VAR_DONE;
}

void BppCompiler::_codeVarInfo(BppVector& codeVec, BppVarAccessType act, BppVarInfo& varInfo)
{
	BppBuffer* cbuf = codeVec[IDX_CODE];
	///// nothing, variable doesn't exist
	//VAR_NONE = 0,
	//	/// argument
	//	VAR_ARG,
	//	/// temporary
	//	VAR_LOCAL,
	//	/// static local variable in function
	//	VAR_SLOCAL,
	//	/// global variable or static member variable
	//	VAR_VAR,
	//	/// member variable
	//	VAR_MEMBER,
	//	/// index variable
	//	VAR_INDEX,
	//	/// property member
	//	VAR_PROP,
	//	/// ignore - variable already processed
	//	VAR_DONE
		
		switch (varInfo.type)
   {
   case VAR_ARG:
	   if (/*varInfo.type*/act == VA_STORE && varInfo.varvtype != BVT_NULL)
	   {
		   _codeIndexedOpcode(codeVec, OP_PUSH, varInfo.varvtype);
		   _codeArgument(codeVec, act, varInfo.index);
		   cbuf->push_back(BppOpCodeByte(OP_POP));
	   }
      _codeArgument(codeVec,act,varInfo.index);
      break;
   case VAR_LOCAL:
	   if (act == VA_STORE && varInfo.varvtype != BVT_NULL)
	   {
		   _codeIndexedOpcode(codeVec, OP_PUSH, varInfo.varvtype);
		   _codeTemporary(codeVec, act, varInfo.index);
		   cbuf->push_back(BppOpCodeByte(OP_POP));
	   }
	   _codeTemporary(codeVec,act,varInfo.index);
      break;
   case VAR_SLOCAL:
	   if (act == VA_STORE && varInfo.varvtype != BVT_NULL)
	   {
		   _codeIndexedOpcode(codeVec, OP_PUSH, varInfo.varvtype);
		   _codeStaticLocal(codeVec, act, varInfo.index);
		   cbuf->push_back(BppOpCodeByte(OP_POP));
	   }
	   _codeStaticLocal(codeVec,act,varInfo.index);
      break;
   case VAR_VAR:
	   if (act == VA_STORE && varInfo.varvtype != BVT_NULL)
	   {
		   _codeIndexedOpcode(codeVec, OP_PUSH, varInfo.varvtype);
		   _codeVariable(codeVec, act, varInfo.index);
		   cbuf->push_back(BppOpCodeByte(OP_POP));
	   }
	   _codeVariable(codeVec,act,varInfo.index);
      break;
   case VAR_MEMBER:
	   if (act == VA_STORE && varInfo.varvtype != BVT_NULL)
	   {
		   _codeIndexedOpcode(codeVec, OP_PUSH, varInfo.varvtype);
		   _codeVariable(codeVec, act, varInfo.index);
		   cbuf->push_back(BppOpCodeByte(OP_POP));
	   }
	   _codeMember(codeVec,act,varInfo.index);
      break;
   case VAR_INDEX:
      _codeIndexRef(codeVec,act);
      break;
   case VAR_PROP:
      _codePropertyRef(codeVec,act);
      break;
   default:
      break;
   }
}


   /// compile an argument
void BppCompiler::_codeArgument(BppVector& codeVec, BppVarAccessType act, BppNWord index)
{
   switch (act) 
   {
   case VA_LOAD:        
      _codeIndexedOpcode(codeVec,OP_AREF,index);
      break;
   case VA_STORE:
      _codeIndexedOpcode(codeVec,OP_ASET,index);
      break;
   }
}
   /// compile a temporary (local variable)
void BppCompiler::_codeTemporary(BppVector& codeVec, BppVarAccessType act, BppNWord index)
{
   switch (act) 
   {
   case VA_LOAD:        
      _codeIndexedOpcode(codeVec,OP_TREF,index);
      break;
   case VA_STORE:
      _codeIndexedOpcode(codeVec,OP_TSET,index);
      break;
   }
}

void BppCompiler::_codeStaticLocal(BppVector& codeVec, BppVarAccessType act, BppNWord index)
{
   switch (act) 
   {
   case VA_LOAD:        
      _codeIndexedOpcode(codeVec,OP_SREF,index);
      break;
   case VA_STORE:
      _codeIndexedOpcode(codeVec,OP_SSET,index);
      break;
   }
}

   /// compile a variable reference
void BppCompiler::_codeVariable(BppVector& codeVec, BppVarAccessType act, BppNWord index)
{
   switch (act) 
   {
   case VA_LOAD:        
      _codeIndexedOpcode(codeVec,OP_REF,index);
      break;
   case VA_STORE:
      _codeIndexedOpcode(codeVec,OP_SET,index);
      break;
   }
}
void BppCompiler::_codeMember(BppVector& codeVec, BppVarAccessType act, BppNWord index)
{
   switch (act) 
   {
   case VA_LOAD:        
      _codeIndexedOpcode(codeVec,OP_MREF,index);
      break;
   case VA_STORE:
      _codeIndexedOpcode(codeVec,OP_MSET,index);
      break;
   }
}


/// compile a literal reference
void BppCompiler::_codeLiteral(BppVector& codeVec, BppNWord index)
{
   _codeIndexedOpcode(codeVec,OP_LIT,index);
}

   /// compile an indexed reference
/**
Pushes opcode depending on <i>act</i> to end of code buffer (<i>codeVec</i>[IDX_CODE]).
@see _codeRValue, _codeVarInfo, BppVarAccessType
*/
void BppCompiler::_codeIndexRef(BppVector& codeVec, BppVarAccessType act)
{
   BppOpCode oc = OP_NOP;
    switch (act) 
    {
    case VA_LOAD:       
       oc = OP_VREF; 
       break;
    case VA_STORE:      
       oc = OP_VSET; 
       break;
    case VA_PUSH:  
       oc = OP_PUSH; 
       break;
    case VA_DUP:        
       oc = OP_DUP2; 
       break;
	default:
		return;
    }
    BppBuffer* buf = codeVec[IDX_CODE];
    buf->push_back(BppOpCodeByte(oc));
}

  /// compile an indexed reference
/**
Pushes opcode depending on <i>act</i> to end of code buffer (<i>codeVec</i>[IDX_CODE]).
@see _codeRValue, _codeVarInfo, BppVarAccessType
*/
void BppCompiler::_codePropertyRef(BppVector& codeVec, BppVarAccessType act)
{
   BppOpCode oc = OP_NOP;
    switch (act) 
    {
    case VA_LOAD:       
       oc = OP_PREF; 
       break;
    case VA_STORE:      
       oc = OP_PSET; 
       break;
    case VA_PUSH:  
       oc = OP_PUSH; 
       break;
    case VA_DUP:        
       oc = OP_DUP2; 
       break;
	default:
		return;
    }
    BppBuffer* buf = codeVec[IDX_CODE];
    buf->push_back(BppOpCodeByte(oc));
}

void BppCompiler::_codeMethodCall(const BppChar* name, BppVarInfo& thisInfo,  BppFuncContext& context, int& errCnt)
{
   // get the receiver value (this)
   _codeRValue(*context.codeVec,thisInfo);
   if (thisInfo.type != VAR_DONE)
   {
      _error(_T("BppCompiler::_codeMethodCall : no 'this' reference found"), &errCnt);
   }
   BppBuffer* buf = context.codeBuffer;
   // generate code to push the selector (member function name)
   buf->push_back(BppOpCodeByte(OP_PUSH));
   if (name) 
   {
	   int index = (_tcscmp(name,_T("self")) == 0) ? IDX_SELFREFERENCE : context.codeVec->getStringIndex(name);
	   if (index < 0)
	   {
		  context.codeVec->push_back(new BppString(name));
		  index = context.codeVec->size() - 1;
	   }
	   _codeLiteral(*context.codeVec,index);
   }
   else // evaluate expression to get method or method name 
   {
        _codeExpr(context, errCnt);
        if (!_fetchRequire(T_ANY,_T(')'))) errCnt++;   
   }
   int argc = 1; // first argument is 'this'
   // compile the argument list 
   if (!_fetchRequire(T_ANY,_T('('))) errCnt++;
   BppToken tkn = _scanner->getToken();
   if (!((tkn == T_ANY) && (_scanner->getChar() == _T(')')))) 
   {
      _scanner->back();
      do 
      {
         buf->push_back(BppOpCodeByte(OP_PUSH));
		 //size_t bsize = buf->size();
         _codeExpr2_assign(context,thisInfo,errCnt);
         _codeRValue(*context.codeVec,thisInfo);
         ++argc;
		 //_adjustAssignment(*buf,bsize);
      } while (((tkn = _scanner->getToken()) == T_ANY) && (_scanner->getChar() == _T(',')));
   }
   if (!_require(T_ANY,tkn,_T(')'))) errCnt++;
   
   // do the call
   _codeIndexedOpcode(*context.codeVec,OP_SEND,(BppNWord)argc);
   thisInfo.type = VAR_DONE;
}

// funcop ::= "(" [ <funcargs> ] ")" .
// funcargs ::= <expr2> { "," <expr2> } .
void BppCompiler::_codeFunctionCall(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
   /* get the value of the function */
   _codeRValue(*context.codeVec,varInfo);

   BppToken tkn = _scanner->getToken();
   BppChar c = (tkn == T_ANY) ? _scanner->getChar() : 0;
   BppNWord n = 0;
   BppBuffer& buf = *context.codeBuffer;
   /* compile each argument expression */
   if (c != _T(')')) 
   {
      _scanner->back();
      do 
      {
         buf.push_back(BppOpCodeByte(OP_PUSH));
			//size_t bsize = buf.size();
         _codeExpr2_assign(context,varInfo,errCnt);
         _codeRValue(*context.codeVec,varInfo);
         ++n;
		 //_adjustAssignment(buf,bsize);
        tkn = _scanner->getToken();
         c = (tkn == T_ANY) ? _scanner->getChar() : 0;
      } while (c == _T(','));
   }
   if (!_require(T_ANY,tkn,_T(')'))) errCnt++;
   _codeIndexedOpcode(*context.codeVec,OP_CALL,n);

   /* we've got an rvalue now */
   varInfo.type = VAR_DONE;
}

// indexop ::= "[" <expr> "]" .
/**
parses indexed operation
*/
void BppCompiler::_codeIndexedOp(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
//    int code_index();
    _codeRValue(*context.codeVec,varInfo);
	BppBuffer* buf = context.codeBuffer;
	size_t bsize = buf->size();
	if ((bsize > 1) && (BppOpCodeByte((*buf)[bsize-2]).opCode() == OP_LIT))
	{
		_error(_T("Indexed access to literals not allowed"), &errCnt);
	}
    context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));
    _codeExpr(context,errCnt);
    if (!_fetchRequire(T_ANY,_T(']'))) errCnt++;
    varInfo.type = VAR_INDEX;
}

// indexop ::= "." <ident>  .
/**
parses indexed operation
*/
void BppCompiler::_codePropertyOp(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
//    int code_index();
    _codeRValue(*context.codeVec,varInfo);
    context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));
    if (!_fetchRequire(T_IDENTIFIER)) errCnt++;

      // reuse string literal if it already exists
      const BppChar* str = _scanner->getTokenString();
      int idx = context.codeVec->getStringIndex(str);
      if (idx < 0)
      {
         idx = context.codeVec->size();
         context.codeVec->push_back(new BppString(str));
      }
      _codeLiteral(*context.codeVec,(BppNWord) idx);
      varInfo.type = VAR_PROP;
}

// idlist ::= <identifier> { "," <identifier> } .
int BppCompiler::_getIdList(BppVector& dest, const BppChar* termStr, int& errCnt,BppVector* defaults)
{
   int cnt = 0;
    BppToken tkn = _scanner->getToken();
    if ((tkn != T_ANY) || (!_tcschr(termStr,_scanner->getChar())))
    {
       _scanner->back();
       do 
       {
          BppToken t = _scanner->getToken();
		  BppValueType vtype = BVT_NULL;
		  BppClass* pvClass = 0;
          if (t == T_VAR)
             t = _scanner->getToken();
		  else if (t == T_IDENTIFIER || t == T_FUNCTION)
		  {
			  vtype = _vm.findVTypeFromName(_scanner->getTokenString(),true,&pvClass);
			  if (vtype != BVT_NULL)
				  t = _scanner->getToken();
		  }
          if (!_require(T_IDENTIFIER,t))
          {
             errCnt++;
             continue;
          }
          const BppChar* p = _scanner->getTokenString();
          if (dest.getStringIndex(p) >= 0)
          {
             _error(BppString::format(_T("Duplicated identifier '%s' found."),p).c_str(), &errCnt);
          }
          else
          {
             dest.push_back(new BppString(p));
             ++cnt;
             if (defaults != 0)
             {
                if ((_scanner->getToken() != T_ANY) || (_scanner->getChar() != _T('=')))
                {
	               _scanner->back();
                   defaults->push_back(BppValue(vtype,(BppValueFlags)(vtype == BVT_NULL ? BVF_NONE : BVF_STRONGTYPED),false));
				   BppValue& vr = (*defaults)[defaults->size()-1];
				   vr.setFlags(vr.getFlags() | BVF_INVALID);
                   continue;
                }
                BppDictionaryEntry tmp(0,BppValue());
                _doLitExpr(_scanner->getToken(),tmp,errCnt);
				if (vtype != BVT_NULL)
				{
					BppRefObject* pro = 0;
					if (vtype == BVT_STRONGOBJECT && pvClass)
					{
						BppVector* pdesc = new BppVector(2,true);
						(*pdesc)[IDX_CLASS] = pvClass;
						pro = pdesc;
					}
				   BppValue tmpv(pro,vtype);
				   tmpv.setFlags(BVF_STRONGTYPED);
				   if (!tmpv.assign(tmp.getValue()))
				   {
					   BppString* pname = dest[dest.size()-1];
					   _error(BppString::format(_T("Type mismatch: %s(%s) = (%s)."),
						 pname->c_str(),
						 getValueTypeName(tmpv.getType()),
						 getValueTypeName(tmp.getValue().getType())).c_str(), &errCnt);
				   }
                   defaults->push_back(tmpv);
				}
                else defaults->push_back(tmp.getValue());
             }
          }
       } while (((tkn = _scanner->getToken()) == T_ANY) && (_scanner->getChar() == _T(',')));
    }
    _scanner->back();
    return cnt;
}

/**
Pushes opcode with index argument, if index is greater than byte opcode (or forceNativeSize set) it is prefixed by
argument width code.
*/
void BppCompiler::_codeIndexedOpcode(BppVector& codeVec, BppOpCode opcode, BppNWord index, bool forceNativeSize)
{
   BppBuffer* buf = codeVec[IDX_CODE];
   BppOpCodeByte ocb = (BppByte)opcode;
   if (ocb.opDataType() != ODT_NONE)
   {
	   buf->push_back(BppOpCodeByte(OP_NOP,ocb.opDataType()));
	   ocb._v._op.opDataType = ODT_NONE;
   }
   if (forceNativeSize)
   {
	   ocb._v._op.opDataType = ODT_NWORD;
	   buf->push_back(ocb);
	   buf->push_backNW(index);
   }
   else if ((index & 0xFF) == index)// Byte
   {
	   ocb._v._op.opDataType = ODT_BYTE;
	   buf->push_back(ocb);
	   buf->push_back((BppByte)index);
   }
   else if ((index & 0xFFFF) == index)// Word
   {
	   ocb._v._op.opDataType = ODT_WORD;
	   buf->push_back(ocb);
	   buf->push_backW(index);
   }
   else // DWord
   {
	   ocb._v._op.opDataType = ODT_DWORD;
	   buf->push_back(ocb);
	   buf->push_backDW(index);
   }
}

size_t BppCompiler::_codeOpcode(BppVector& codeVec, BppOpCode opcode)
{
   BppBuffer* buf = codeVec[IDX_CODE];
   BppOpCodeByte ocb = (BppByte)opcode;
   if (ocb.opDataType() != ODT_NONE)
   {
	   buf->push_back(BppOpCodeByte(OP_NOP,ocb.opDataType()));
	   ocb._v._op.opDataType = ODT_NONE;
   }
   return buf->push_back(ocb);
}

bool BppCompiler::_fetchRequire(BppToken rtoken, BppChar ch)
{
   return _require(rtoken,_scanner->getToken(),ch);
}

bool BppCompiler::_require(BppToken rtoken, BppToken token, BppChar ch)
{
   if ((token != rtoken) || ( (token == T_ANY) && (ch != 0) && (ch != _scanner->getChar()) ))
   {
      if (rtoken == T_ANY)
      {
		  BppChar chbuf[2];
         chbuf[0] = ch;
         chbuf[1] = 0;
		 _error(BppString::format(_T("expected '%s', found '%s'"), (const BppChar*)chbuf, _scanner->getTokenName(token)).c_str());
      }
      else
      {
		  _error(BppString::format(_T("expected '%s', found '%s'"), _scanner->getTokenName(rtoken), _scanner->getTokenName(token)).c_str());
      }
      return false;
   }
   return true;
}

bool BppCompiler::_checkLValue(const BppVarInfo& varInfo)
{
   if ((varInfo.type == VAR_NONE) || (varInfo.type == VAR_DONE))
   {
      _error(_T("lvalue expected"));
      return false;
   }
   return true;
}

void BppCompiler::_error(const BppChar * msg, int * errCnt)
{
   _scanner->parseError(msg);
   _print(_T("%s"),_scanner->getTokenString());
   if (errCnt) ++(*errCnt);
}

#ifndef DOXY
/**
If last instruction in current code buffer is one of assignment operations
OP_ASET, OP_SET, OP_TSET, OP_SSET or OP_MSET, function adds a OP_POPC operation
for stack correction (pops TOS and assigns its value to new TOS).
It is called on end of compiling a statement or argument of function call.
@param codeBuffer Bytecode buffer to modify
@param startIdx Start index in codeBuffer for beginning search for last instruction
(must point to opcode, not to an index value).
*/
//void BppCompiler::_adjustAssignment(BppBuffer& codeBuffer,size_t startIdx)
//{
//  
//	size_t bsize2 = codeBuffer.size();
//	unsigned lastopcode = 0;
//	for (size_t ib = startIdx; ib<bsize2;++ib)
//	{
//		BppOpCodeByte lastop = codeBuffer[ib];
//		lastopcode = lastop.opCode();
//		unsigned ltype = lastop.opDataType();
//		switch (ltype)
//		{
//		case ODT_BYTE:
//			ib++;
//			break;
//		case ODT_WORD:
//			ib+=2;
//			break;
//		case ODT_DWORD:
//			ib+=4;
//			break;
//		}
//	}
//	switch (lastopcode)
//	{
//	case OP_ASET:
//	case OP_SET:
//	case OP_TSET:
//		//case OP_VSET:
//		//case OP_PSET:
//	case OP_SSET:
//	case OP_MSET:
//		codeBuffer.push_back(BppOpCodeByte(OP_POPC));
//	}
//}

#endif
