#include "bppint.h"
#include "bppvm.h"
#include "bppdbg.h"
#include "bppdict.h"
#include "bppprint.h"
#include "bppbuf.h"
#include "bppobjct.h"
#include "bppclass.h"
#include "bppcomp.h"
#include "bppconst.h"
#include <math.h>
#include <stdlib.h>
#ifndef WINCE
  #include <assert.h>
#endif
#ifdef _MSC_VER
   #if _MSC_VER < 1300
      #define __FUNCTION__ ""
   #endif
#else
   #ifdef __TURBOC__
      #define __FUNCTION__ ""
   #endif
#endif

#ifdef GCC
#define T_FUNCTION BppString(__FUNCTION__).c_str()
#else
#define T_FUNCTION _T(__FUNCTION__)
#endif
/// Helper structure for implementation of exception handler stack.
struct BppErrorHandler
{
   BppErrorHandler(BppNWord pcount, const ValPtr& pstack, BppErrorHandler* nxt)
   :pc(pcount),sp(pstack),next(nxt)
   {
   }
   BppNWord pc; ///< Program counter (index in Bytecode buffer) to start of error handling code (vatch block)
   ValPtr sp; ///< Stack pointer at entry of protected (try) code block
   BppErrorHandler* next; ///< next item in exception handler stack
};

// procinfo implementation

BppProcInfo::BppProcInfo(BppInterpreter& interpreter, BppVector* bytecode,int nargs,ValPtr bp,ValPtr fp,bool nocaller)
:_interpreter(interpreter),_bytecode(bytecode),_nargs(nargs),_bp(bp), _fp(fp),_nocaller(nocaller),_parent(0),
_errHandlerStack(0)
{
}

BppProcInfo::~BppProcInfo()
{
   while (_errHandlerStack != 0)
   {
      popErrorHandler();
   }
   if (_parent != 0)
      delete _parent;
}

/**
Function checks last element of Bytecode vector for type BVT_LOCALVARINFO
and returns descriptor containing names of local variables.
@return Descriptor of local variable names on succes, otherwise 0.
The result is a vector of two elements, where first element contains vector of declared argument names and second
contains vector of local variable names. If function defines static local variables, last item of second element
is a dictionary (instead of a string), containing names of static variables as keys and indexes of variable values
in functions Bytecode vector as values.
@note Local variables descriptions are generated only if generation is switched on before running compiler.
@see BppVirtualMachine::setGenerateLocalVarInfo, getLocalVarName
*/
const BppVector* BppProcInfo::getLocalVarInfo() const
{
   const BppValue& v = (*_bytecode)[_bytecode->size()-1];
   if (v.getType() == BVT_LOCALVARINFO)
      return v;
   return 0;
}

/**
Function is intended for debugging purposes. It requires code to be compiled with generating local variables information enabled.
@param Index of variable to get name for.
@param Type of local variable.
@return Variable name on success, otherwise 0.
@see getLocalVarInfo, getLovalVarName(unsigned idx), getArgumentName, getLocalStaticVarName, BppVirtualMachine::setGenerateLocalVarInfo
*/
const BppChar * BppProcInfo::getLocalVarName(unsigned idx, LocalVarType lvt) const
{
   const BppVector* pi = getLocalVarInfo();
   if (pi != 0)
   {
      BppVector* pnames = (*pi)[lvt ? 1 : 0];
      if ((pnames != 0) && (pnames->size() > idx))
      {
         if (lvt == LVT_STATIC)
         {
            ValPtr lastTemp = pnames->last();
            if (lastTemp->getType() == BVT_DICTIONARY)
            {
               BppDictionary* statics = *lastTemp;
               if (statics != 0)
               {
                  for (BppDictionaryEntry* p = statics->getFirst(); p != 0; p = p->getNext())
                  {
                     if (((BppInt)p->getValue()) == ((BppInt)idx))
                        return p->getKey()->c_str();
                  }
               }
            }
         }
         else {
            const BppString* nm = (*pnames)[idx];
            if (nm != 0) return nm->c_str();
         }
      }
   }
   return 0;
}


BppProcInfo::operator BppByte*() const
{
   BppBuffer* cbuf = (*_bytecode)[IDX_CODE];
   if (cbuf != 0)
      return cbuf->data();
   return 0;
}


const BppChar* BppProcInfo::getName() const
{
   BppString* pname = (*_bytecode)[IDX_NAME];
   if (pname != 0) return pname->c_str();
   return _T("<unnamed>");
}

BppClass* BppProcInfo::getClass() const
{
   return (BppClass*)((*_bytecode)[IDX_CLASS]);
}

void BppProcInfo::pushErrorHandler( BppNWord pc, const ValPtr& pStack )
{
   _errHandlerStack = new BppErrorHandler(pc,pStack,_errHandlerStack);
}

bool BppProcInfo::popErrorHandler()
{
   if (_errHandlerStack == 0) return false;
   BppErrorHandler* p = _errHandlerStack;
   _errHandlerStack = p->next;
   delete p;
   return true;
}


// BppInterpreter implementation
BppInterpreter::BppInterpreter(BppVirtualMachine& vm)
:_stack(32),_vm(vm),_classes(vm.classes()),_symbols(vm.symbols()),_printer(vm.printer()),_procStack(0),
_debugger(0),_trace(false),_step(false),_allowBreakRequest(false),_requestCancel(false)
{
   _stack.addRef();
   _errBuffer.addRef();
   _sp=_stack.last();
   ++_sp;
}

BppInterpreter::~BppInterpreter()
{
   _stack.clear();
   if (_procStack != 0)
      delete _procStack;
}


void BppInterpreter::setDebugger(BppDebugger* debugger)
{
   if (_debugger == debugger) return;
   if (_debugger != 0)
   {
      _debugger->setInterpreter(0);
      _debugger->setCurrentProc(0);
      _trace = false;
      _step = 0;
   }
   _debugger = debugger;
   if (_debugger != 0)
   {
      _debugger->setInterpreter(this);
      _debugger->setCurrentProc(0);
      _trace = _debugger->active;
   }
}

void BppInterpreter::reset()
{
   _stack.clear();
   _sp = _stack.last();
   ++_sp;
   _errBuffer.clear();
   _errStack.clear();
   _requestCancel = false;
}

bool BppInterpreter::execute(const BppChar* name, int argc, const BppChar* argv[])
{
   C_LOCALE;
   reset();
   if (!name) return false;
    /* lookup the symbol */
   BppDictionaryEntry* sym = _symbols.find(name);
   if (!sym) return false;
   //ValPtr savedSp = _sp;
   //ValPtr savedLast = _stack.last();
   for (int i=0; i<argc;++i) // push arguments to stack
      _stack.push_back(new BppString(argv[i]));
   _sp = _stack.last(); // _sp points to TOS
   const BppValue& val = sym->getValue();
    /* dispatch on its data type */
   bool ok = false;
    switch (val.getType())
    {
    case BVT_CODE:
       {
          BPPUSERFUNC f = val;
          ok = f(&_vm,argc,_sp);
       }
    case BVT_BYTECODE:
    case BVT_BYTECODEREF:
	    ok = call(argc,val);
    }
    //if (_sp.current() != savedSp.current()) {
    //   *savedSp = *_sp;
    //   _sp = savedSp;
    //}
    //int j = _stack.last().current() - savedLast.current();
    //while (--j >= 0) _stack.pop_back();
    return ok;
}

void BppInterpreter::enterProc(BppVector* bytecode,int nargs,ValPtr bp, ValPtr fp)
{
   BppProcInfo* pi = new BppProcInfo(*this,bytecode,nargs,bp,fp);
   pi->setParent(_procStack);
   _procStack = pi;
   if (_trace || _step)
      _debugger->setCurrentProc(_procStack);

}

void BppInterpreter::leaveProc()
{
   if (!_procStack) return;
   BppProcInfo* pi = _procStack;
   _procStack = pi->getParent();
   pi->setParent(0);
   if (_trace || _step)
      _debugger->setCurrentProc(_procStack);
   delete pi;
}

bool BppInterpreter::safeAssign(BppValue & dest, const BppValue & src)
{
   if (!dest.assign(src))
   {
         return badType(src, dest);
   }
	return true;
}

struct __BppProcHelper
{
   __BppProcHelper(BppInterpreter* owner) : _owner(owner) {}
   ~__BppProcHelper()
   {
      _owner->leaveProc();
   }
   BppInterpreter* _owner;
};

bool BppInterpreter::call(int nargs, BppVector* byteCode)
{

      if (byteCode == 0)
         return returnError(_T("no bytecode: "), _T("interpret(BppVector* byteCode)\n"));
      /* make a dummy call frame */

   //  int NARGS = 0;
      BppVector* defargs = (*byteCode)[IDX_DEFAULTARGS];
      int definedArgs = (int)defargs->size();
      ValPtr BP = _sp;
      BP -= nargs; // points before 1st argument (i.e. to object in method calls)
      ++BP;
      for (int argi = 0; argi < nargs && argi < definedArgs; ++argi, ++BP)
      {
         if ((*defargs)[argi].getFlags() & BVF_STRONGTYPED)
         {
            // parameter type comversion
            BppValue tmp = (*defargs)[argi];
            tmp.v_flags = BVF_STRONGTYPED;
            if (!tmp.assign(*BP))
            {
                 return returnError(BppString::format(_T("Type mismatch for argument %d(%s) = (%s)."),
                  argi + 1,
                  getValueTypeName(tmp.getType()),
                  getValueTypeName(BP->getType())).c_str());
            }
            if (BP->getType() != tmp.getType())
               *BP = tmp;
            BP->setFlags(BP->getFlags() | BVF_STRONGTYPED);
         }
      }
      for (; nargs < definedArgs; ++nargs)
      {
         const BppValue& vref = (*defargs)[nargs];
         if ((vref.getFlags() & BVF_INVALID) == BVF_INVALID)
            return returnError(_T("Too few arguments specified."));
         switch (vref.v_type)
         {
         case BVT_BYTECODE:
         case BVT_LOCALVARINFO:
         case BVT_BUFFER:
         case BVT_CHARBUFFER:
         case BVT_OBJECT:
         case BVT_VECTOR:
         case BVT_STRING:
         case BVT_DICTIONARY:
            *(++_sp) = BppValue(vref, true);
            break;
         default:
            *(++_sp) = vref;
         }
         if (vref.getFlags() & BVF_STRONGTYPED)
            _sp->setFlags(_sp->getFlags() | BVF_STRONGTYPED);
      }
      BP = _sp;
      BP -= nargs; // points before 1st argument (i.e. to object in method calls)
      ValPtr FP = _sp; // FP points before 1st local (temporary) variable)
      int frameTop = FP.current() + 1;
      BppBuffer* codebuf = (*byteCode)[IDX_CODE];
      BppByte* pcode = codebuf->data();
      enterProc(byteCode, nargs, BP, FP);
      __BppProcHelper prochelper(this);
      BppClass* ownClass = (*byteCode)[IDX_CLASS];
      unsigned pc = 0;

      bool pending = false;
      while (!_requestCancel)
      {
         if (_allowBreakRequest && (!_step))
            _step = _vm.breakRequest();
         if (pending)
         {
            if (_procStack != 0)
            {
               BppErrorHandler* ehStack = _procStack->getEHStack();
               if (ehStack)
               {
                  pc = ehStack->pc;
                  const ValPtr& vp = ehStack->sp;
                  int n = vp.current() + 1;
                  //int  i = _stack.last().current() - vp.current();
                  while (_stack.size() > n) {
                     BppValue v = *_stack.last();
                     _stack.pop_back();
                  }
                  _sp = vp;
                  _procStack->popErrorHandler();
                  pending = false;
                  continue;
               }
               BppProcInfo* pparent = _procStack->getParent();
               if (pparent)
               {
                  // copy exception to parent procedure
                  pparent->fp()[1] = _procStack->fp()[1];
               }
            }
            return false;
         }
         if (_trace || _step)
         {
            if (_debugger)
            {
               if (_debugger->_currentProc != _procStack)
                  _debugger->setCurrentProc(_procStack);
               _debugger->traceInstruction(_sp, BP, FP, pcode, pc);
            }
            if (_step)
            {
               _step = _vm.dbgWait();
            }
         }
         BppOpCodeByte opcodebyte = pcode[pc++];
         BppOpCode opcode = (BppOpCode)opcodebyte._v._op.opCode;
         if (opcode == OP_NOP && opcodebyte.opDataType() != ODT_NONE) // NOP + opDataType-Bits means extended code
         {
            BppByte ocb = opcodebyte;
            opcodebyte = pcode[pc++];
            opcode = (BppOpCode)(((BppByte)opcodebyte.opCode()) | ocb);
         }
         BppDWord opdata = 0;
         switch (opcodebyte._v._op.opDataType)
         {
         case ODT_BYTE:
            opdata = (BppDWord)(BppByte)pcode[pc++];
            break;
         case ODT_WORD:
            opdata = BytePtr2W(pcode + pc);
            pc += sizeof(BppWord);
            break;
         case ODT_DWORD:
            opdata = BytePtr2DW(pcode + pc);
            pc += sizeof(BppDWord);
            break;
         }

         switch (opcode)
         {
         case OP_NOP: // 0x00, // no operation
            break;
         case OP_BRT: // 0x01, /* branch on true */
         case OP_BRF: // 0x02, /* branch on false */
         {
            bool t = false;
            if (_sp->inTypeRange(BVT_INTEGRAL_MIN, BVT_UINT))
            {
               t = BppInt(*_sp) != 0;
            }
            else if (_sp->isIntegralType())
            {
               t = BppLong(0L) != BppLong(*_sp);
            }
            else if (_sp->isRealType())
            {
               t = double(*_sp) != 0.0;
            }
            else
            {
               void* p = _sp->ptr();
               if (p && _sp->v_type == BVT_STRONGOBJECT)
                  p = (BppObject*)*_sp;
               t = (p != 0);
            }
            if (t != (opcode == OP_BRT))
               break;
         }
         case OP_BR: // 0x03, /* branch unconditionally */
            pc = (BppNWord)opdata;
            break;
         case OP_TRY:
            _procStack->pushErrorHandler(opdata, _sp);
            break;
         case OP_ENDTRY:
            _procStack->popErrorHandler();
            pc = (BppNWord)opdata;
            break;
         case OP_ENDCATCH:
            if (_errStack.size() == 0)
               FP[1] = BppValue();
            else {
               FP[1] = *_errStack.last();
               _errStack.pop_back(); // 0x33, dispose exception, if any
            }
            break;
         case OP_THROW:
            returnError(*_sp);
            pending = true;
            break;
         case OP_SWITCH:
            //pc += sizeof(BppNWord); --> pc bereits oben korrigiert
            break;
         case OP_NIL: // 0x04, /* load top of stack with nil */
            (*_sp).dispose();
            if (opdata)
            {
               BppValue v((BppRefObject*)0, (BppValueType)(opdata & 0xFF));
               v.setFlags(BVF_STRONGTYPED);
               (*_sp) = v;
            }
            break;
         case OP_PUSH: // 0x05, /* push nil onto stack */
            ++_sp;
            if (opdata)
            {
               (*_sp).dispose();
               (*_sp) = BppValue((BppValueType)(opdata & 0xFF), BVF_STRONGTYPED, false);
            }
            break;
         case OP_DUP: // 0x2D, /* duplicate TOS */
            (++_sp)->dispose();
            *_sp = _sp[-1];
            break;
         case OP_POPC: // remove topmost entry from stack, copy TOS value to new TOS
            if (_sp.current() <= frameTop) break;
            _sp[-1] = *_sp;
         case OP_POP: // 0x2E, remove topmost entry from stack
            _sp->dispose();
            --_sp;
            break;
         case OP_NOT: // 0x06, /* logical negate top of stack */
         {
            BppObject* po = (BppObject*)*_sp;
            if (po)
            {
               BppDictionaryEntry* pe = po->getClass()->findMember(ownClass,_T("OP_NOT"));
               if (pe && pe->getSymbolType() == ST_FUNCTION)
               {
                  pending = !send(0, _T("OP_NOT"), &pe->getValue());
                  break;
               }
            }
			
			switch (_sp->v_type) {
			case BVT_STRONGOBJECT:
			case BVT_OBJECT:
			case BVT_OBJECTREF:
			{
				(*_sp) = po ?  BppInt(0) : BppInt(1);
			}
			break;
			case BVT_FLOAT:
			case BVT_DOUBLE:
				(*_sp) = ((double)*_sp) == 0.0 ? BppInt(1) : BppInt(0);
				break;
			case BVT_LONG:
			case BVT_ULONG:
				(*_sp) = ((BppULong)*_sp) == BppULong(0) ? BppInt(1) : BppInt(0);
				break;
			default:
				(*_sp) = ((BppIntPtr)*_sp) == 0 ? BppInt(1) : BppInt(0);
			}
         }
         break;
         case OP_NEG: // 0x07, /* negate top of stack */
            switch (_sp->v_type)
            {
            case BVT_SBYTE:
            case BVT_BYTE:
            case BVT_SHORT:
            case BVT_USHORT:
               (*_sp) = -((BppInt)*_sp);
               break;
            case BVT_INT:
            case BVT_UINT:
               _sp->_v.v_integer = -_sp->_v.v_integer;
               break;
            case BVT_INTPTR:
#ifndef BPP_I64
               _sp->_v.v_integer = -_sp->_v.v_integer;
               break;
#endif
            case BVT_LONG:
            case BVT_ULONG:
               _sp->_v.v_long = -BppLong(_sp->_v.v_long);
               break;
            case BVT_FLOAT:
               _sp->_v.v_float = -_sp->_v.v_float;
               break;
            case BVT_DOUBLE:
               _sp->_v.v_double = -_sp->_v.v_double;
               break;
            case BVT_OBJECT:
            case BVT_OBJECTREF:
            case BVT_STRONGOBJECT:
               pending = (!send(0, _T("OP_NEG")));
               break;
            default:
               badType(_sp->v_type, BVT_INT);
               pending = true;
               break;
            }
            break;
         case OP_PLUS: // 0x3C, /* unary plus */
            switch (_sp->v_type)
            {
            case BVT_SBYTE:
            case BVT_BYTE:
            case BVT_SHORT:
            case BVT_USHORT:
               (*_sp) = ((BppInt)*_sp);
               break;
            case BVT_INT:
            case BVT_UINT:
            case BVT_LONG:
            case BVT_ULONG:
            case BVT_INTPTR:
            case BVT_FLOAT:
            case BVT_DOUBLE:
               break;
            case BVT_OBJECT:
            case BVT_OBJECTREF:
            case BVT_STRONGOBJECT:
               pending = (!send(0, _T("OP_PLUS")));
               break;
            default:
               badType(_sp->v_type, BVT_INT);
               pending = true;
               break;
            }
            break;
         case OP_ADD: // 0x08, /* add top two stack entries */
            pending = (!doOP_ADD());
            break;
         case OP_SUB: // 0x09, /* subtract top two stack entries */
            pending = (!doOP_SUB());
            break;
         case OP_MUL: // 0x0A, /* multiply top two stack entries */
            pending = (!doOP_MUL());
            break;
         case OP_DIV: // 0x0B, /* divide top two stack entries */
            pending = (!doOP_DIV());
            break;
         case OP_REM: // 0x0C, /* remainder of top two stack entries */
            pending = (!doOP_REM());
            break;
         case OP_BAND: // 0x0D, /* bitwise and of top two stack entries */
            pending = (!doOP_BAND());
            break;
         case OP_BOR: // 0x0E, /* bitwise or of top two stack entries */
            pending = (!doOP_BOR());
            break;
         case OP_XOR: // 0x0F, /* bitwise xor of top two stack entries */
            pending = (!doOP_XOR());
            break;
         case OP_BNOT: // 0x10, /* bitwise not of top two stack entries */
            pending = (!doOP_BNOT());
            break;
         case OP_SHL: // 0x11, /* shift left top two stack entries */
            pending = (!doOP_SHL());
            break;
         case OP_SHR: // 0x12, /* shift right top two stack entries */
            pending = (!doOP_SHR());
            break;
         case OP_MKREF: // 0x46, /* make reference from TOS*/
            pending = (!doOP_MKREF());
            break;
         case OP_UNREF: // make value from reference on TOS
            pending = (!doOP_UNREF(ownClass));
            break;
         case OP_LT: // 0x13, /* less than */
            if (doCompare(ownClass))
               (*_sp) = (BppInt)((_sp->_v.v_integer < 0) ? 1 : 0);
            else pending = true;
            break;
         case OP_LE: // 0x14, /* less than or equal to */
            if (doCompare(ownClass))
               (*_sp) = (BppInt)((_sp->_v.v_integer <= 0) ? 1 : 0);
            else pending = true;
            break;
         case OP_EQ: // 0x15, /* equal to */
            if (!doEquals(ownClass))
               pending = true;
            break;
         case OP_NE: // 0x16, /* not equal to */
            if (doEquals(ownClass))
               (*_sp) = (BppInt)((_sp->_v.v_integer == 0) ? 1 : 0);
            else pending = true;
            break;
         case OP_GE: // 0x17, /* greater than or equal to */
            if (doCompare(ownClass))
               (*_sp) = (BppInt)((_sp->_v.v_integer >= 0) ? 1 : 0);
            else pending = true;
            break;
         case OP_GT: // 0x18, /* greater than */
            if (doCompare(ownClass))
               (*_sp) = (BppInt)((_sp->_v.v_integer > 0) ? 1 : 0);
            else pending = true;
            break;
         case OP_INC: // 0x19, /* increment */
         case OP_PINC: // 0x37, post increment
            switch (_sp->v_type)
            {
            case BVT_SBYTE:
            case BVT_BYTE:
               _sp->_v.v_sbyte++;
               break;
            case BVT_SHORT:
            case BVT_USHORT:
               _sp->_v.v_short++;
               break;
            case BVT_INT:
            case BVT_UINT:
               _sp->_v.v_integer++;
               break;
            case BVT_INTPTR:
#ifndef BPP_I64
               _sp->_v.v_integer++;
               break;
#endif
            case BVT_LONG:
            case BVT_ULONG:
#ifdef EMULATE_I64
               _sp->_v.v_long = ++BppLong(_sp->_v.v_long);
#else
               _sp->_v.v_long++;
#endif
               break;
            case BVT_FLOAT:
               _sp->_v.v_float++;
               break;
            case BVT_DOUBLE:
               _sp->_v.v_double++;
               break;
            case BVT_OBJECT:
            case BVT_OBJECTREF:
            case BVT_STRONGOBJECT:
               pending = (!send(0, (opcode == OP_INC) ? _T("OP_INC") : _T("OP_PINC")));
               break;
            default:
               badType(_sp->v_type, BVT_INT);
               pending = true;
               break;
            }
            break;
         case OP_DEC: // 0x1A, /* decrement */
         case OP_PDEC: // 0x38, post decrement
            switch (_sp->v_type)
            {
            case BVT_SBYTE:
            case BVT_BYTE:
               _sp->_v.v_sbyte--;
               break;
            case BVT_SHORT:
            case BVT_USHORT:
               _sp->_v.v_short--;
               break;
            case BVT_INT:
            case BVT_UINT:
               _sp->_v.v_integer--;
               break;
            case BVT_INTPTR:
#ifndef BPP_I64
               _sp->_v.v_integer--;
               break;
#endif
            case BVT_LONG:
            case BVT_ULONG:
#ifdef EMULATE_I64
               _sp->_v.v_long = --BppLong(_sp->_v.v_long);
#else
               _sp->_v.v_long--;
#endif
               break;
            case BVT_FLOAT:
               _sp->_v.v_float--;
               break;
            case BVT_DOUBLE:
               _sp->_v.v_double--;
               break;
            case BVT_OBJECT:
            case BVT_OBJECTREF:
            case BVT_STRONGOBJECT:
               pending = (!send(0, (opcode == OP_DEC) ? _T("OP_DEC") : _T("OP_PDEC")));
               break;
            default:
               badType(_sp->v_type, BVT_INT);
               pending = true;
               break;
            }
            break;
         case OP_LIT: // 0x1B, /* load literal */
         {
            const BppValue& vref = (*byteCode)[opdata];
            switch (vref.v_type)
            {
            case BVT_BYTECODE:
            case BVT_LOCALVARINFO:
            case BVT_BUFFER:
            case BVT_CHARBUFFER:
            case BVT_OBJECT:
            case BVT_VECTOR:
            case BVT_STRING:
            case BVT_DICTIONARY:
               *_sp = BppValue(vref, true);
               break;
            default:
               *_sp = vref;
            }
            break;
         }
         case OP_SREF:
            *_sp = (*byteCode)[opdata];
            break;
         case OP_SSET:
            pending = !safeAssign((*byteCode)[opdata], *_sp);
            break;
         case OP_RETURN: // 0x1C, /* return from interpreter */
            if (FP[1].getType() != BVT_NULL)
            {
               FP[1].dispose(); // dispose exception, if any
               if (_errStack.size() > 0) {
                  BppValue v = *_stack.last();
                  _errStack.pop_back();
               }
            }
            if (opdata)
            {
               BppValueType rtype = (BppValueType)opdata;
               BppValueType currtype = _sp->getType();
               if (currtype != rtype)
               {
                  if (currtype == BVT_NULL &&
                     rtype > _BVTVALMAX &&
                     rtype <= _BVTREFMAX) return true;
                  if ((rtype == BVT_CODE && currtype == BVT_BYTECODE) ||
                     (rtype == BVT_BYTECODE && currtype == BVT_CODE) ||
                     (rtype == BVT_OBJECT && currtype == BVT_STRONGOBJECT) ||
                     (rtype == BVT_STRONGOBJECT && currtype == BVT_OBJECT))
                     return true;

                  BppString buf = BppString::format(_T("invalid return type %s, %s expected"), getValueTypeName(_sp->getType()), getValueTypeName(rtype));
                  returnError(buf.c_str());
                  return false;
               }
            }
            return true;
         case OP_CALL: // 0x1D, /* call a function */
         {
            // BppNWord n = opdata;
            ValPtr vp = _sp;
            vp -= opdata;
            switch ((*vp).v_type)
            {
            case BVT_CODE:
            //case BVT_CODEREF:
               pending = (!(vp->_v.v_code)(&_vm, opdata, _sp));
               break;
            case BVT_BYTECODE:
            case BVT_BYTECODEREF:
               pending = (!call(opdata, *vp));
               break;
            case BVT_OBJECT:
            case BVT_OBJECTREF:
            case BVT_STRONGOBJECT:
               pending = (!send(opdata, _T("OP_CALL")));
               break;
            case BVT_CLASS:
            case BVT_CLASSREF: // handle explicit cast to class type
               if (opdata == 1) {
                  BppClass* pcl = *vp;
                  if (pcl)
                  {
                     BppString* pclname = pcl->getClassName();
                     BppObject* pobj = *_sp;
                     if (pobj)
                     {
                        BppClass* psc = pobj->getClass();
                        *vp = pobj;
                        _sp->dispose();
                        --_sp;
                        if (!psc->isSubtypeOf(pcl))
                        {
                           pending = (!send(0, pclname->c_str()));
                        }
                        break;
                     }
                  }
               }
            default:
               returnError(_T("call to non-procedure, type "), getValueTypeName(vp->getType()));
               pending = true;
               break;
            }
            // cleanup stack
            if (_sp.current() > (vp.current()/*+1*/))
            {
               // vp++;
               *vp = *_sp;
               int n = vp.current()+1;
               //int n = _stack.size() - (vp.current());
               while (_stack.size() > n) {
                  BppValue v = * _stack.last(); // prevent from object deletion while stack is modified
                  _stack.pop_back();
               }
               _sp = vp;
            }
            else if (_sp.current() < vp.current())
               pending = (!returnError(_T("stack corrupted")));
         }
         break;
         case OP_REF: // 0x1E, /* load a variable value */
         case OP_SET: // 0x1F, /* set the value of a variable */
         {
            //BppNWord n = opdata;
            BppDictionaryEntry* pvar = ((*byteCode)[opdata]);
            if (!pvar)
            {
               pending = (!returnError(_T("invalid variable reference "), getValueTypeName((BppValueType)((*byteCode)[opdata]).v_type)));
               break;
            }
            if (opcode == OP_REF)
               *_sp = pvar->getValue();
            else
            {
               pending = !safeAssign((BppValue&)(pvar->getValue()), *_sp);
            }
         }
         break;
         case OP_VREF: // 0x20, /* load a vector element */
            pending = (!doOP_VREF());
            break;
         case OP_VSET: // 0x21, /* set a vector element */
            pending = (!doOP_VSET());
            break;
         case OP_PREF: //load property value */
            pending = (!doOP_PREF());
            break;
         case OP_PSET: //set property */
            pending = (!doOP_PSET());
            break;
         case OP_MREF: // 0x22, /* load a member variable value */
         case OP_MSET: // 0x23, /* set a member variable */
         {
            BppObject* obj = BP[1];
            if (obj == 0)
            {
               pending = (!returnError(_T("object expected")));
               break;
            }
            //BppNWord n = opdata;
            BppVector* mdata = (obj->getMemberData());
            if (mdata == 0)
            {
               pending = (!returnError(_T("member data expected")));
               break;
            }
            BppValue& v = (*mdata)[opdata];
            if (opcode == OP_MREF)
               *_sp = v;
            else
               pending = !safeAssign(v, *_sp);
         }
         break;
         case OP_AREF: // 0x24, /* load an argument value */
         case OP_ASET: // 0x25, /* set an argument value */
         {
            //BppNWord n = opdata;
            if (opdata >= (unsigned)nargs)
            {
               pending = (!returnError(_T("too few arguments")));
               break;
            }
            if (opcode == OP_AREF)
               *_sp = BP[(int)opdata + 1]; // BP points before 1st arg
            else
               pending = !safeAssign(BP[(int)opdata + 1], *_sp);
         }
         break;
         case OP_TREF: // 0x26, /* load a temporary variable value */
            *_sp = FP[(int)opdata + 1]; // FP points before 1st temporary
            break;
         case OP_TSET: // 0x27, /* set a temporary variable */
            pending = !safeAssign(FP[(int)opdata + 1], *_sp);
            break;
         case OP_TSPACE: // 0x28, /* allocate temporary variable space */
         {
            //BppNWord n = opdata;
            FP = _sp; // FP points before 1st temporary
            for (BppNWord i = 0; i < opdata; ++i)
               (++_sp)->dispose(); // inc _sp and set value to NULL
            frameTop += opdata;
         }
         break;
         case OP_SEND: // 0x29, /* send a message to an object */
            pending = (!send((int)opdata));
            break;
         case OP_DUP2: // 0x2A, /* duplicate top two elements on the stack */
            _sp += 2;
            _sp[-1] = _sp[-3];
            *_sp = _sp[-2];
            break;
         case OP_NEW: // 0x2B, /* create a new class object */
         {
            BppClass* pclass = *_sp;
            if (pclass)
               *_sp = new BppObject(pclass);
            else
               pending = !checkType(*_sp, BVT_CLASS);
         }
         break;
         case OP_DELETE: // 0x2C,    /* delete object instance */
            if (_sp->v_type >= _BVTREFMIN)
            {
               BppRefObject* robj = _sp->getRef();
               if ((robj != 0) && (robj->getRefCnt() == 0))
                  delete robj;
               _sp->dispose();  // make NULL result
            }
            else
                pending = !(_sp->assign(BppValue()));
            break;
            // trace codes
         case OP_TRON:
            if (_debugger != 0)
            {
               if (!(_trace || _step)) _debugger->setCurrentProc(_procStack);
               _trace = true;
            }
            break;
         case OP_TROFF:
            _trace = false;
            break;
         case OP_TRSTEP:
            if (_debugger != 0)
            {
               if (!(_trace || _step)) _debugger->setCurrentProc(_procStack);
               _step = true;
            }
            break;
         case OP_NEWLINE: // 0x83, /* notify new line */
         {
            BppString* s = *_sp;
            if (s != 0)
               _step = _vm.notifySourceLineChanged(s->c_str(), (long)opdata);
            --_sp;
         }
         break;
         }
      }

   return true;
}

// opcode handler functions
bool BppInterpreter::doOP_ADD()
{
   BppValue& right = *_sp;
   BppValue& left = _stack[_sp.current()-1];
   BppValueType tl = (BppValueType)left.v_type;
   BppValueType tr = (BppValueType)right.v_type;
   BppValueType rt = tl < tr ? tr : tl;
   if ((tl != BVT_OBJECT) && (tl != BVT_OBJECTREF) && (tl != BVT_STRONGOBJECT))
   {
      if ((tr == BVT_OBJECT) || (tr == BVT_OBJECTREF) || tr == BVT_STRONGOBJECT)
      {
         BppValue tmp = right;
         right = left;
         left = tmp;
         return send(1,_T("OP_ADD_R"));
      }
   }
   else rt = tl;
   if (tl == BVT_NULL || tr == BVT_NULL) rt = BVT_NULL;
   else if (rt < BVT_INT) rt = BVT_INT;
   switch (rt)
   {
   case BVT_OBJECT:
   case BVT_OBJECTREF:
   case BVT_STRONGOBJECT:
      return send(1,_T("OP_ADD"));
   case BVT_INT:
	   left = (BppInt)left + (BppInt)right;
	   break;
   case BVT_UINT:
	   left = (BppUInt)left + (BppUInt)right;
	   break;
   case BVT_LONG:
	   left = BppLong(left) + BppLong(right);
	   break;
   case BVT_ULONG:
	   left = BppULong(left) + BppULong(right);
	   break;
   case BVT_INTPTR:
	   left = BppIntPtr(left) + BppIntPtr(right);
	   break;
   case BVT_FLOAT:
	   left = (float)left + (float)right;;
	   break;
   case BVT_DOUBLE:
	   left = (double)left + (double)right;
	   break;
   case BVT_STRING:
   case BVT_STRINGREF:
	   if (tl == BVT_STRING || tl == BVT_STRINGREF)
	   {
           BppString* sl = left;
		   if (!sl) return returnError(_T("left string operand is NULL"));
           if (tr == BVT_STRING || tr == BVT_STRINGREF)
		   {
			   BppString* sr = right;
			   if (!sr) return returnError(_T("right string operand is NULL"));
			   BppString* s = sl->clone();
			   (*s) += *sr;
			   left = s;
		   }
		   else 
		   {
			   if (!right.isIntegralType()) return badType(tr,BVT_STRING);
			   BppString* s = sl->clone();
			   (*s) += (BppChar)(BppInt)right;
			   left = s;
		   }
	   }
	   else
	   {
		   if (!left.isIntegralType()) return badType(tr,BVT_STRING);
		   BppString* sr = right;
		   if (!sr) return returnError(_T("right string operand is NULL"));
		   BppChar c[2] = {0,0};
		   c[0] = (BppChar)(BppInt)left;
		   BppString* s = new BppString(c);
		   (*s) += *sr;
		   left = s;
	   }
	   break;
   default:
      return badType(rt,BVT_INT);
   }
   --_sp; // _sp points to result (position of left argument)
   return true;
}


bool BppInterpreter::doOP_SUB()
{
   BppValue& right = *_sp;
   BppValue& left = _stack[_sp.current()-1];
   BppValueType tl = (BppValueType) left.v_type;
   BppValueType tr = (BppValueType) right.v_type;
   BppValueType rt = tl < tr ? tr : tl;
   if ((tl != BVT_OBJECT) && (tl != BVT_OBJECTREF) && (tl != BVT_STRONGOBJECT))
   {
      if ((tr == BVT_OBJECT) || (tr == BVT_OBJECTREF) || tr == BVT_STRONGOBJECT)
      {
         BppValue tmp = right;
         right = left;
         left = tmp;
         return send(1,_T("OP_SUB_R"));
      }
   }
   else rt = tl;
   if (tl == BVT_NULL || tr == BVT_NULL) rt = BVT_NULL;
   else if (rt < BVT_INT) rt = BVT_INT;
   switch (rt)
   {
   case BVT_OBJECT:
   case BVT_OBJECTREF:
   case BVT_STRONGOBJECT:
      return send(1,_T("OP_SUB"));
   case BVT_INT:
	   left = (BppInt)left - (BppInt)right;
	   break;
   case BVT_UINT:
	   left = (BppUInt)left - (BppUInt)right;
	   break;
   case BVT_LONG:
	   left = BppLong(left) - BppLong(right);
	   break;
   case BVT_ULONG:
	   left = BppULong(left) - BppULong(right);
	   break;
   case BVT_INTPTR:
	   left = BppIntPtr(left) - BppIntPtr(right);
	   break;
   case BVT_FLOAT:
	   left = (float)left - (float)right;;
	   break;
   case BVT_DOUBLE:
	   left = (double)left - (double)right;
	   break;
   default:
      return badType(rt,BVT_INT);
   }
   --_sp; // _sp points to result (position of left argument)
   return true;
}


bool BppInterpreter::doOP_MUL()
{
   BppValue& right = *_sp;
   BppValue& left = _stack[_sp.current()-1];
   BppValueType tl = (BppValueType)left.v_type;
   BppValueType tr = (BppValueType)right.v_type;
   BppValueType rt = tl < tr ? tr : tl;
   if ((tl != BVT_OBJECT) && (tl != BVT_OBJECTREF) && (tl != BVT_STRONGOBJECT))
   {
      if ((tr == BVT_OBJECT) || (tr == BVT_OBJECTREF) || tr == BVT_STRONGOBJECT)
      {
         BppValue tmp = right;
         right = left;
         left = tmp;
         return send(1,_T("OP_MUL_R"));
      }
   }
   else rt = tl;
   if (tl == BVT_NULL || tr == BVT_NULL) rt = BVT_NULL;
   else if (rt < BVT_INT) rt = BVT_INT;
   switch (rt)
   {
   case BVT_OBJECT:
   case BVT_OBJECTREF:
   case BVT_STRONGOBJECT:
      return send(1,_T("OP_MUL"));
   case BVT_INT:
	   left = (BppInt)left * (BppInt)right;
	   break;
   case BVT_UINT:
	   left = (BppUInt)left * (BppUInt)right;
	   break;
   case BVT_LONG:
	   left = BppLong(left) * BppLong(right);
	   break;
   case BVT_ULONG:
	   left = BppULong(left) * BppULong(right);
	   break;
   case BVT_INTPTR:
	   left = BppIntPtr(left) * BppIntPtr(right);
	   break;
   case BVT_FLOAT:
	   left = (float)left * (float)right;;
	   break;
   case BVT_DOUBLE:
	   left = (double)left * (double)right;
	   break;
   default:
      return badType(rt,BVT_INT);
   }
   --_sp; // _sp points to result (position of left argument)
   return true;
}


bool BppInterpreter::doOP_DIV()
{
   BppValue& right = *_sp;
   if (right.isIntegralType() && ((BppInt)right) == 0 && ((double)right) == 0.0)
      return returnError(_T("divide 0 error"));
   BppValue& left = _stack[_sp.current()-1];
   BppValueType tl = (BppValueType)left.v_type;
   BppValueType tr = (BppValueType)right.v_type;
   BppValueType rt = tl < tr ? tr : tl;
   if ((tl != BVT_OBJECT) && (tl != BVT_OBJECTREF) && (tl != BVT_STRONGOBJECT))
   {
      if ((tr == BVT_OBJECT) || (tr == BVT_OBJECTREF) || tr == BVT_STRONGOBJECT)
      {
         BppValue tmp = right;
         right = left;
         left = tmp;
         return send(1,_T("OP_DIV_R"));
      }
   }
   else rt = tl;
   if (tl == BVT_NULL || tr == BVT_NULL) rt = BVT_NULL;
   else if (rt < BVT_INT) rt = BVT_INT;
   switch (rt)
   {
   case BVT_OBJECT:
   case BVT_OBJECTREF:
   case BVT_STRONGOBJECT:
      return send(1,_T("OP_DIV"));
   case BVT_INT:
	   left = (BppInt)left / (BppInt)right;
	   break;
   case BVT_UINT:
	   left = (BppUInt)left / (BppUInt)right;
	   break;
   case BVT_LONG:
	   left = BppLong(left) / BppLong(right);
	   break;
   case BVT_ULONG:
	   left = BppULong(left) / BppULong(right);
	   break;
   case BVT_INTPTR:
	   left = BppIntPtr(left) / BppIntPtr(right);
	   break;
   case BVT_FLOAT:
	   left = (float)left / (float)right;;
	   break;
   case BVT_DOUBLE:
	   left = (double)left / (double)right;
	   break;
   default:
      return badType(rt,BVT_INT);
   }
   --_sp; // _sp points to result (position of left argument)
   return true;
}


bool BppInterpreter::doOP_REM()
{
   BppValue& right = *_sp;
   if (right.isIntegralType() && ((BppInt)right)== 0 && ((double)right)==0.0)
      return returnError(_T("divide 0 error"));
   BppValue& left = _stack[_sp.current()-1];
   BppValueType tl = (BppValueType)left.v_type;
   BppValueType tr = (BppValueType)right.v_type;
   BppValueType rt = tl < tr ? tr : tl;
   if ((tl != BVT_OBJECT) && (tl != BVT_OBJECTREF) && (tl != BVT_STRONGOBJECT))
   {
      if ((tr == BVT_OBJECT) || (tr == BVT_OBJECTREF) || tr == BVT_STRONGOBJECT)
      {
         BppValue tmp = right;
         right = left;
         left = tmp;
         return send(1,_T("OP_REM_R"));
      }
   }
   else rt = tl;
   if (tl == BVT_NULL || tr == BVT_NULL) rt = BVT_NULL;
   else if (rt < BVT_INT) rt = BVT_INT;
   switch (rt)
   {
   case BVT_OBJECT:
   case BVT_OBJECTREF:
   case BVT_STRONGOBJECT:
      return send(1,_T("OP_REM"));
   case BVT_INT:
	   left = (BppInt)left % (BppInt)right;
	   break;
   case BVT_UINT:
	   left = (BppUInt)left % (BppUInt)right;
	   break;
   case BVT_LONG:
	   left = BppLong(left) % BppLong(right);
	   break;
   case BVT_ULONG:
	   left = BppULong(left) % BppULong(right);
	   break;
   case BVT_INTPTR:
	   left = BppIntPtr(left) % BppIntPtr(right);
	   break;
   case BVT_FLOAT:
	   left = (float) fmod((float)left,(float)right);
	   break;
   case BVT_DOUBLE:
	   left = (double) fmod((double)left,(double)right);
	   break;
   default:
      return badType(rt,BVT_INT);
   }
   --_sp; // _sp points to result (position of left argument)
   return true;
}


bool BppInterpreter::doOP_BAND()
{
   BppValue& right = *_sp;
   BppValue& left = _stack[_sp.current()-1];
   BppValueType tl = (BppValueType)left.v_type;
   BppValueType tr = (BppValueType)right.v_type;
   BppValueType rt = tl < tr ? tr : tl;
   if ((tl != BVT_OBJECT) && (tl != BVT_OBJECTREF) && (tl != BVT_STRONGOBJECT))
   {
      if ((tr == BVT_OBJECT) || (tr == BVT_OBJECTREF) || tr == BVT_STRONGOBJECT)
      {
         BppValue tmp = right;
         right = left;
         left = tmp;
         return send(1,_T("OP_BAND_R"));
      }
   }
   else rt = tl;
   if (tl == BVT_NULL || tr == BVT_NULL) rt = BVT_NULL;
   else if (rt < BVT_INT) rt = BVT_INT;
   switch (rt)
   {
   case BVT_OBJECT:
   case BVT_OBJECTREF:
   case BVT_STRONGOBJECT:
      return send(1,_T("OP_BAND"));
   case BVT_INT:
	   left = (BppInt)left & (BppInt)right;
	   break;
   case BVT_UINT:
	   left = (BppUInt)left & (BppUInt)right;
	   break;
   case BVT_LONG:
	   left = BppLong(left) & BppLong(right);
	   break;
   case BVT_ULONG:
	   left = BppULong(left) & BppULong(right);
	   break;
   case BVT_INTPTR:
	   left = BppIntPtr(left) & BppIntPtr(right);
	   break;
   default:
      return badType(rt,BVT_INT);
   }
   --_sp; // _sp points to result (position of left argument)
   return true;
}

bool BppInterpreter::doOP_BOR()
{
   BppValue& right = *_sp;
   BppValue& left = _stack[_sp.current()-1];
   BppValueType tl = (BppValueType)left.v_type;
   BppValueType tr = (BppValueType)right.v_type;
   BppValueType rt = tl < tr ? tr : tl;
   if ((tl != BVT_OBJECT) && (tl != BVT_OBJECTREF) && (tl != BVT_STRONGOBJECT))
   {
      if ((tr == BVT_OBJECT) || (tr == BVT_OBJECTREF) || tr == BVT_STRONGOBJECT)
      {
         BppValue tmp = right;
         right = left;
         left = tmp;
         return send(1,_T("OP_BOR_R"));
      }
   }
   else rt = tl;
   if (tl == BVT_NULL || tr == BVT_NULL) rt = BVT_NULL;
   else if (rt < BVT_INT) rt = BVT_INT;
   switch (rt)
   {
   case BVT_OBJECT:
   case BVT_OBJECTREF:
   case BVT_STRONGOBJECT:
      return send(1,_T("OP_BOR"));
   case BVT_INT:
	   left = (BppInt)left | (BppInt)right;
	   break;
   case BVT_UINT:
	   left = (BppUInt)left | (BppUInt)right;
	   break;
   case BVT_LONG:
	   left = BppLong(left) | BppLong(right);
	   break;
   case BVT_ULONG:
	   left = BppULong(left) | BppULong(right);
	   break;
   case BVT_INTPTR:
	   left = BppIntPtr(left) | BppIntPtr(right);
	   break;
   default:
      return badType(rt,BVT_INT);
   }
   --_sp; // _sp points to result (position of left argument)
   return true;
}

bool BppInterpreter::doOP_XOR()
{
   BppValue& right = *_sp;
   BppValue& left = _stack[_sp.current()-1];
   BppValueType tl = (BppValueType)left.v_type;
   BppValueType tr = (BppValueType)right.v_type;
   BppValueType rt = tl < tr ? tr : tl;
   if ((tl != BVT_OBJECT) && (tl != BVT_OBJECTREF) && (tl != BVT_STRONGOBJECT))
   {
      if ((tr == BVT_OBJECT) || (tr == BVT_OBJECTREF) || tr == BVT_STRONGOBJECT)
      {
         BppValue tmp = right;
         right = left;
         left = tmp;
         return send(1,_T("OP_XOR_R"));
      }
   }
   else rt = tl;
   if (tl == BVT_NULL || tr == BVT_NULL) rt = BVT_NULL;
   else if (rt < BVT_INT) rt = BVT_INT;
   switch (rt)
   {
   case BVT_OBJECT:
   case BVT_OBJECTREF:
   case BVT_STRONGOBJECT:
      return send(1,_T("OP_XOR"));
   case BVT_INT:
	   left = (BppInt)left ^ (BppInt)right;
	   break;
   case BVT_UINT:
	   left = (BppUInt)left ^ (BppUInt)right;
	   break;
   case BVT_LONG:
	   left = BppLong(left) ^ BppLong(right);
	   break;
   case BVT_ULONG:
	   left = BppULong(left) ^ BppULong(right);
	   break;
   case BVT_INTPTR:
	   left = BppIntPtr(left) ^ BppIntPtr(right);
	   break;
   default:
      return badType(rt,BVT_INT);
   }
   --_sp; // _sp points to result (position of left argument)
   return true;
}

bool BppInterpreter::doOP_BNOT()
{
   BppValue& op = *_sp;
   if ((op.v_type == BVT_OBJECT) || (op.v_type == BVT_OBJECTREF) || op.v_type == BVT_STRONGOBJECT)
      return send(0,_T("OP_BNOT"));
   BppValueType vt = op.getType();
   if (!op.isIntegralType()) return badType(vt,BVT_INT);
   if (vt < BVT_INT) vt = BVT_INT;
   switch (vt)
   {
	   case BVT_INT:
		   op = ~((BppInt)op);
		   break;
	   case BVT_UINT:
		   op = ~((BppUInt)op);
		   break;
	   case BVT_LONG:
		   op = ~(BppLong(op));
		   break;
	   case BVT_ULONG:
		   op = ~(BppULong(op));
		   break;
	   case BVT_INTPTR:
		   op = BppValue(~(BppIntPtr(op)),true);
		   break;
	   default:
		   return badType(vt,BVT_INT);
   }
   return true;
}

bool BppInterpreter::doOP_SHL()
{
	BppValue& right = *_sp;
	BppValue& left = _stack[_sp.current() - 1];
	BppValueType tl = (BppValueType)left.v_type;
	BppValueType tr = (BppValueType)right.v_type;
	if ((tl != BVT_OBJECT) && (tl != BVT_OBJECTREF) && (tl != BVT_STRONGOBJECT))
	{
		if ((tr == BVT_OBJECT) || (tr == BVT_OBJECTREF) || tr == BVT_STRONGOBJECT)
		{
			BppValue tmp = right;
			right = left;
			left = tmp;
			return send(1, _T("OP_SHL_R"));
		}
		else if (!right.isIntegralType()) return badType(tr, BVT_INT);
	}
	else if (tl > BVT_NULL && tl < BVT_INT) tl = BVT_INT;
	switch (tl)
	{
	case BVT_OBJECT:
	case BVT_OBJECTREF:
	case BVT_STRONGOBJECT:
		return send(1, _T("OP_SHL"));
	case BVT_SBYTE:
		left = (BppSByte) ((BppSByte)left << (BppInt)right);
		break;
	case BVT_BYTE:
		left = (BppByte) ((BppByte)left << (BppInt)right);
		break;
	case BVT_SHORT:
		left = (BppShort) ((BppShort)left << (BppInt)right);
		break;
	case BVT_USHORT:
		left = (BppUShort) ((BppUShort)left << (BppInt)right);
		break;
	case BVT_INT:
		left = (BppInt)((BppInt)left << (BppInt)right);
		break;
	case BVT_UINT:
		left = (BppUInt) ((BppUInt)left << (BppInt)right);
		break;
	case BVT_LONG:
		left = BppLong(left) << BppInt(right);
		break;
	case BVT_ULONG:
		left = (BppULong) (BppULong(left) << BppInt(right));
		break;
	case BVT_INTPTR:
		left = (BppIntPtr) (BppIntPtr(left) << BppInt(right));
		break;
	default:
		return badType(tl, BVT_INT);
	}
	--_sp; // _sp points to result (position of left argument)
	return true;
}
bool BppInterpreter::doOP_SHR()
{
	BppValue& right = *_sp;
	BppValue& left = _stack[_sp.current() - 1];
	BppValueType tl = (BppValueType)left.v_type;
	BppValueType tr = (BppValueType)right.v_type;
	if ((tl != BVT_OBJECT) && (tl != BVT_OBJECTREF) && (tl != BVT_STRONGOBJECT))
	{
		if ((tr == BVT_OBJECT) || (tr == BVT_OBJECTREF) || tr == BVT_STRONGOBJECT)
		{
			BppValue tmp = right;
			right = left;
			left = tmp;
			return send(1, _T("OP_SHR_R"));
		}
		else if (!right.isIntegralType()) return badType(tr, BVT_INT);
	}
	else if (tl > BVT_NULL && tl < BVT_INT) tl = BVT_INT;
	switch (tl)
	{
	case BVT_OBJECT:
	case BVT_OBJECTREF:
	case BVT_STRONGOBJECT:
		return send(1, _T("OP_SHR"));
	case BVT_SBYTE:
		left = (BppSByte) ((BppSByte)left >> (BppInt)right);
		break;
	case BVT_BYTE:
		left = (BppByte) ((BppByte)left >> (BppInt)right);
		break;
	case BVT_SHORT:
		left = (BppShort) ((BppShort)left >> (BppInt)right);
		break;
	case BVT_USHORT:
		left = (BppUShort) ((BppUShort)left >> (BppInt)right);
		break;
	case BVT_INT:
		left = (BppInt) ((BppInt)left >> (BppInt)right);
		break;
	case BVT_UINT:
		left = (BppUInt) ((BppUInt)left >> (BppInt)right);
		break;
	case BVT_LONG:
		left = (BppLong) (BppLong(left) >> BppInt(right));
		break;
	case BVT_ULONG:
		left = (BppULong) (BppULong(left) >> BppInt(right));
		break;
	case BVT_INTPTR:
		left = (BppIntPtr) (BppIntPtr(left) >> BppInt(right));
		break;
	default:
		return badType(tl, BVT_INT);
	}
	--_sp; // _sp points to result (position of left argument)
	return true;
}


bool BppInterpreter::doOP_MKREF()
{
   BppValue& val = *_sp;
   BppValueType vt = (BppValueType)val.v_type;
   switch (vt)
   {
   //case BVT_CODE:
   //   val.v_type = BVT_CODEREF;
   //   break;
   case BVT_BYTECODE:
      val.v_type = BVT_BYTECODEREF;
      break;
   case BVT_BUFFER:
      val.v_type = BVT_BUFFERREF;
      break;
   case BVT_CHARBUFFER:
      val.v_type = BVT_CHARBUFFERREF;
      break;
   case BVT_CLASS:
      val.v_type = BVT_CLASSREF;
      break;
   case BVT_STRONGOBJECT:
   {
      BppObject* o = val;
      if (o/* && val._v.v_refObj->getRefCnt() == 1*/) o->addRef();
      val.dispose();
      val.v_type  = BVT_OBJECTREF;
      val._v.v_object = o;
   }
      break;
   case BVT_OBJECT:
      val.v_type = BVT_OBJECTREF;
      break;
   case BVT_VECTOR:
      val.v_type = BVT_VECTORREF;
      break;
   case BVT_STRING:
      val.v_type = BVT_STRINGREF;
      break;
   case BVT_DICTIONARY:
      val.v_type = BVT_DICTIONARYREF;
      break;
   default:
      return true; // ignore all other types
   }
   BppRefObject* robj = val.getRef();
   if (robj != 0)
   {
      long rcnt = robj->getRefCnt(); // decrement refcount without delete object
      if (rcnt > 0)
         robj->setRefCnt(rcnt-1);
   }
   val.v_exttype._flags &= ~BVF_STRONGTYPED;
   return true;
}

bool BppInterpreter::doOP_UNREF(BppClass* ownClass)
{
   static const BppString unreffunc = _T("OP_UNREF");
   BppValue& val = *_sp;
   BppValueType vt = (BppValueType)val.v_type;
   switch (vt)
   {
   //case BVT_CODEREF:
      //val.v_type = BVT_CODE;
      //break;
   case BVT_BYTECODEREF:
      val.v_type = BVT_BYTECODE;
      break;
   case BVT_BUFFERREF:
      val.v_type = BVT_BUFFER;
      break;
   case BVT_CHARBUFFERREF:
      val.v_type = BVT_CHARBUFFER;
      break;
   case BVT_CLASSREF:
      val.v_type = BVT_CLASS;
      break;
   case BVT_OBJECTREF:
      val.v_type = BVT_OBJECT;
      break;
   case BVT_VECTORREF:
      val.v_type = BVT_VECTOR;
      break;
   case BVT_STRINGREF:
      val.v_type = BVT_STRING;
      break;
   case BVT_DICTIONARYREF:
      val.v_type = BVT_DICTIONARY;
      break;
   case BVT_OBJECT:
   case BVT_STRONGOBJECT:
   {
      BppObject* obj = val;
      if (obj != 0)
      {
         BppDictionaryEntry* e = obj->getClass()->findMember(ownClass,unreffunc);
         if (e != 0)
         {
            if (e->getSymbolType() == ST_FUNCTION)
            {
               return send(0, unreffunc.c_str(), &e->getValue());
            }
         }
      }
   }
   default:
      return true; // ignore all other types
   }
   BppRefObject* robj = val.getRef();
   if (robj != 0 && vt >= _BVTREFMIN /*robj->getRefCnt() == 0*/)
   {
      robj->addRef(); // increment refcount
   }
   return true;
}


bool BppInterpreter::doOP_VREF()
{
   ValPtr vp = _sp;
   --vp; // points to target
   switch (maskWeakRef(BppValueType(vp->v_type)))
   {
   case BVT_DICTIONARY:
      {
         if (!checkType(*_sp,BVT_STRING)) return false;
         BppDictionary* pdict = vp->_v.v_dict;
         BppString* i = *_sp;
         *vp = (*pdict)[*i].getValue();
         --_sp;
      }
      break;
   case BVT_VECTOR:
      {
         if (!_sp->isIntegralType()) return checkType(*_sp,BVT_UINT);
         BppVector* pvec = vp->_v.v_vector;
         BppUInt i = *_sp;
         if (i >= pvec->size())
            return returnError(_T("subscript out of range"));
         *vp = (*pvec)[i];
         --_sp;
      }
      break;
   case BVT_STRING:
      {
		  if (!_sp->isIntegralType()) return checkType(*_sp, BVT_UINT);
         BppString* s = vp->_v.v_string;
         BppUInt i = *_sp;
         if (i >= s->length())
            return returnError(_T("subscript out of range"));
         *vp = BppCharVal((*s)[i]);
         --_sp;
      }
      break;
   case BVT_BUFFER:
      {
		  if (!_sp->isIntegralType()) return checkType(*_sp, BVT_UINT);
		  BppBuffer* b = vp->_v.v_buffer;
         BppUInt i = *_sp;
         if (i >= b->size())
            return returnError(_T("subscript out of range"));
         *vp = (BppByte) (*b)[i];
         --_sp;
      }
      break;
   case BVT_CHARBUFFER:
      {
		  if (!_sp->isIntegralType()) return checkType(*_sp, BVT_UINT);
		  BppCharBuffer* cb = vp->_v.v_charbuffer;
         BppUInt i = *_sp;
         if (i >= cb->size())
            return returnError(_T("subscript out of range"));
		  *vp = BppCharVal((*cb)[i]);
         --_sp;
      }
      break;
   default:
      if (!checkType(*vp,BVT_OBJECT)) return false;
      if (!send(1,_T("OP_VREF"))) return false;
      break;
   }
   return true;
}

bool BppInterpreter::doOP_VSET()
{
   ValPtr vp = _sp;
   vp -= 2; // points to target
   BppValue& idxval = _sp[-1];
   switch (maskWeakRef(BppValueType(vp->v_type)))
   {
   case BVT_DICTIONARY:
      {
         if (!checkType(idxval,BVT_STRING)) return false;
         BppDictionary* pdict = vp->_v.v_dict;
         BppString* i = idxval;
         (*pdict)[*i].setValue(*_sp);
         *vp = *_sp;
         _sp = vp;
      }
      break;
   case BVT_VECTOR:
      {
		  if (!idxval.isIntegralType()) return checkType(idxval, BVT_UINT);
		  BppVector* pvec = vp->_v.v_vector;
         BppUInt i = idxval;
         if (i >= pvec->size())
            return returnError(_T("subscript out of range"));
         (*pvec)[i] = *_sp;
         *vp = *_sp;
         _sp = vp;
      }
      break;
   case BVT_STRING:
      {
		  if (!idxval.isIntegralType()) return checkType(idxval, BVT_UINT);
		  BppString* s = vp->_v.v_string;
         BppUInt i = idxval;
         if (i >= s->length())
            return returnError(_T("subscript out of range"));
         BppChar c = (BppChar) (BppInt) *_sp;
         (*s)[i] = c;
		 *vp = BppCharVal(c);
         _sp = vp;
      }
      break;
   case BVT_BUFFER:
      {
		  if (!idxval.isIntegralType()) return checkType(idxval, BVT_UINT);
         BppBuffer* b = vp->_v.v_buffer;
         BppUInt i = idxval;
         if ( i >= b->size())
            return returnError(_T("subscript out of range"));
         BppByte c =  *_sp;
         (*b)[i] = c;
         *vp =  c;
         _sp = vp;
      }
      break;
   case BVT_CHARBUFFER:
      {
		  if (!idxval.isIntegralType()) return checkType(idxval, BVT_UINT);
         BppCharBuffer* cb = vp->_v.v_charbuffer;
         BppUInt i = idxval;
         if (i >= cb->size())
            return returnError(_T("subscript out of range"));
         BppChar c = (BppChar) (BppInt) *_sp;
         (*cb)[i] = c;
         *vp = BppCharVal(c);
         _sp = vp;
      }
      break;
   default:
      if (!checkType(*vp,BVT_OBJECT)) return false;
      if (!send(2,_T("OP_VSET"))) return false;
      break;
   }
   return true;
}

bool BppInterpreter::doOP_PREF()
{
   ValPtr vp = _sp;
   --vp; // points to target
    if (!checkType(*_sp,BVT_STRING)) return false;
   switch (vp->v_type)
   {
   case BVT_DICTIONARY:
   case BVT_DICTIONARYREF:
      {
         BppDictionary* pdict = vp->_v.v_dict;
         BppString* i = *_sp;
         *vp = (*pdict)[*i].getValue();
         --_sp;
      }
      break;
   default:
      if (!checkType(*vp,BVT_OBJECT)) return false;
      {
         BppObject* po = *vp;
         BppString* key = *_sp;
         if (!po || !key) return false;
         BppClass* pclass = po->getClass();
         BppClass* refClass = _procStack ? _procStack->getClass() : 0;
         BppDictionaryEntry* e = pclass->findMember(refClass, *key);
         if (e && e->getSymbolType() == ST_DATA) {
            BppVector* v = e->getValue();
            if (v) {
               BppInt idx = (*v)[0];
               if (idx >= 0)
               {
                  *vp = (*po->getMemberData())[idx];
                  --_sp;
                  return true;
               }
            }
            return false;
         }
      }
      if (!send(1,_T("OP_PREF"))) return false;
      break;
   }
   return true;
}

bool BppInterpreter::doOP_PSET()
{
   ValPtr vp = _sp;
   vp -= 2; // points to target
   BppValue& idxval = _sp[-1];
   if (!checkType(idxval,BVT_STRING)) return false;
   switch (vp->v_type)
   {
   case BVT_DICTIONARY:
   case BVT_DICTIONARYREF:
      {
         BppDictionary* pdict = vp->_v.v_dict;
         BppString* i = idxval;
         (*pdict)[*i].setValue(*_sp);
         *vp = *_sp;
         _sp = vp;
      }
      break;
   default:
      if (!checkType(*vp,BVT_OBJECT)) return false;
      {
         BppObject* po = *vp;
         BppString* key = idxval;
         if (!po || !key) return false;
         BppClass* pclass = po->getClass();
         BppClass* refClass = _procStack ? _procStack->getClass() : 0;
         BppDictionaryEntry* e = pclass->findMember(refClass, *key);
         if (e && e->getSymbolType() == ST_DATA) {
            BppVector* v = e->getValue();
            if (v) {
               BppInt idx = (*v)[0];
               if (idx >= 0)
               {
                  bool ok = (*po->getMemberData())[idx].assign(*_sp);
                  *vp = *_sp;
                  _sp = vp;
                  return ok;
               }
            }
            return false;
         }
      }
      if (!send(2,_T("OP_PSET"))) return false;
      break;
   }
   return true;
}

static const BppString equFunc = _T("equals");

bool BppInterpreter::doEquals(BppClass* ownClass)
{
   BppValue& right = *_sp;
   BppValue& left = _stack[_sp.current()-1];
   BppValueType tl = (BppValueType)left.v_type;
   BppValueType tl0 = tl;
   BppValueType tr = (BppValueType)right.v_type;
   BppValueType tr0 = tr;
   if ((tr == tl || (tl >_BVTVALMAX && tr > _BVTVALMAX)) && right._v.v_addr == left._v.v_addr)
   {
      *(--_sp) = (BppInt)1;
      return true;
   }
   bool haveobject = false;
   BppObject* obj = left;
   if (obj != 0)
   {
      haveobject = true;
      BppDictionaryEntry* e = obj->getClass()->findMember(ownClass,equFunc);
      if (e !=  0)
      {
         if (e->getSymbolType() == ST_FUNCTION)
         {
			 if (!send(1,equFunc.c_str(),&e->getValue())) return false;
            if (!checkType(*_sp,BVT_INT)) return false;
            return true;
         }
      }
   }
   else if (tl == BVT_STRONGOBJECT) tl = BVT_NULL;
   obj = right;
   if (obj != 0)
   {
      haveobject = true;
      BppDictionaryEntry* e = obj->getClass()->findMember(ownClass,equFunc);
      if (e !=  0)
      {
         if (e->getSymbolType() == ST_FUNCTION)
         {
            BppValue tmp = right;
            right = left;
            left = tmp;
			if (!send(1,equFunc.c_str(),&e->getValue())) return false;
            if (!checkType(*_sp,BVT_INT)) return false;
            left = *_sp;
            return true;
         }
      }
   }
   else if (tr == BVT_STRONGOBJECT) tr = BVT_NULL;
   if (tl == BVT_NULL)
   {
	   if (right._v.v_addr == 0 && tl0 > _BVTVALMAX)
	   {
		   *(--_sp) = (BppInt) 1;
		   return true;
	   }
   }
   else if (tr == BVT_NULL)
   {
	   if (left._v.v_addr == 0 && tr0 > _BVTVALMAX)
	   {
		   *(--_sp) = (BppInt) 1;
		   return true;
	   }
   }
   if (tl == tr && tl > _BVTVALMAX && (!obj))
   {
	   if (right == left)
          *(--_sp) = (BppInt) 1;
	   else
          *(--_sp) = (BppInt) 0;
	   return true;
   }
   if (haveobject) {
      if (!doCompare(ownClass)) return false;
      if (((BppInt)*_sp) == 0)
         *_sp = (BppInt)1;
      else
         *_sp = (BppInt)0;
   }
   else *_sp = BppInt(left.compare(right) ? 0 : 1);
   return true;
}

static const BppString cmpFunc = _T("compare");

bool BppInterpreter::doCompare(BppClass* ownClass)
{
  BppValue& right = *_sp;
   BppValue& left = _stack[_sp.current()-1];
   BppValueType tl = (BppValueType)left.v_type;
   BppValueType tr = (BppValueType)right.v_type;
   BppObject* obj = left;
   BppValueType rt = tl < tr ? tr : tl;
   if (obj != 0)
   {
      BppDictionaryEntry* e = obj->getClass()->findMember(ownClass,cmpFunc);
      if (e !=  0)
      {
         if (e->getSymbolType() == ST_FUNCTION)
         {
			 if (!send(1,cmpFunc.c_str(),&e->getValue())) return false;
            if (!checkType(*_sp,BVT_INT)) return false;
            return true;
         }
      }
   }
   else if (tl == BVT_STRONGOBJECT) tl = BVT_NULL;
   BppObject* robj = right;
   if (robj != 0)
   {
      BppDictionaryEntry* e = robj->getClass()->findMember(ownClass,cmpFunc);
      if (e !=  0)
      {
         if (e->getSymbolType() == ST_FUNCTION)
         {
            BppValue tmp = right;
            right = left;
            left = tmp;
			if (!send(1,cmpFunc.c_str(),&e->getValue())) return false;
            if (!checkType(*_sp,BVT_INT)) return false;
            left = *_sp;
            left._v.v_integer = -left._v.v_integer;
            return true;
         }
      }
   }
   else if (tr == BVT_STRONGOBJECT) tr = BVT_NULL;
   if (tl == BVT_NULL || tr == BVT_NULL) rt = BVT_NULL;
   else if (rt < BVT_INT) rt = BVT_INT;
   switch (rt)
   {
   case BVT_NULL:
	   left = (BppInt) (tr < tl ? 1 : (tl < tr ? -1 : 0));
	   break;
   case BVT_INT:
	   {
		   BppInt til = left;
		   BppInt tir = right;
		   left = (BppInt) (til > tir ? 1 : (til<tir ? -1 : 0));
	   }
	   break;
   case BVT_UINT:
	   {
		   BppUInt tul = left;
		   BppUInt tur = right;
		   left = (BppInt) (tul > tur ? 1 : (tul<tur ? -1 : 0));
	   }
	   break;
   case BVT_LONG:
	   {
		   BppLong tll = left;
		   BppLong tlr = right;
		   left = (BppInt) (tll > tlr ? 1 : (tll<tlr ? -1 : 0));
	   }
	   break;
   case BVT_ULONG:
	   {
		   BppULong tull = left;
		   BppULong tulr = right;
		   left = (BppInt) (tull > tulr ? 1 : (tull<tulr ? -1 : 0));
	   }
	   break;
   case BVT_INTPTR:
	   {
		   BppIntPtr tpl = left;
		   BppIntPtr tpr = right;
		   left = (BppInt) (tpl > tpr ? 1 : (tpl<tpr ? -1 : 0));
	   }
	   break;
   case BVT_FLOAT:
   case BVT_DOUBLE:
	   {
		   double tfl = left;
		   double tfr = right;
		   left = (BppInt) (tfl > tfr ? 1 : (tfl<tfr ? -1 : 0));
	   }
	   break;
   default:
	   switch (tl)
	   {
	   case BVT_STRING:
	   case BVT_STRINGREF:
		   {
			   if (tr != BVT_STRING && tr != BVT_STRINGREF) return badType(tr,BVT_STRING);
			   BppString* sl = left;
			   BppString* sr = right;
			   if (sl && sr) left = (BppInt) sl->compare(*sr);
			   else left = (BppInt) (sl ? 1 : (sr ? -1 : 0));
		   }
		   break;
	   case BVT_BUFFER:
	   case BVT_BUFFERREF:
		   {
			   if (tr != BVT_BUFFER && tr != BVT_BUFFERREF) return badType(tr,BVT_BUFFER);
			   BppBuffer* bl = left;
			   BppBuffer* br = right;
			   if (bl && br) left = (BppInt) bl->compare(*br);
			   else left = (BppInt) (bl ? 1 : (br ? -1 : 0));
		   }
		   break;
	   case BVT_CHARBUFFER:
	   case BVT_CHARBUFFERREF:
		   {
			   if (tr != BVT_CHARBUFFER && tr != BVT_CHARBUFFERREF) return badType(tr,BVT_CHARBUFFER);
			   BppCharBuffer* cbl = left;
			   BppCharBuffer* cbr = right;
			   if (cbl && cbr) left = (BppInt) cbl->compare(*cbr);
			   else left = (BppInt) (cbl ? 1 : (cbr ? -1 : 0));
		   }
		   break;
      case BVT_VECTOR:
      case BVT_VECTORREF:
      {
         if (tr != BVT_VECTOR && tr != BVT_VECTORREF) return badType(tr, tl);
         BppVector* vl = left;
         BppVector* vr = right;
         if (vl == vr || (vl && vr && *vl == *vr)) left = (BppInt)0;
         else {
            left = (BppInt) (vl > vr ? 1 :-1);
         }
      }
      break;
      case BVT_DICTIONARY:
      case BVT_DICTIONARYREF:
      {
         if (tr != BVT_DICTIONARY && tr != BVT_DICTIONARYREF) return badType(tr, tl);
         BppDictionary* dl = left;
         BppDictionary* dr = right;
         left = (BppInt) (dl > dr ? 1 : dl < dr ? -1 : 0);
      }
      break;
      default:
         if (obj && robj) {
            left = (BppInt) (obj > robj ? 1 : obj < robj ? -1: 0);
            break;
         }
         
		   return badType(BVT_INT,tl);
	   }
   }
   --_sp;
   return true;
}
// helper

/* send - send a message to an object */
// nargs Argumente liegen auf dem Stack, fehlt sel, ist 1. Argument der Name
// Annahme: Objekt-Wert liegt auf Position (_sp.current() - nargs), also unterhalb der Argumente
/*
   argn
   argn-1
   ..
   arg1 (name)
   object
*/
bool BppInterpreter::send(int nargs, const BppChar* sel, const BppValue* funcVal)
{
   ValPtr vp = _sp;
   vp -= nargs; // vp points to object value
  // if (!checkType(*vp,BVT_OBJECT)) return false;
   BppObject* pobj = *vp;
   if (pobj == 0)
      return returnError(_T("BppInterpreter::send : object expected\n"));
   BppClass* pclass = pobj->getClass();
   if (!pclass) return returnError(_T("BppInterpreter::send : no class for object\n"));
   BppClass* contextClass = _procStack ? _procStack->getClass() : 0;
   BppString selector = sel;
   BppValue codeVal;
   if (sel == 0) // first arg is name or code
    {
       ValPtr vp2 = vp;
       ++vp2; // points to name or coderef
       BppValueType vt = vp2->getType();
       if ((vt == BVT_BYTECODE) || (vt == BVT_BYTECODEREF))
       {
          codeVal = *vp2;
          selector = (const BppString&) *(BppString*) ((*(BppVector*) codeVal)[IDX_NAME]);
       }
	   else if ((vt == BVT_CODE)/* || (vt == BVT_CODEREF)*/)
	   {
		   codeVal = *vp2;
		   selector =_T("native");
	   }
       else
       {
          if (!checkType(*vp2,BVT_STRING))
             return false;
          selector = (const BppString&) (*(BppString*) *vp2);
       }
       *vp2 = *vp; // object replaces name
	}
	else // name explicitly specified
	{
	   ++_sp; // add entry to stack
      unsigned k= (unsigned) _sp.current();
	   for(int i=0;i<=nargs;i++,k--) // copy all args + object one entry up
         _stack[k] = _stack[k-1];
	   ++nargs;
   }
    do
    {
		if (funcVal)
			codeVal = *funcVal;
       if (codeVal.v_type == BVT_NULL)
       {
          BppDictionaryEntry* de = pclass->findMember(contextClass,selector,true);//pclass->funcMembers().find(selector);
          if (de == 0)
          {
             const BppChar* psel = selector.c_str();
             if ((_tcsncmp(psel,_T("BC_"),3) == 0) || (_tcsncmp(psel,_T("DC_"),3)== 0))
             {
                BppChar c0 = psel[0];
                psel += 3;
                selector = BppString(psel);
                if (contextClass)
                {
                   pclass = contextClass;
                      if (c0 == _T('D'))
                      {
                         de = contextClass->funcMembers().find(selector);
                      }
                }
             }
             else break;
             if (de == 0) continue;
          }
          codeVal = de->getValue();
       }
       switch (codeVal.v_type)
       {
       case BVT_CODE:
      //case BVT_CODEREF:
          {
             BPPUSERFUNC f = codeVal;
             if (f == 0)
                return returnError(selector.c_str(),_T(" : NULL pointer to native function\n"));
             // set entry before object to method name
             // i.e. bp[0] in native method implementation contains method name
             (*vp) = new BppString(BPPMOVE(selector));
             bool result =  f(&_vm,nargs,_sp);
             *vp = *_sp;
             // Argumente freigeben 
             int n = vp.current() +1;
             while (_stack.size() > n) {
                // prevent from object deletion while stack is modified
                BppValue v = *_stack.last();
                _stack.pop_back();
             }
             _sp = vp;
             return result;
          }
       case BVT_BYTECODE:
       case BVT_BYTECODEREF:
          {
             // check(3);
             BppVector* code = codeVal;
             if (code == 0)
                return returnError(selector.c_str(),_T(" : NULL pointer to bytecode\n"));
             bool result =  call(nargs,code);
             *vp = *_sp;
             int n = vp.current() + 1;
             while (_stack.size() > n) {
                // prevent from object deletion while stack is modified
                BppValue v = *_stack.last();
                _stack.pop_back();
             }
             _sp = vp;
             return result;
          }
          break;
       default:
          return returnError(selector.c_str(), _T("has no method implementation\n"));
       }
    }
    while ((pclass = pclass->getBaseClass()) != NULL);
    const BppChar* s = selector.c_str();
    if ((_tcscmp(s,_T("dtor")) == 0) && (_tcscmp(s,_T("BC_dtor")) == 0))
       return true;
    return noMethod(BppString::format(_T("%s::%s"),pobj->getClass()->getClassName()->c_str(),s).c_str());
}

#define BVTRANGE(T1,T2) if (type >= T1 && type <= T2) return true; break
bool BppInterpreter::checkType(const BppValue& val, BppValueType vtype)
{
   BppValueType type = val.getType();
   if (type == vtype || vtype == maskWeakRef(type))
      return true;
   //if ((vtype < _BVTREFMIN) && (vtype == maskWeakRef(type))) return true;/*&& (type >= _BVTREFMIN)*/ // base type of reference type is allowed too
   switch (vtype)
   {
   case BVT_OBJECT:
      if (type == BVT_STRONGOBJECT) return true;
      break;
   case BVT_NUMBER: BVTRANGE(BVT_INTEGRAL_MIN,BVT_REAL_MAX);
   case BVT_INTNUMBER: BVTRANGE(BVT_SBYTE,BVT_ULONG);
   case BVT_UINTNUMBER:
	   switch (type)
	   {
	   case BVT_BYTE:  
	   case BVT_USHORT:
	   case BVT_UINT:
	   case BVT_ULONG:
		   return true;
	   }
	   break;
   case BVT_INTEGRALTYPE: BVTRANGE(BVT_INTEGRAL_MIN,BVT_INTEGRAL_MAX);
   case BVT_PTRTYPE:
	   if (type == BVT_NULL) return true;
#ifndef BPP_I64
       BVTRANGE(BVT_INT,BVT_ULONG);
#else
	   BVTRANGE(BVT_INTPTR,BVT_ULONG);
#endif
   case BVT_REALNUMBER: BVTRANGE(BVT_REAL_MIN,BVT_REAL_MAX);
   case BVT_REFCNTTYPE: BVTRANGE(BVT_RCNT_MIN,BVT_RCNT_MAX);
   case BVT_REFERENCE: BVTRANGE(_BVTREFMIN,_BVTREFMAX);
   }
   return badType(type,vtype);
}
#undef BVTRANGE

bool BppInterpreter::badType(BppValueType type, BppValueType expected)
{
	const BppChar* t = _vm.getTypeNameFromVType(type);
	if ((!t) || type >= _BVTREFMIN) t = getValueTypeName(type);
	const BppChar* e = _vm.getTypeNameFromVType(expected);
	if ((!e) || expected >= _BVTREFMIN) e = getValueTypeName(expected);

	BppString buf = BppString::format(_T("invalid value type %s, %s expected"),t,e);
   return returnError(buf.c_str());
}

bool BppInterpreter::badType(const BppValue& val, const BppValue& expected)
{
	BppValueType vt = val.getType();
    BppValueType et = expected.getType();
	BppVector* desc = expected.getStrongObjDesc();
	if (!desc) return badType(vt,et);
	BppClass* pcexp = (*desc)[IDX_CLASS];
	if (!pcexp) return badType(vt,et);
	const BppChar* t = _vm.getTypeNameFromVType(vt);
	if ((!t) || vt >= _BVTREFMIN) t = getValueTypeName(vt);
	BppObject* vobj = val;
	if (vobj)
	{
		    BppString buf = BppString::format(_T("invalid %s type %s, object of class %s expected"),
			t,vobj->getClass()->getClassName()->c_str(),
			pcexp->getClassName()->c_str());
	    returnError(buf.c_str());
	}
	else
	{
		BppString buf = BppString::format(_T("invalid value type %s, object of class %s expected"),
			t, pcexp->getClassName()->c_str());
		returnError(buf.c_str());
	}
	return false;
}

bool BppInterpreter::noMethod(const BppChar* name)
{
       return returnError(_T("no method for name "),name);

}

bool BppInterpreter::returnError(const BppChar* msg, const BppChar* context)
{
   BppString* s = 0;
   if (context && (*context))
      s = new BppString(BppString::format(_T("%s: %s"), context, msg));
   else s= new BppString(msg);   
    if (_vm.getCurrentLine() > 0)
	  (*s) += BppString::format(_T("\nin \"%s\" line: %d"), _vm.getCurrentSource().c_str(), _vm.getCurrentLine());
   return returnError(BppValue(s));
}

bool BppInterpreter::returnError( const BppValue& errValue )
{
   if (_procStack != 0)
   {
	   BppValue& currerr = _procStack->fp()[1];
	   if (currerr.getType() != BVT_NULL) _errStack.push_back(currerr);
      currerr = errValue;
   }
   BppValue tmp;
   if (doStrCast(tmp, errValue)) {
      BppString* s = tmp;
      _errBuffer = *s;
   }
   else
   {
      _errBuffer = BppString::format(_T("%s:%p"), getValueTypeName(errValue.getType()), errValue.ptr());
   }
   _errBuffer += _T('\n');
   return false;
}


/**
Function Function retrieves a string pointer from specified value if possible.
@param val Value to get string from
@return string pointer on success, otherwise NULL
@note <p>This function is for use from inside native function implementation only!</p>
<p>If <i>val</i> is of string type, it simply returns  pointer to contained string.
If <i>val</i> is of object type it checks if object has a "toString" method. If so
that method is called.</p>
<p>Implementation of native Function should store returned string into a value object immediately
to prevent it from being released.</p>
*/
BppString* BppInterpreter::getStringFromValue(BppClass* ownClass, BppValue& val )
{
	static const BppString _string = _T("string");
	static const BppString _tostring = _T("toString");
   BppString* s = val;
   if (s == 0)
   {
      BppObject*pobj = val;
      if (pobj != 0)
      {
         BppClass* pclass = pobj->getClass();
		 BppDictionaryEntry* pe = pclass->findMember(ownClass,_string);
         if (!pe) pe = pclass->findMember(ownClass,_tostring);
         if ((pe != 0) && (pe->getSymbolType() == ST_FUNCTION))
         {
            ValPtr savedSp = _sp;
            _sp = _stack.last();
			*(++_sp) = val;
			*(++_sp) = pe->getValue();
            if (send(1,0))
            {
               s = *_sp;
            }
            _sp = savedSp;
         }
      } // if pobj
   }
   return s;
}


bool BppInterpreter::doNumCast(BppValue & dest, const BppValue & src, BppValueType dstType, const BppChar * fname)
{
   // convert from string
   BppString* s = src;
   if (s != 0)
   {
      int base = 0;
      const BppChar* ps = s->c_str();
      if (dstType < BVT_FLOAT && s->length() > 2) {
         if (ps[0] == _T('0') && (ps[1] == _T('B') || ps[1] == _T('b'))) {
            base = 2;
            ps += 2;
         }
      }

      dest.dispose();
      C_LOCALE;
      BppChar* rptr = 0;
      switch (dstType)
      {
      case BVT_SBYTE:
      case BVT_SHORT:
      case BVT_INT:
#if SIZE_MAX <= UINT_MAX
      case BVT_INTPTR:
#endif
         dest._v.v_integer = (BppInt)_tcstol(ps, &rptr, base);
         break;
      case BVT_BYTE:
      case BVT_USHORT:
      case BVT_UINT:
          dest._v.v_uint = (BppUInt)_tcstoul(ps, &rptr, base);
         break;
      case BVT_LONG:
#if SIZE_MAX > UINT_MAX
      case BVT_INTPTR:
#endif
            dest._v.v_long = (BppLong)_tcstoi64(ps, &rptr, base);
         break;
      case BVT_ULONG:
          dest._v.v_ulong = (BppULong)_tcstoui64(ps, &rptr, base);
         break;
      case BVT_FLOAT:
         dest._v.v_float = (float)_tcstod(s->c_str(), &rptr);
         break;
      case BVT_DOUBLE:
         dest._v.v_double = _tcstod(s->c_str(), &rptr);
         break;
      default:
         return returnError(_T("invalid argument type"), fname);
      }
      if (rptr && *rptr != 0)
         return returnError(BppString::format(_T("unable to convert \"%s\" into number of type %s."),
            s->c_str(), getValueTypeName(dstType)).c_str(), fname);
      dest.v_type = dstType;
      return true;
   }
   BppValue srcVal = src;
   // check buffer, convert buffer to number
   BppBuffer* b = srcVal;
   if (b)
   {
      size_t sz = b->size();
      switch (sz)
      {
      case sizeof(BppSByte) :
         srcVal = *(BppSByte*)b->data();
         break;
      case sizeof(BppShort) :
         srcVal = *(BppShort*)b->data();
         break;
      case sizeof(BppInt) :
         srcVal = *(BppInt*)b->data();
         break;
      case sizeof(BppLong) :
         srcVal = (*(BppLong*)b->data())
#ifdef EMULATE_I64
         .toLong()
#endif
         ;
         break;
      default:
         return returnError(_T("invalid buffer size"), fname);
      }
   }
   // convert from number
   BppValueType srctype = srcVal.getType();
   if (srctype >= BVT_INTEGRAL_MIN && srctype <= BVT_REAL_MAX)
   {
      switch (dstType)
      {
      case BVT_SBYTE:
         dest = (BppSByte)srcVal;
         break;
      case BVT_BYTE:
         dest = (BppByte)srcVal;
         break;
      case BVT_SHORT:
         dest = (BppShort)srcVal;
         break;
      case BVT_USHORT:
         dest = (BppUShort)srcVal;
         break;
      case BVT_INT:
         dest = (BppInt)srcVal;
         break;
      case BVT_UINT:
         dest = (BppUInt)srcVal;
         break;
      case BVT_LONG:
         dest = (BppLong)srcVal;
         break;
      case BVT_ULONG:
         dest = (BppULong)srcVal;
         break;
      case BVT_INTPTR:
         dest = (BppIntPtr)srcVal;
         break;
      case BVT_FLOAT:
         dest = (float)srcVal;
         break;
      case BVT_DOUBLE:
         dest = (double)srcVal;
         break;
      default:
         return returnError(BppString::format(_T("unable to convert type %s into number of type %s."),
            getValueTypeName(srctype), getValueTypeName(dstType)).c_str(),
            fname);
      }
      dest.v_exttype._type = dstType;

      return true;
   }
   // convert from object
   BppObject* pobj = srcVal;
   if (pobj != 0)
   {
      BppClass* pclass = pobj->getClass();
      BppClass* ownClass = _procStack ? _procStack->getClass() : 0;
      BppDictionaryEntry* pe = pclass->findMember(ownClass,fname);
      if ((pe != 0) && (pe->getSymbolType() == ST_FUNCTION))
      {
         ValPtr sp = _sp;
         *(++_sp) = pe->getValue();
         if (send(1, 0))
         {
            dest = *sp;
         }
         _sp = sp;
      }
      else return returnError(BppString::format(_T("conversion operator %s not defined."),fname).c_str(), 
         pclass->getClassName()->c_str());
   } // if pobj
   else return returnError(BppString::format(_T("invalid argument of type %s."),
      getValueTypeName(srctype)).c_str(), fname);
   return true;
}

bool BppInterpreter::doStrCast(BppValue & dest, const BppValue & src)
{
   BppValueType vt = maskWeakRef(src.getType());
   BppChar buf[65];
   C_LOCALE;
   switch (vt)
   {
   case BVT_INT:
   case BVT_SHORT:
   case BVT_SBYTE:
#if (INT_MAX == SHRT_MAX)
      _stprintf(buf, _T("%ld"), (BppInt)src);
#else
      _stprintf(buf, _T("%d"), (BppInt)src);
#endif
      break;
   case BVT_UINT:
   case BVT_USHORT:
   case BVT_BYTE:
#if (INT_MAX == SHRT_MAX)
      _stprintf(buf, _T("%lu"), (BppUInt)src);
#else
      _stprintf(buf, _T("%u"), (BppUInt)src);
#endif
      break;
   case BVT_INTPTR:
      _stprintf(buf, _T("%p"), (void*)(BppIntPtr)src);
      break;
   case BVT_LONG:
#ifdef EMULATE_I64
   {
      ((BppLong)src).toa(buf);
   }
#else
      _stprintf(buf, _T("%lld"), (BppLong)src);
#endif
      break;
   case BVT_ULONG:
#ifdef EMULATE_I64
   {
      ((BppULong)src).toa(buf);
   }
#else
      _stprintf(buf, _T("%llu"), (BppULong)src);
#endif
      break;
   case BVT_FLOAT:
   case BVT_DOUBLE:
      _stprintf(buf, _T("%g"), (double)src);
      break;
   case BVT_STRING:
   {
      BppString* s = src;
      if (s)
      {
         dest = new BppString(*s);
         break;
      }
   }
   case BVT_CHARBUFFER:
   {
      BppCharBuffer* cb = src;
      if (cb != 0)
      {
         BppString* s = new BppString(*cb);
         dest = s;
         break;
      }
   }
   case BVT_OBJECT:
   case BVT_STRONGOBJECT:
   {
      BppObject*pobj = src;
      if (pobj != 0)
      {
         BppClass* pclass = pobj->getClass();
         BppClass* ownClass = _procStack ? _procStack->getClass() : 0;
         BppDictionaryEntry* pe = pclass->findMember(ownClass,_T("string"));
         if ((pe == 0) || (pe->getSymbolType() != ST_FUNCTION))
            pe = pclass->findMember(ownClass,_T("toString"));
         if ((pe != 0) && (pe->getSymbolType() == ST_FUNCTION))
         {
            ValPtr sp = _sp;
            *(++_sp) = pe->getValue();
            if (send(1, 0))
            {
               dest = *_sp;
               vt = _sp->getType();
               _sp = sp;
               if (vt == BVT_STRING || vt == BVT_STRINGREF)
                  break;
            }
            _sp = sp;
         }
      } // if pobj
   }
   default:
      return returnError(_T("invalid argument type"));
      break;
   }
   if (vt <= _BVTVALMAX) dest = new BppString((const BppChar*)buf);

   return true;
}

bool BppInterpreter::invokeFunc(BppValue & result, const BppValue & funcval, BppVector * args)
{
   int nargs = 0;
   if (args != 0)
      nargs = (int)args->size();
   BppValueType ftype = funcval.getType();
   switch (ftype) {
   case BVT_CODE:
   case BVT_BYTECODE:
   case BVT_BYTECODEREF:
      break;
   default:
      {
         BppValue strVal;
         if (!doStrCast(strVal, funcval)) return false;
         BppString* s = strVal;
         if (!s) return badType(ftype, BVT_BYTECODE);
         return invoke(result, s->c_str(), args);
      }
   }
   //unsigned savedSize = _stack.size();
   ValPtr savedSp = _sp;
   ++_sp;
   BppValue fval = funcval; // make additional reference
   for (int i = 0; i < nargs; ++i) // push arguments to stack
      *(++_sp) = (*args)[i];
      //_stack.push_back((*args)[i]);
   //_sp = _stack.last(); // _sp points to TOS
   bool ok = false;
   /* dispatch on its data type */
   if (ftype == BVT_CODE)
   {
      BPPUSERFUNC f = funcval;
      ok = f(&_vm, nargs, _sp);
   }
   else {
      ok = call(nargs, funcval);
   }
   if (ok) result = *_sp;
   _sp = savedSp;
   //unsigned s = _stack.size();
   //while (s > savedSize) { _stack.pop_back(); --s; }
   return ok;

}

/**
@param result Buffer to put result value into.
@param name Name of function to invoke
@param args Vector of arguments.
@return true on success
@note Function does not change value of stack pointer nor value where stack pointer points to.
If result value is needed to be on stack, caller must push it manually.
*/
bool BppInterpreter::invoke(BppValue& result, const BppChar* name, BppVector* args)
{
   if (name == 0)
   {
      return returnError(_T("Missing function name "),T_FUNCTION);
   }
   BppDictionaryEntry* pe = _vm.symbols().find(name);
   if (pe == 0)
      return returnError(BppString::format(_T("Symbol '%s' not found. "),name).c_str(),T_FUNCTION);
   const BppValue& val = pe->getValue();
    /* dispatch on its data type */
    switch (val.getType())
    {
    case BVT_CODE:
    case BVT_BYTECODE:
    case BVT_BYTECODEREF:
       return invokeFunc(result, val, args);
    default:
       badType(val.getType(),BVT_BYTECODE);
    }
    return false;
}

/**
@param result Buffer to put result value into.
@param pobj Pointer to object where method is to be invoked.
@param name Name of method to invoke.
@param args Vector uf arguments.
@return true on success
@note Function does not change value of stack pointer nor value where stack pointer points to.
If result value is needed to be on stack, caller must push it manually.
*/
bool BppInterpreter::invokeMember(BppValue& result, BppObject* pobj, const BppChar* name, BppVector* args)
{
   if (pobj == 0)
       return returnError(_T("Missing function object reference "),T_FUNCTION);
   int nargs = 0;
   if (args != 0)
      nargs = (int) args->size();
//   unsigned savedSize = _stack.size();
   ValPtr savedSp = _sp;
   ++_sp;
   *_sp = pobj;
   for (int i=0;i<nargs;++i) // push arguments to stack
      *(++_sp) = ((*args)[i]);
   //_sp = _stack.last(); // _sp points to TOS
   //++_sp;
   //*_sp = pobj;
   bool ok = send(nargs,name);
   if (ok)
   {
      result = *_sp;
   }
   _sp = savedSp;
   //unsigned s = _stack.size();
   //while (s > savedSize) { _stack.pop_back(); --s; }
   return ok;
}
/**
@param result Buffer to put result value into.
@param pclass Pointer to class where function is to be invoked.
@param name Name of function to invoke.
@param args Vector uf arguments.
@return true on success
@note Function does not change value of stack pointer nor value where stack pointer points to.
If result value is needed to be on stack, caller must push it manually.
*/
bool BppInterpreter::invokeStatic(BppValue& result, BppClass* pclass, const BppChar* name, BppVector* args)
{
   if (pclass == 0)
       return returnError(_T("Missing class reference "),T_FUNCTION);
   BppClass* ownClass = _procStack ? _procStack->getClass() : 0;
   BppDictionaryEntry* pe = pclass->findMember(ownClass,name);
   if (pe == 0)
      return returnError(BppString::format(_T("Symbol '%s' not found in class '%s'. "),name,pclass->getClassName()->c_str()).c_str(),T_FUNCTION);
   BppSymbolType st = pe->getSymbolType();
   if ((st != ST_SFUNCTION) && (st != ST_SDATA))
      return returnError(BppString::format(_T("Symbol '%s' not a static member of class '%s'. "),name,pclass->getClassName()->c_str()).c_str(),T_FUNCTION);
   const BppValue& val = pe->getValue();
    /* dispatch on its data type */
   switch (val.getType())
   {
   case BVT_CODE:
   case BVT_BYTECODE:
   case BVT_BYTECODEREF:
      return invokeFunc(result, val, args);
   default:
      badType(val.getType(), BVT_BYTECODE);
   }
   return false;
}

/**
@param result Buffer to put result value into.
@param className Name of class to get instance of.
@param args Vector uf arguments to be passed to constructor.
@return true on success
@note Function does not change value of stack pointer nor value where stack pointer points to.
If result value is needed to be on stack, caller must push it manually.
*/
bool BppInterpreter::createObject(BppValue& result, const BppString& className, BppVector* args)
{
   if (className.length() == 0)
   {
      return returnError(_T("Missing class name. "),T_FUNCTION);
   }
   BppDictionaryEntry* pe = _vm.classes().find(className);
   if (pe == 0)
      return returnError(BppString::format(_T("Class '%s' not found. "),className.c_str()).c_str(),T_FUNCTION);
   BppClass* pClass = pe->getValue();
   return createObject(result,pClass,args);
}

/**
@param result Buffer to put result value into.
@param pclass Pointer to class to get instance of.
@param args Vector uf arguments to be passed to constructor.
@return true on success
@note Function does not change value of stack pointer nor value where stack pointer points to.
If result value is needed to be on stack, caller must push it manually.
*/
bool BppInterpreter::createObject(BppValue& result, BppClass* pclass, BppVector* args)
{
   if (pclass == 0)
      return returnError(_T("Invalid argument (pclass == NULL). "),T_FUNCTION);
   BppObject* pobj = new BppObject(pclass);
   bool ok = invokeMember(result,pobj,pclass->getClassName()->c_str(),args);
   if (!ok) delete pobj;
   return ok;
}

