#include "bppdef.h"
#include "bppscan.h"
#include "bppstrng.h"
#include "bppcomp.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <float.h>
#include <math.h>
#include "bppufunc.h"
#include "bppvalue.h"
#ifdef __TURBOC__
#include <dir.h>
#else
	#ifndef WINCE
		#include <direct.h>
	#endif
#endif

#define DEFAULTBUFSIZE 50
class BppCharReader
{
public:
   virtual ~BppCharReader() {}
   virtual BppChar get() =0;
};

struct BppSavedToken
{
	BppSavedToken() : dblBuffer(0.0),intnumBuffer(0),charBuffer(0),
		ttype(T_NOTOKEN),next(0) {}
	~BppSavedToken() { if (next) delete next; }
   BppCharBuffer stringBuffer;
   double dblBuffer;
   BppLong   intnumBuffer;
   BppChar charBuffer;
   BppToken ttype;
   BppSavedToken* next;
};
class BppStringCharReader : public BppCharReader
{
public:
   BppStringCharReader(const BppChar* src)
      :_src(src)
   {
      p = (BppChar*) _src.c_str();
   }
   virtual BppChar get()
   {
      BppChar r = *p;
      if (r != 0)
      {
         p++;
         return r;
      }
      return _TEOF;
   }
private:
   BppString _src;
   BppChar* p;
};

class BppFileCharReader : public BppCharReader
{
public:
   BppFileCharReader(FILE* src)
      :p(src)
   {
   }
   virtual ~BppFileCharReader()
   {
      fclose(p);
   }

   virtual BppChar get()
   {
      return feof(p) ? (BppChar) _TEOF : getc(p);
   }
private:
   FILE* p;
};

/* token name table */
static const BppChar *t_names[] = 
{
   _T("<notoken>"),
   _T("<error>"),
   _T("<eof>"),
   _T("<string>"),
   _T("<ident>"),
   _T("<long>"),
   _T("<float>"),
   _T("class"),
   _T("static"),
   _T("if"),
   _T("else"),
   _T("while"),
   _T("return"),
   _T("for"),
   _T("break"),
   _T("continue"),
   _T("do"),
   _T("switch"),
   _T("case"),
   _T("default"),
   _T("new"),
   _T("delete"),
   _T("try"),
   _T("catch"),
   _T("throw"),
   _T("var"),
   _T("function"),
   _T("operator"),
   _T("eval"),
   _T("public"),
   _T("protected"),
   _T("private"),
   _T("TRON"),
   _T("TROFF"),
   _T("TRSTEP"),
   _T("null"),
   _T("void"),
   _T("auto"),
//   _T("rem"),
   _T("<="),
   _T("=="),
   _T("!="),
   _T(">="),
   _T("<<"),
   _T(">>"),
   _T("&&"),
   _T("||"),
   _T("++"),
   _T("--"),
   _T("+="),
   _T("-="),
   _T("*="),
   _T("/="),
   _T("%="),
   _T("&="),
   _T("|="),
   _T("^="),
   _T("~="),
   _T("<<="),
   _T(">>="),
   //_T("=>"),
   //_T("=*"),
   _T("::"),
   _T("->"),
   _T("<any>")
};

/* keyword table */
static struct { const BppChar* kt_keyword; BppToken kt_token; } ktab[] = 
{
   { _T("class"),	   T_CLASS    },
   { _T("static"),	T_STATIC	  },
   { _T("if"),		   T_IF       },
   { _T("else"),	   T_ELSE     },
   { _T("while"),	   T_WHILE    },
   { _T("return"),	T_RETURN	  },
   { _T("for"),	   T_FOR      },
   { _T("break"),	   T_BREAK    },
   { _T("continue"), T_CONTINUE },
   { _T("do"),       T_DO       },
   { _T("switch"),   T_SWITCH   },
   { _T("case"),     T_CASE     },
   { _T("default"),  T_DEFAULT  },
   { _T("new"),	   T_NEW      },
   { _T("delete"),	T_DELETE	  },
   { _T("try"),	   T_TRY  	  },
   { _T("catch"), 	T_CATCH	  },
   { _T("throw"),	   T_THROW	  },
   { _T("var"),	   T_VAR 	  },
   { _T("function"),	T_FUNCTION },
   { _T("operator"),	T_OPERATOR },

 //  { _T("eval"),	   T_EVAL 	  },
   { _T("public"),	T_PUBLIC	  },
   { _T("protected"), T_PROTECTED  },
   { _T("private"),	T_PRIVATE },
   { _T("void"),	   T_VOID      },
   { _T("null"),     T_NIL      },
   { _T("auto"),     T_AUTO      },
 //  { _T("rem"),      T_REMARK   },
   { _T("TRON"),     T_TRON     },
   { _T("TROFF"),    T_TROFF    },
   { _T("TRSTEP"),   T_TRSTEP   },
   { NULL,		      T_NOTOKEN  }
};

static bool chFileDir(const BppChar* filePath)
{
 	const BppChar* dirpath = getFileDir(filePath,false);
    if (!dirpath) return false;
    return (_tchdir(dirpath) == 0);
}

BppCompiler* BppScanner::_notifyCompiler = 0;

BppScanner::BppScanner(FILE* fp, const BppChar* srcName, long lnum, unsigned lpos)
:_dblBuffer(0),_intBuffer(0),_tokenStack(0),_token(T_NOTOKEN),_savedCh(0),_lastCh(0), _lnum(lnum),
_linepos(lpos),_line(DEFAULTBUFSIZE),_stringBuffer(DEFAULTBUFSIZE),_srcName(srcName),_next(0)
{
   _charBuffer[0] = 0;
   _charBuffer[1] = 0;
   _reader = new BppFileCharReader(fp);
   _stringBuffer.addRef();
   _callerDir.addRef();
   BppChar* pdir = _tgetcwd(NULL,0);
   if (pdir)
   {
	   _callerDir = pdir;
	   freemem(pdir);
   }
   chFileDir(srcName);
}

BppScanner::BppScanner(const BppChar* src, const BppChar* srcName, long lnum, unsigned lpos)
:_dblBuffer(0),_intBuffer(0),_tokenStack(0),_token(T_NOTOKEN),_savedCh(0),_lastCh(0), _lnum(lnum),
_linepos(lpos),_line(DEFAULTBUFSIZE),_stringBuffer(DEFAULTBUFSIZE),_srcName(srcName), _next(0)
{
   _charBuffer[0] = 0;
   _charBuffer[1] = 0;
   _reader = new BppStringCharReader(src);
   _stringBuffer.addRef();
   _callerDir.addRef();
   BppChar* pdir = _tgetcwd(NULL,0);
   if (pdir)
   {
	   _callerDir = pdir;
	   freemem(pdir);
   }
}

void BppScanner::push(FILE* fp, const BppChar* srcName)
{
   if (_next != 0)
      _next->push(fp,srcName);
   else
      _next = new BppScanner(fp,srcName);
}

void BppScanner::push(const BppChar* src, const BppChar* srcName)
{
   if (_next != 0)
      _next->push(src,srcName);
   else
      _next = new BppScanner(src,srcName);
}

BppScanner::~BppScanner()
{
   if (_next != 0)
      delete _next;
   if (_callerDir.length() > 0)
	   _tchdir(_callerDir.c_str()); 
   if (_notifyCompiler) _notifyCompiler->onScannerDelete(this);
   delete _reader;
   delete _tokenStack;
}

bool BppScanner::back(bool append)
{
   if (_next != 0)
      return _next->back();
   if (_token == T_NOTOKEN) return false;
	BppSavedToken* tmp = new BppSavedToken();
	tmp->ttype = _token;
	tmp->charBuffer = _charBuffer[0];
	tmp->dblBuffer = _dblBuffer;
	tmp->intnumBuffer = _longBuffer;
	tmp->stringBuffer = _stringBuffer;
    if (!_tokenStack)
	  _tokenStack = tmp;
	else
	{
		if (append)
		{
			BppSavedToken* p = _tokenStack;;
			while (p->next) p=p->next;
			p->next = tmp;
		}
		else
		{
			tmp->next = _tokenStack;
			_tokenStack = tmp;
		}
	}
   return true;
}

BppToken BppScanner::getToken(bool ignoreBuffer)
{
   if (_next != 0)
   {
      BppToken result = _next->getToken();
      if (result != T_EOF) return result;
      delete _next;
      _next = 0;
   }
   if ((!ignoreBuffer) && _tokenStack)
   {
	   BppSavedToken* p = _tokenStack;
	   _tokenStack = p->next;
	   p->next = 0;
	   _token = p->ttype;
	   _charBuffer[0] = p->charBuffer;
	   _dblBuffer =  p->dblBuffer;
	   _longBuffer = p->intnumBuffer;
	   _stringBuffer = p->stringBuffer;
	   delete p;
   }
   else
      _token = readToken();
   return _token;
}

const BppChar* BppScanner::getSourceName() const
{
   if (_next != 0)
      return _next->getSourceName();
   return _srcName.c_str();
}

long BppScanner::getLineNumber() const
{
   if (_next != 0)
      return _next->getLineNumber();
   return _lnum;
}

unsigned BppScanner::getLinePos() const
{
   if (_next != 0)
      return _next->getLinePos();
   return _linepos;
}

BppToken BppScanner::readToken()
{
   BppChar ch,ch2;

   /* check the next character */
   while (1)
   {
      switch (ch = skipSpaces()) 
      {
      case _TEOF:	return (T_EOF);
      case _T('"'):	return (readString());
      case _T('\''):	return (readChar());
      case _T('<'):	
         switch (ch = get()) 
         {
         case _T('='):
            return (T_LE);
         case _T('<'):
            switch (ch = get())
            {
            case _T('='):
               return (T_SHLEQ);
            default:
               _savedCh = ch;
               return (T_SHL);
            }
         default:
            _savedCh = ch;
            _charBuffer[0] = _T('<');
            return (T_ANY);
         }
      case _T('='):	
         switch (ch = get()) 
         {
         case _T('='):
            return (T_EQ);
         //case _T('>'):
         //   return (T_REFEQ);
         //case _T('*'):
         //   return (T_UNREFEQ);
         default:
            _savedCh = ch;
            _charBuffer[0] = _T('=');
            return (T_ANY);
         }
      case _T('!'):	
         switch (ch = get()) 
         {
         case _T('='):
            return (T_NE);
         default:
            _savedCh = ch;
            _charBuffer[0] = _T('!');
            return (T_ANY);
         }
      case _T('>'):	
         switch (ch = get()) 
         {
         case _T('='):
            return (T_GE);
         case _T('>'):
            switch (ch = get())
            {
            case _T('='):
               return (T_SHREQ);
            default:
               _savedCh = ch;
               return (T_SHR);
            }
         default:
            _savedCh = ch;
            _charBuffer[0] = _T('>');
            return (T_ANY);
         }
      case _T('&'):	
         switch (ch = get()) 
         {
         case _T('&'):
            return (T_AND);
         case _T('='):
            return (T_BANDEQ);
         default:
            _savedCh = ch;
            _charBuffer[0] = _T('&');
            return (T_ANY);
         }
      case _T('|'):	
         switch (ch = get()) 
         {
         case _T('|'):
            return (T_OR);
         case _T('='):
            return (T_BOREQ);
         default:
            _savedCh = ch;
            _charBuffer[0] = _T('|');
            return (T_ANY);
         }
      case _T('^'):	
         switch (ch = get()) 
         {
         case _T('='):
            return (T_XOREQ);
         default:
            _savedCh = ch;
            _charBuffer[0] = _T('^');
            return (T_ANY);
         }
      case _T('~'):	
         switch (ch = get()) 
         {
         case _T('='):
            return (T_BNOTEQ);
         default:
            _savedCh = ch;
            _charBuffer[0] = _T('~');
            return (T_ANY);
         }
      case _T('+'):	
         switch (ch = get()) 
         {
         case _T('+'):
            return (T_INC);
         case _T('='):
            return (T_ADDEQ);
         default:
            _savedCh = ch;
            _charBuffer[0] = _T('+');
            return (T_ANY);
         }
      case _T('-'):	
         switch (ch = get()) 
         {
         case _T('-'):
            return (T_DEC);
         case _T('='):
            return (T_SUBEQ);
         case _T('>'):
            return (T_MEMREF);
         default:
            _savedCh = ch;
            _charBuffer[0] = _T('-');
            return (T_ANY);
         }
      case _T('*'):	
         switch (ch = get()) 
         {
         case _T('='):
            return (T_MULEQ);
         default:
            _savedCh = ch;
            _charBuffer[0] = _T('*');
            return (T_ANY);
         }
      case _T('%'):	
         switch (ch = get()) 
         {
         case _T('='):
            return (T_REMEQ);
         default:
            _savedCh = ch;
            _charBuffer[0] = _T('%');
            return (T_ANY);
         }
      case _T('@'):
         while ((ch = get()) != _TEOF)
            if (ch == _T('\n'))
               break;
         break;
      case _T('#'):
         if ((ch=get()) == _T('!'))
         {
            while ((ch = get()) != _TEOF)
               if (ch == _T('\n'))
                  break;
            break;
         }
         _savedCh = ch;
         _charBuffer[0] = _T('#');
         return (T_ANY);
      case _T('/'):	
         switch (ch = get()) 
         {
         case _T('='):
            return (T_DIVEQ);
         case _T('/'):
            while ((ch = get()) != EOF)
               if (ch == _T('\n'))
                  break;
            break;
         case _T('*'):
            ch = ch2 = _TEOF;
            for (; (ch2 = get()) != _TEOF; ch = ch2)
               if (ch == _T('*') && ch2 == _T('/'))
                  break;
            break;
         default:
            _savedCh = ch;
            _charBuffer[0] = _T('/');
            return (T_ANY);
         }
         break;
      case _T(':'):	
         switch (ch = get()) 
         {
         case _T(':'):
            return (T_CC);
         default:
            _savedCh = ch;
            _charBuffer[0] = _T(':');
            return (T_ANY);
         }
      default:	
         if (_istdigit((BppUChar) ch))
            return (readNumber(ch));
         else if (isIdChar(ch))
         {
			 if (ch == _T('R'))
			 {
				 if ((ch = get()) == _T('"')) // read raw string
				 {
                     BppString endtoken = _T(")");
					 while (isIdChar(ch = get())) endtoken += ch;
					 endtoken += _T('"');
					 size_t lend = endtoken.length();
					 if (ch != _T('(')) return parseError(_T("invalid raw string literal"));
                     _stringBuffer.clear();
					 size_t cnt = 0;
					 const BppChar* pend = endtoken.c_str();
					 while ((ch = get()) != _TEOF)
					 {
						 _stringBuffer.push_back(ch);
						 if (++cnt < lend) continue;
						 if (_tcsncmp(_stringBuffer.data()+(cnt-lend),pend,lend) == 0)
						 {
                             _stringBuffer[cnt-lend] = 0;
							 _stringBuffer.adjustSize();
							 return (T_STRING);
						 }
					 }
					 return parseError(_T("unexpected EOF in raw string literal"));
				 }
				 else
				 {
					 _savedCh = ch;
					 ch = _T('R');
				 }
			 }
            BppToken t = readIdentifier(ch);
            //if (t == T_REMARK)
            //{
            //   while ((ch = get()) != _TEOF)
            //      if (ch == _T('\n'))
            //         break;
            //   break;
            //}
            return t;
         }
         else {
            _charBuffer[0] = ch;
            return (T_ANY);
         }
      }
   }
}

const BppChar* BppScanner::getTokenName(BppToken tkn) const
{
   if (_next != 0)
      return _next->getTokenName(tkn);

   if ((tkn >= T_NOTOKEN) && (tkn < T_ANY))
      return t_names[tkn];
   else if (tkn == T_ANY)
   {
	   if (_charBuffer[0] == 0) return _T("special character");
      return _charBuffer;
   }
   return _T("invalid token");
}

const BppChar* BppScanner::getTokenString() const
{
   if (_next != 0)
      return _next->getTokenString();
   return (const BppChar*) _stringBuffer.data();
}

double BppScanner::getDouble() const
{
   if (_next != 0)
      return _next->getDouble();
   return _dblBuffer;
}

BppInt BppScanner::getLong() const
{
   if (_next != 0)
      return _next->getLong();
   return _intBuffer;
}

BppUInt BppScanner::getULong() const
{
   if (_next != 0)
      return _next->getULong();
   return _uintBuffer;
}

BppLong BppScanner::getLLong() const
{
   if (_next != 0)
      return _next->getLLong();
   return _longBuffer;
}

BppULong BppScanner::getULLong() const
{
   if (_next != 0)
      return _next->getULLong();
   return _ulongBuffer;
}

BppValue BppScanner::getNumberValue() const
{
	switch (getNumberType())
	{
	case BVT_UINT:
		return getULong();
	case BVT_LONG:
		return getLLong();
	case BVT_ULONG:
		return getULLong();
	case BVT_FLOAT:
		return (float) getDouble();
	case BVT_DOUBLE:
		return getDouble();
	}
	return getLong();
}

BppValueType BppScanner::getNumberType() const 
{ 
   if (_next != 0)
      return _next->getNumberType();
	return _numType; 
}

BppChar BppScanner::getChar() const
{
   if (_next != 0)
      return _next->getChar();
   return _charBuffer[0];
}

BppChar BppScanner::get()
{
   BppChar ch;

   /* check for a lookahead character */
   if ((ch = _savedCh) != _T('\0'))
      _savedCh = _T('\0');

   /* check for a buffered character */
   else {
      while ((ch = _line[_linepos++]) == _T('\0')) 
      {
         /* check for being at the end of file */
         if (_lastCh == _TEOF)
            return (_lastCh);

         /* read the next line */
         _line.clear();
         _linepos=0 ;
         while ((_lastCh = _reader->get()) != _TEOF && _lastCh != _T('\n'))
         {
            _line.push_back(_lastCh);
            _linepos++;
         }
         _line.push_back(_T('\n'));
         _line.push_back(_T('\0'));
         _linepos = 0;
         ++_lnum;
      }
   }

   /* return the current character */
   return (ch);
}


BppChar BppScanner::skipSpaces()
{
   BppUChar ch;
   while ((ch = get()) != _T('\0') && _istspace(ch));
   if (ch == _T('\\'))
   {
	   BppChar c = get();
	   if (c == _T('\n')) return skipSpaces();
	   _savedCh = c;
   }
   return (ch);
}

BppToken BppScanner::readString()
{
   BppChar ch;
   bool bslash = false;
   _stringBuffer.clear();
   /* get the string */
   while (1)
   {
      ch = get();
      _savedCh = ch;
      bslash = (ch == _T('\\'));
      ch = readLiteralChar();
      if ((ch == _TEOF) || ((!bslash) && (ch == _T('"'))))
      {
         if (ch == _T('"'))
         {
            ch = skipSpaces();
            if (ch == _T('"')) continue;
            _savedCh = ch;
         }
         break;
      }
      if (ch == 0)
      {
         BppCharBuffer b = _stringBuffer;
		 if (_savedCh == _T('x') || _istdigit(_savedCh))
		 {
			 ch = _T('0');
			 if (_savedCh == _T('0')) _savedCh = 0;
		 }
		 else {
			 ch = _savedCh;
			 _savedCh = 0;
		 }
         if (readNumber(ch,true) != T_NUMBER)
            return parseError(_T("number literal expected"));
         ch = (BppChar) _intBuffer;
         _stringBuffer = b;
      }
      _stringBuffer.push_back(ch);
   }
   //*p++ = ch;
   if (ch == _TEOF) _savedCh = ch;
   _stringBuffer.push_back(_T('\0'));
   _stringBuffer.adjustSize();
   return (T_STRING);
}

/// character constant literal
BppToken BppScanner::readChar()
{
   _charBuffer[0] = readLiteralChar();
   if (_charBuffer[0] == 0)
   {
       BppChar ch = _savedCh;  
	   if (_savedCh == _T('x') || _istdigit(_savedCh)) {
		   ch = _T('0');
		   if (_savedCh == _T('0')) _savedCh = 0;
	   }
	   else {
           _savedCh = 0;
	   }
       if (readNumber(ch,true) != T_NUMBER)
          return parseError(_T("number literal expected"));
   }
   else
   {
	   _numType = BVT_INT;
	   _intBuffer = _charBuffer[0];
   }
   if (get() != _T('\''))
      return parseError(_T("closing single quote expected"));
   return (T_NUMBER);
}

BppChar BppScanner::readLiteralChar()
{
   BppUChar ch = get();
   if (ch == _T('\\'))
   {
      switch (ch = get()) 
      {
      case _T('n'):  ch = _T('\n'); break;
      case _T('r'):  ch = _T('\r'); break;
      case _T('t'):  ch = _T('\t'); break;
      case _T('v'):  ch = _T('\v'); break;
      case _T('f'):  ch = _T('\f'); break;
      case _T('a'):  ch = _T('\a'); break;
      case _T('b'):  ch = _T('\b'); break;
	  case _T('\n'): return readLiteralChar(); // ch = get(); break;
	  case _T('x'): 
	  case _T('u'):
	  case _T('U'):
		  _savedCh = _T('x'); ch = 0; break;
      case _TEOF:  ch = _T('\\'); _savedCh = _TEOF; break;
      default:
            if (ch >= _T('0') && ch <= _T('7'))
            {
               _savedCh = ch;
               ch = 0;
            }
            break;
      }
   }
   return (ch);
}

BppToken BppScanner::readIdentifier(BppChar firstChar)
{
   BppChar ch;
   /* get the identifier */
   _stringBuffer.clear();
   _stringBuffer.push_back(firstChar);
   while ((ch = get()) != _TEOF && isIdChar(ch))
      _stringBuffer.push_back(ch);
   _savedCh = ch;
   _stringBuffer.push_back(_T('\0'));
   _stringBuffer.adjustSize();

   /* check to see if it is a keyword */
   const BppChar* pt = (const BppChar*) _stringBuffer.data();
   size_t i;
   for (i = 0; ktab[i].kt_keyword != NULL; ++i)
   {
      if (_tcscmp(ktab[i].kt_keyword,pt) == 0)
         return (ktab[i].kt_token);
   }
   if (_tcscmp(pt,_T("__FILE__")) == 0)
   {
      _stringBuffer.clear();
      size_t n = _srcName.length();
      pt = _srcName.c_str();
      for (i=0;i<n;++i)
         _stringBuffer.push_back(pt[i]);
      _stringBuffer.push_back(_T('\0'));
      _stringBuffer.adjustSize();
      return T_STRING;
   }
   if (_tcscmp(pt,_T("__func__")) == 0)
   {
      _stringBuffer.clear();
      size_t n = _funcName.length();
      pt = _funcName.c_str();
      for (i=0;i<n;++i)
         _stringBuffer.push_back(pt[i]);
      _stringBuffer.push_back(_T('\0'));
      _stringBuffer.adjustSize();
      return T_STRING;
   }
   if (_tcscmp(pt,_T("__LINE__")) == 0)
   {
     _intBuffer = _lnum;
     _dblBuffer = 0;
     BppString nstr = BppString::format(_T("%d"),_lnum);
     _stringBuffer.clear();
      size_t n = nstr.length();
      pt = nstr.c_str();
      for (i=0;i<n;++i)
         _stringBuffer.push_back(pt[i]);
      _stringBuffer.push_back(_T('\0'));
      _stringBuffer.adjustSize();
      return T_NUMBER;
   }
   return (T_IDENTIFIER);
}

BppToken BppScanner::readNumber(BppChar firstChar, bool isCharcode)
{
   bool xnum = false;
   bool bnum = false;
   _numType = BVT_NULL;
   BppUChar ch;
   BppToken ttype = T_NUMBER;
   /* get the number */
   //p = t_token; *p++ = ch; t_value = 0;
   _stringBuffer.clear();
   _stringBuffer.push_back(firstChar);
   _intBuffer = 0;
   _dblBuffer = 0;
   BppChar lastChar = (BppChar)_totlower(firstChar);
   while ((ch = get()) != _TEOF)
   {
	   if ( (!isCharcode) && ch == _T('\''))
	   {
		   if (!(_istdigit(lastChar) || (xnum && _istxdigit(lastChar)))) break;
		   lastChar = ch;
		   continue;
	   }
	   lastChar = (BppChar)_totlower(ch);
      if (!_istdigit(ch))  // current character is not a digit
      {
         if (!(xnum || bnum))
         {
            if (ch == _T('.'))
            {
               if (ttype == T_FNUMBER)  // only one decimal point allowed
                  break;
               ttype = T_FNUMBER;
            }
            else if (lastChar  == _T('e')) // exponent
            {
               ttype = T_FNUMBER;
               _stringBuffer.push_back(ch);
               lastChar = (BppChar)_totlower(ch = get());
               if (!(_istdigit(ch) || (ch == _T('-')) || (ch == _T('+'))))
                  break; // exponent must start with digit or sign
            }
            else
			{
               if  ((lastChar == _T('x')) && (_stringBuffer.size() < 3 && (_stringBuffer[_stringBuffer.size()-1]==_T('0'))) )
                  xnum = true;
               else if ((lastChar == _T('b')) && (_stringBuffer.size() < 3 && (_stringBuffer[_stringBuffer.size()-1]==_T('0'))) )
				   bnum = true;
			   else
                  break;
			}
         }
         else if (bnum && (ch != '0') && (ch != '1'))
			 break;
	     else if (!_istxdigit(ch))
            break;
      }
      _stringBuffer.push_back(ch);
   }
   _savedCh = ch;
   _stringBuffer.push_back(_T('\0'));
   _stringBuffer.adjustSize();
   bool isL = false;
   bool isU = false;
   bool isF = false;
   if (!isCharcode)
   {
	   if (ch == _T('\''))  return parseError(_T("invalid quoted number literal"));

	   // check suffixes
	   isL = lastChar == _T('l');
	   isU = (ttype == T_NUMBER) && (lastChar == _T('u'));
	   isF = (ttype == T_FNUMBER) && (lastChar == _T('f'));
	   if (isL || isU || isF)
	   {
		   get(); // skip buffered char
		   if (ttype == T_NUMBER) // first suffix char was l or u
		   {
			   lastChar =(BppChar) _totlower(ch = get());
			   if (isL) // first suffix was l
			   {
				   if (lastChar == _T('l')) // skip second l
				   {
					   lastChar =(BppChar) _totlower(ch = get());
				   }
				   if (lastChar == _T('u')) // check u
					   isU = true;
				   else _savedCh = ch; // push char
			   }
			   else if (isU) // first suffix was u
			   {
				   if (lastChar == _T('l'))
				   {
					   isL = true;
					   lastChar = (BppChar)_totlower(ch = get()); // skip second l if exists
					   if (lastChar != _T('l'))
					   {
						   _savedCh = ch; // push char
					   }
				   }
				   else _savedCh = ch; // push char
			   }
		   }
	   }
   }
   /// convert to number

   static const BppULong uintmax_i64 = BppULong((unsigned long)0xFFFFFFFFUL);
   static const BppULong intmax_i64 = BppULong((unsigned long)0x7FFFFFFFUL);
   static const BppLong zero = BppLong(0);
   BppChar* endp = 0;
   if (ttype == T_NUMBER)
   {
	   if (bnum) 
		   _ulongBuffer = _tcstoui64(_stringBuffer.data()+2,&endp,2);
	   else
           _ulongBuffer = _tcstoui64(_stringBuffer.data(),&endp,0);
	   _numType = BVT_LONG;
	   if (
#ifndef EMULATE_I64
	#ifndef WINCE
		   (errno == ERANGE) || 
	#endif
#endif
		   ((!isU) && (*((BppLong*)&_longBuffer) < zero)))
          return parseError(_T("integer literal out of range"));
	   if (isU)
	   {
		   if ((!isL) && (*((BppULong*)&_ulongBuffer) <= uintmax_i64))
			   _numType = BVT_UINT;
		   else
			   _numType = BVT_ULONG;
	   }
	   else if ((!isL) && (*((BppULong*)&_ulongBuffer) <= intmax_i64))
		   _numType = BVT_INT;
   }
   else
   {
	   _dblBuffer = _tcstod(_stringBuffer.data(),&endp);
	   _numType = (isF) ? BVT_FLOAT : BVT_DOUBLE;
	   if (
#ifndef __OLDCPP__
#ifndef WINCE
		   (errno == ERANGE) ||
#endif
#endif
		   (isF && (fabs(_dblBuffer) > FLT_MAX)))
          return parseError(_T("real literal out of range"));
      //_stscanf(_stringBuffer.data(),_T("%lf"),&_dblBuffer);
   }
   if (endp && *endp != _T('\0'))
	   return parseError(_T("invalid number literal"));
   return (ttype);
}


bool BppScanner::isIdChar(BppUChar ch)
{

   return (_istupper(ch)
      || _istlower(ch)
      || _istdigit(ch)
      || ch == _T('_')
	  || ch == _T('$'));
}

BppToken BppScanner::parseError(const BppChar* msg)
{
   if (_next != 0)
      return _next->parseError(msg);
   _stringBuffer.setCapacity(_tcslen(msg)+ _srcName.length() + _line.size() + 60);
   /* redisplay the line with the error */
   _stprintf(_stringBuffer.data(),_T("--> %s \nin %s line %ld pos %u \n%s\n"),msg,_srcName.c_str(),_lnum,_linepos,_line.data());
   _stringBuffer.adjustSize();
   return T_ERROR;
}

