#include <malloc.h>
#include <limits.h>

#include "bppvm.h"
#include "bppprint.h"
#include "bppobjct.h"
#include "bppint.h"
#include "bppvec.h"
#include "bppclass.h"
#include "bppstrng.h"
#include "bppcomp.h"
#include "bppdbg.h"
#include "dictio.h"
#include "bppdfunc.h"
#include "bppufunc.h"
#include "bppregex.h"
#ifdef WIN32
#include <windows.h>
UINT BppVirtualMachine::_breakRequestMsg = RegisterWindowMessage(_T("BppBreakRequestMsg"));
UINT BppVirtualMachine::_dbgWaitMsg = RegisterWindowMessage(_T("BppDbgWaitMsg"));
UINT BppVirtualMachine::_lineChangedMsg = RegisterWindowMessage(_T("BppLineChangedMsg"));
#endif
#ifdef __MSDOS__
   #include <usedll.h>
   typedef HDLL HMODULE;
   typedef int BOOL;
#endif

#ifndef DOXY
struct TypeIdEntry
{
	const BppChar* key;
	BppValueType id;
};
static const TypeIdEntry _typeMapping[] = {
    { _T("int"), BVT_INT },
    { _T("uint"), BVT_UINT },
	{ _T("bool"), BVT_INT },
    { _T("short"), BVT_SHORT },
    { _T("ushort"), BVT_USHORT },
	{ _T("float"), BVT_FLOAT }, 
    { _T("sbyte"), BVT_SBYTE },
    { _T("byte"), BVT_BYTE },
    { _T("long"), BVT_LONG },
	{ _T("ulong"), BVT_ULONG },
	{ _T("intptr"), BVT_INTPTR },
	{ _T("double"), BVT_DOUBLE }, 
	{ _T("FILE"), BVT_FILE }, 
	{ _T("buffer"), BVT_BUFFER }, 
	{ _T("charbuffer"), BVT_CHARBUFFER }, 
	{ _T("object"), BVT_OBJECT }, 
	{ _T("vector"), BVT_VECTOR }, 
	{ _T("string"), BVT_STRING }, 
	{ _T("dictionary"), BVT_DICTIONARY },
	{ _T("function"), BVT_BYTECODE },
	{ _T("function"), BVT_CODE },
	{ _T("void"), BVT_VOID }, 
	{ _T("auto"), BVT_AUTO }, 
	//{ _T("function"), BVT_CODEREF },     ///< reference to bytecode vector
	{ _T("function"), BVT_BYTECODEREF }, ///< reference to bytecode vector
	{ _T("buffer"), BVT_BUFFERREF }, ///< reference to byte buffer
	{ _T("charbuffer"), BVT_CHARBUFFERREF }, ///< reference to character buffer
	{ _T("object"), BVT_OBJECTREF }, ///< reference to an object
	{ _T("vector"), BVT_VECTORREF }, ///< reference to vector
	{ _T("string"), BVT_STRINGREF }, ///< reference to string
	{ _T("dictionary"), BVT_DICTIONARYREF } ///< dictionary reference
};
static const size_t _typeMapSize = sizeof(_typeMapping) / sizeof(TypeIdEntry);

#endif

/**
Constructor initializes internal symbol tables for virtual machine, installs
default printer and registers predefined B++ functions for current instance.
@param printer Pointer to IBppPrinter derived object used for default printer.
If <i>printer</i> is <tt>NULL</tt> a dummy printer will be installed, i.e.
all output is discarded by default.
@note The printer installed here is used for debug and error output too if 
no special printers are specified therefore.
@see IBppPrinter, BppDummyPrinter, execute, compileFile, compileText
*/
BppVirtualMachine::BppVirtualMachine(IBppPrinter* printer)
:_genLineInfo(false),_genLocalVarInfo(false),_printer(printer),_ownsPrinter(printer == 0),_interpreter(0),_cmdargc(0),_cmdargv(0),
_classes(),_symbols(),_dbgWaitProc(0),_breakRequestProc(0),_lineChangedProc(0),_defaultInput(stdin),_currentLine(-1),_inDbg(false)
#ifdef WIN32
,_hwndDbg(0)
#endif
{
   _currentInput = &_defaultInput;
   if (_printer == 0)
      _printer = new BppDummyPrinter();
   _classes.addRef();
   _symbols.addRef();
   _modules.addRef();
   _currentSource.addRef();
   BppDefaultFunctions::registerFunctions(this);
   Bpp_RegEx::createClass(*this);
}

/**
Destructor releases all native modules loaded by virtual machine and clears
internal symbol tables.
*/
BppVirtualMachine::~BppVirtualMachine()
{
   _classes.clear();
   _symbols.clear();
   for (BppDictionaryEntry* pentry= _modules.getFirst();pentry != 0;pentry = pentry->getNext())
      FreeLibrary((HMODULE)pentry->getValue().ptr());
   _modules.clear();
   if (_ownsPrinter)
      delete _printer;
   _printer = 0;
}

BppValueType BppVirtualMachine::findVTypeFromName(const BppString& name, bool includeClasses, BppClass** classType)
{
	if (includeClasses)
	{
		BppDictionaryEntry* pe = _classes.find(name);
		if (pe)
		{
		   if (classType)
		   {
			   *classType = (BppClass*) pe->getValue();
			   return BVT_STRONGOBJECT;
		   }
	       return BVT_OBJECT;
		}
	}
	const BppChar* s = name.c_str();
	for (size_t it=0;it<_typeMapSize;++it)
	{
        const TypeIdEntry& entry = _typeMapping[it];
		if (_tcscmp(s,entry.key) == 0)
		{
			return _typeMapping[it].id;
		}
		if (entry.id >= _BVTREFMIN) break;
	}
    return BVT_NULL;    
}

const BppChar* BppVirtualMachine::getTypeNameFromVType(BppValueType vtype)
{
	for (size_t it=0;it<_typeMapSize;++it)
	{
		if (_typeMapping[it].id == vtype)
		{
			return _typeMapping[it].key;
		}
	}
	return 0;
}

bool BppVirtualMachine::getGenerateLineInfo() const
{
   return _genLineInfo;
}

void BppVirtualMachine::setGenerateLineInfo(bool genInfo)
{
   _genLineInfo = genInfo;
}

bool BppVirtualMachine::getGenerateLocalVarInfo() const
{
   return _genLocalVarInfo;
}

void BppVirtualMachine::setGenerateLocalVarInfo(bool genInfo)
{
   _genLocalVarInfo = genInfo;
}

const BppString& BppVirtualMachine::getCurrentSource() const
{
   return _currentSource;
}

long BppVirtualMachine::getCurrentLine() const
{
   return _currentLine;
}

/**
Function releases all native modules loaded by virtual machine and clears
internal symbol tables. At last it re-registers predefined B++ functions for current instance.
@note This function should be used in order to clean up all precompiled stuff
in virtual machine without delete it, e.g. in order to compile and execute
a new program.
*/
void BppVirtualMachine::reinit()
{
   _classes.clear();
   _symbols.clear();
   _currentLine = -1;
   _currentSource.clear();
   for (BppDictionaryEntry* pentry= _modules.getFirst();pentry != 0;pentry = pentry->getNext())
      FreeLibrary((HMODULE)pentry->getValue().ptr());
   _modules.clear();
   BppDefaultFunctions::registerFunctions(this);
   Bpp_RegEx::createClass(*this);
}

/**
Method tries to register a native function (written in C++) for use with B++ in current
virtual machine. Client code may use it to install user defined extension functions.
@param name Name of user defined function for use in B++. The name must be unique
in context of VM.
@param pfunc Function pointer. Please note that function must use STDCALL convention
on Win32 platforms.
@return <tt>true</tt> on success
@note Function fails if <i>name</i> is invalid, <i>pfunc</i> is NULL or
there is already a symbol defined for <i>name</i>.
*/
bool BppVirtualMachine::registerNativeFunction(const BppChar* name, BPPUSERFUNC pfunc)
{
   if ((name == 0) || (_tcslen(name) == 0))
      return false;
   if (pfunc == 0)
      return false;
   BppDictionaryEntry& de = _symbols[name];
   if (de.getSymbolType() != ST_NONE) return false;
   de.setValue(pfunc);
   de.setSymbolType(ST_SFUNCTION);
   return true;
}

/**
Method tries to register a native class (written in C++) for use with B++ in current
virtual machine. Client code may use it to install user defined extension classes.
@param name Name of user defined function for use in B++. The name must be unique
in context of VM.
@param pClass Pointer to a class object. Class object must be created using
current virtual machine in its constructor.
@return <tt>true</tt> on success
@note Function fails if <i>name</i> is invalid, <i>pClass</i> is NULL or
there is already a symbol defined for <i>name</i>.
*/
bool BppVirtualMachine::registerNativeClass( const BppChar* name, BppClass* pClass )
{
   if ((name == 0) || (_tcslen(name) == 0))
      return false;
   if (pClass == 0) return false;
   BppDictionaryEntry& de = _classes[name];
   if (de.getSymbolType() != ST_NONE) return false;
   pClass->setClassName(name);
   pClass->setIsNative(true);
   de.setValue(pClass);
   de.setSymbolType(ST_CLASS);
   return true;
}

/// name of destructor function
static const BppString dtorname = _T("dtor");
/**
Function is called from BppClass whenever an object instance is to be deleted.
It calls objects destructor (if any).
@param obj object to delete
*/
void BppVirtualMachine::notifyObjDelete(BppObject* obj)
{
   if (_interpreter != 0)
   {
      BppClass* pclass = obj->getClass(); // should never return NULL
      if (pclass->findMemberPL(BVF_PRIVATE,dtorname) == 0) return; // return if no dtor defined
      ValPtr savedSP =_interpreter->_sp;
      _interpreter->_sp = _interpreter->_stack.last();
#ifndef NOCPPEXCEPTION  // catch all exceptions in destructor call
      try {
#endif
      ++(_interpreter->_sp);  // push object to stack
      (*_interpreter->_sp) = BppValue(obj,true); // make reference to prevent from delete
      if (!_interpreter->send(0,dtorname.c_str()))
      {
         // TODO: send error message if dtor call fails
      }       

#ifndef NOCPPEXCEPTION
      } catch (...) {}
#endif
      // cleanup stack
      //while (_interpreter->_sp.current() > savedSP.current())
      //{
      //   _interpreter->_sp->dispose();
      //   --_interpreter->_sp;
      //}
      _interpreter->_sp = savedSP;
   }
}

/**
Function provides read/write access to variable for number of command line
arguments. It is typically used by an application during initializing
of VM.
@return writeable reference to variable for number of command line arguments
@see cmdargv
*/
int& BppVirtualMachine::cmdargc()
{
   return _cmdargc;
}

/**
Function provides read/write access to variable for array of command line
arguments. It is typically used by an application during initializing
of VM.
@return writeable reference to variable for array of command line arguments
@note <p>Client code should provide arguments array in order to make
argument related B++ functions useable.</p>
<p>Virtual machine does not copy nor delete arguments array, i.e. client
code is responsible to provide the array for entire live time of VM.</p>
@see cmdargc, Bpp_getargs, Bpp_getusrargs, Bpp_argcnt, Bpp_arg
*/
BppChar**& BppVirtualMachine::cmdargv()
{
   return _cmdargv;
}

/**
@return writeable reference to class table of VM
@note This function is intended for internal use only.
*/
BppDictionary& BppVirtualMachine::classes()
{
   return _classes;
}

/**
@return writeable reference to symbol table of VM
@note This function is intended for internal use only.
*/
BppDictionary& BppVirtualMachine::symbols()
{
   return _symbols;
}

/**
@return writeable reference to module table of VM
@note This function is intended for internal use only.
*/
BppDictionary& BppVirtualMachine::modules()
{
   return _modules;
}

/**
@return reference to default printer of VM
@note Using default printer allows formatted output operations
in implementation of native functions.
*/
IBppPrinter& BppVirtualMachine::printer()
{
   return *_printer;
}

/**
@return pointer to interpreter currently used by VM (or NULL if none)
@note This function is mainly for internal use. However, in some cases
access to interpreter may be useful in implementation of native
functions too, e.g. in order to get execution context or to emit
an error message.
@see BppInterpreter::getProcInfo, BppInterpreter::returnError
*/
BppInterpreter* BppVirtualMachine::interpreter() const
{
   return _interpreter;
}

/**

*/
void BppVirtualMachine::cloneSymbols(const BppVirtualMachine& other)
{
   _symbols.clear();
   _classes.clear();
   BppDictionaryEntry* pe;
   for (pe= _modules.getFirst();pe != 0;pe = pe->getNext())
      FreeLibrary((HMODULE)pe->getValue().ptr());
   _modules.clear();
   for (pe = other._modules.getFirst();pe != 0;pe = pe->getNext())
      loadNativeModule(pe->getKey()->c_str());
   for (pe = other._symbols.getFirst();pe != 0;pe = pe->getNext())
      (*_symbols[pe->getKey()]) = *pe;
   for (pe = other._classes.getFirst();pe != 0;pe = pe->getNext())
      (*_classes[pe->getKey()]) = *pe;
}

#ifndef DOXY
// Helper struct for compileXXX functions
// It ensures releasing of debugger and closing source file.
struct CompileHelper
{
   CompileHelper(FILE* fp, BppDebugger* dbg) : _file(fp),_dbg(dbg) {}
   ~CompileHelper()
   {
      if (_file != 0) fclose(_file);
      if (_dbg != 0) delete _dbg;
   }
   FILE* _file;
   BppDebugger* _dbg;
};
#endif

bool BppVirtualMachine::compileFile(const BppChar* filename, bool showinfo, bool decode, IBppPrinter* errPrinter, IBppPrinter* dbgPrinter)
{
   if (filename == 0) return false;
   FILE* fp = _tfopen(filename,_T("rt"));
   if (fp==0) 
   {
      (*_printer)(_T("error opening file %s\n"),filename);
      return false;
   }
   BppCompiler comp(*this,errPrinter);
   comp.setShowInfo(showinfo);
   comp.setGenerateLineInfo(_genLineInfo);
   comp.setGenerateLocalVarInfo(_genLocalVarInfo);
   BppDebugger* dbg = 0;
   if (decode)
   {
      dbg = new BppDebugger(*this,dbgPrinter);
      comp.setDebugger(dbg);
   }
   CompileHelper hlp(fp,dbg);
   bool result = comp.compile(fp,filename);
   return result;
}

bool BppVirtualMachine::compileText(const BppChar* src, bool showinfo, bool decode, IBppPrinter* errPrinter, IBppPrinter* dbgPrinter)
{
   if (src == 0) return false;
   BppCompiler comp(*this,errPrinter);
   comp.setShowInfo(showinfo);
   comp.setGenerateLineInfo(_genLineInfo);
   comp.setGenerateLocalVarInfo(_genLocalVarInfo);
   BppDebugger* dbg = 0;
   if (decode)
   {
      dbg = new BppDebugger(*this,dbgPrinter);
      comp.setDebugger(dbg);
   }
   CompileHelper hlp(0,dbg);
   return comp.compile(src);
}

struct ExecHelper
{
   ExecHelper(BppInterpreter*& interpreter, BppDebugger* dbg) : _interpreter(interpreter),_dbg(dbg)
   {
      _interpreter->setDebugger(dbg);
   }
   ~ExecHelper()
   {
      _interpreter->setDebugger(0);
      delete _interpreter;
      _interpreter = 0;
      if (_dbg != 0) delete _dbg;
   }
   BppInterpreter*& _interpreter;
   BppDebugger* _dbg;
private:
   ExecHelper(const ExecHelper&);
   ExecHelper& operator=(const ExecHelper&);
};

bool BppVirtualMachine::execute(const BppChar* funcName, int argc, const BppChar** argv, bool trace, IBppPrinter* errPrinter, IBppPrinter* dbgPrinter)
{
   if (funcName == 0) return false;
   _interpreter = new BppInterpreter(*this);
   if (_breakRequestProc != 0)
      _interpreter->setAllowBreakRequest(true);
   BppDebugger* dbg = 0;
   dbg = new BppDebugger(*this,dbgPrinter);
   dbg->active = trace;
   ExecHelper hlp(_interpreter,dbg);
   bool result = _interpreter->execute(funcName,argc,argv);
   _result = *(_interpreter->_sp);
   if (!result)
   {
      if (errPrinter != 0)
         (*errPrinter)(_T("%s"),_interpreter->errors().c_str());
      else
         (*_printer)(_T("%s"),_interpreter->errors().c_str());
   }
   return result;
}

const BppValue& BppVirtualMachine::getResult() const
{
   return _result;
}

bool BppVirtualMachine::writeModule(const BppChar* filename)
{
   if (filename == 0) return false;
   BppDictionaryIO io(*this);
   return io.writeModule(filename);
}

bool BppVirtualMachine::writeModule(FILE* fp)
{
   if (fp == 0) return false;
   BppDictionaryIO io(*this);
   return io.writeModule(fp);
}

bool BppVirtualMachine::readModule(const BppChar* filename)
{
   if (filename == 0) return false;
   BppDictionaryIO io(*this);
   return io.readModule(filename);
}

bool BppVirtualMachine::readModule(FILE* fp)
{
   if (fp == 0) return false;
   BppDictionaryIO io(*this);
   return io.readModule(fp);
}


bool BppVirtualMachine::loadNativeModule(const BppChar* dllName)
{
   if (dllName == 0) return false;
   BppString _dllname = dllName;
   _dllname.replace(_T("*"),BPP_DLLWILD);
   if (_modules.find(_dllname) != 0) return true;
#ifdef WINCE
   const BppChar* dllPath = searchPath(_dllname.c_str());
   if (dllPath == 0)
	   dllPath = _dllname.c_str();
   HMODULE handle = LoadLibrary(dllPath);
#else
   HMODULE handle = LoadLibrary(_dllname.c_str());
#endif
   if (handle == 0) return false;
   _modules[_dllname] = (BppIntPtr) handle;
   BppRegisterNativeModuleFunc regFunc = (BppRegisterNativeModuleFunc) GetProcAddress(handle,_PT("RegisterNativeModule"));
   if (regFunc != 0)
   {
      regFunc(this);
      return true;
   }
   return false;
}


struct DbgHelper
{
   DbgHelper(BppVirtualMachine* vm):_vm(vm) { _vm->setInDbg(true); }
   ~DbgHelper() { _vm->setInDbg(false); }
   BppVirtualMachine* _vm;
};

bool BppVirtualMachine::dbgWait() const
{
   DbgHelper dh((BppVirtualMachine*) this);
   if (_dbgWaitProc != 0) return _dbgWaitProc();
#ifdef WIN32
   if (_hwndDbg != 0)
   {
      if (SendMessage(_hwndDbg,_dbgWaitMsg,(WPARAM)this,0) == _dbgWaitMsg)
         return true;
   }
#endif
   return false;
}
BppDbgWaitProc BppVirtualMachine::setDbgWaitProc(BppDbgWaitProc dbgWaitProc)
{
   BppDbgWaitProc oldProc = _dbgWaitProc;
   _dbgWaitProc = dbgWaitProc;
   return oldProc;
}



bool BppVirtualMachine::breakRequest() const
{
   DbgHelper dh((BppVirtualMachine*) this);
   if (_breakRequestProc != 0) return _breakRequestProc();
#ifdef WIN32
   if (_hwndDbg != 0)
   {
      if (SendMessage(_hwndDbg,_breakRequestMsg,(WPARAM)this,0) == _breakRequestMsg)
         return true;
   }
#endif
   return false;
}

bool BppVirtualMachine::notifySourceLineChanged(const BppChar* sourceName, long line)
{
   DbgHelper dh((BppVirtualMachine*) this);
   _currentSource = sourceName;
   _currentLine = line;
   if (_lineChangedProc != 0)
   {
      if (_lineChangedProc(_currentSource.a_str(),line))
         return true;
   }
#ifdef WIN32
   else if (_hwndDbg != 0)
   {
      if (SendMessage(_hwndDbg,_lineChangedMsg,(WPARAM)this,(LPARAM)line) == _lineChangedMsg)
         return true;
   }
#endif
   return _interpreter->_step;
}

BppBreakRequestProc BppVirtualMachine::setBreakRequestProc( BppBreakRequestProc breakRequestProc )
{
   BppBreakRequestProc oldProc = _dbgWaitProc;
   _breakRequestProc = breakRequestProc;
   if (_interpreter != 0)
      _interpreter->setAllowBreakRequest(_breakRequestProc != 0);
   return oldProc;
}

BppLineChangedProc BppVirtualMachine::setLineChangedProc(BppLineChangedProc lineChangedProc)
{
   BppLineChangedProc oldProc = _lineChangedProc;
   _lineChangedProc = lineChangedProc;
   return oldProc;
}

IBppInput* BppVirtualMachine::getCurrentInput() const
{
   return _currentInput;
}

IBppInput* BppVirtualMachine::setCurrentInput( IBppInput* currentInput )
{
    IBppInput* oldInput = _currentInput;
    if (currentInput != 0)
       _currentInput = currentInput;
    else
       _currentInput = &_defaultInput;
    return oldInput;
}
#ifdef WIN32
void BppVirtualMachine::setDbgWindow( HWND hwndDbg )
{
   _hwndDbg = hwndDbg;
   if (_interpreter != 0)
      _interpreter->setAllowBreakRequest(_hwndDbg != 0);
}

UINT BppVirtualMachine::getBreakRequestMsg()
{
   return _breakRequestMsg;
}

UINT BppVirtualMachine::getDbgWaitMsg()
{
   return _dbgWaitMsg;
}

UINT BppVirtualMachine::getLineChangedMsg()
{
   return _lineChangedMsg;
}

#endif
void BppVirtualMachine::cancel()
{
   if (_interpreter != 0)
      _interpreter->cancel();
}

bool BppVirtualMachine::invoke(BppValue& result, const BppChar* name, BppVector* args)
{
   if (!_interpreter) return false;
   return _interpreter->invoke(result,name,args);
}

bool BppVirtualMachine::invokeMember(BppValue& result, BppObject* pobj, const BppChar* name, BppVector* args)
{
  if (!_interpreter) return false;
  return _interpreter->invokeMember(result,pobj,name,args);
}

bool BppVirtualMachine::invokeStatic(BppValue& result, BppClass* pclass, const BppChar* name, BppVector* args)
{
  if (!_interpreter) return false;
  return _interpreter->invokeStatic(result,pclass,name,args);
}

bool BppVirtualMachine::isInDbg() const
{
   return _inDbg;
}

void BppVirtualMachine::setInDbg(bool inDbg)
{
   _inDbg = inDbg;
}

// ==============================================================================
//         C-API implementation
// ==============================================================================


void* BPPAPI Bpp_CreateVM(void* pPrinter)
{
   return new BppVirtualMachine((IBppPrinter*) pPrinter);
}

int BPPAPI Bpp_ReleaseVM(void* pVM)
{
   BppVirtualMachine* p = (BppVirtualMachine*) pVM;
   if (p == 0)
      return 0;
   delete p;
   return 1;
}

int BPPAPI Bpp_GetGenerateLineInfo(void* pVM)
{
   BppVirtualMachine* p = (BppVirtualMachine*) pVM;
   if (p == 0)
      return 0;
   return (p->getGenerateLineInfo()) ? 1 : 0;
}

 /// Set value of line info generation flag.
 void BPPAPI Bpp_SetGenerateLineInfo(void* pVM, int bGenInfo)
 {
    BppVirtualMachine* p = (BppVirtualMachine*) pVM;
   if (p == 0)
      return;
   p->setGenerateLineInfo((0 != bGenInfo));
}

 int BPPAPI Bpp_GetGenerateLocalVarInfo(void* pVM)
{
   BppVirtualMachine* p = (BppVirtualMachine*) pVM;
   if (p == 0)
      return 0;
   return (p->getGenerateLocalVarInfo()) ? 1 : 0;
}

 void BPPAPI Bpp_SetGenerateLocalVarInfo(void* pVM, int bGenInfo)
 {
    BppVirtualMachine* p = (BppVirtualMachine*) pVM;
   if (p == 0)
      return;
   p->setGenerateLocalVarInfo((0 != bGenInfo));
}
int BPPAPI Bpp_CompileFileA(void* pVM,
                            char* pcszFileName, 
                            int bShowInfo, 
                            int bDecode, 
                            void* pErrPrinter, 
                            void* pDbgPrinter)
{
   if (pVM == 0) return 0;
   BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
   BppString fName = (const char*) pcszFileName;
   return (pVirt->compileFile(fName.c_str(),bShowInfo != 0,bDecode != 0,(IBppPrinter*) pErrPrinter,(IBppPrinter*)pDbgPrinter)) ? 1 : 0;
}

#ifndef NO_TCHAR_H
int BPPAPI Bpp_CompileFileW(void* pVM,
                            wchar_t* pcszFileName, 
                            int bShowInfo, 
                            int bDecode, 
                            void* pErrPrinter, 
                            void* pDbgPrinter)
{
   if (pVM == 0) return 0;
   BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
   BppString fName = (const wchar_t*) pcszFileName;
   return (pVirt->compileFile(fName.c_str(),bShowInfo != 0,bDecode != 0,(IBppPrinter*) pErrPrinter,(IBppPrinter*)pDbgPrinter)) ? 1 : 0;
}
#endif

int BPPAPI Bpp_CompileTextA(void* pVM,
                            char* pcszSrc, 
                            int bShowInfo, 
                            int bDecode, 
                            void* pErrPrinter, 
                            void* pDbgPrinter)
{
   if (pVM == 0) return 0;
   BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
   BppString src = (const char*) pcszSrc;
   return (pVirt->compileText(src.c_str(),bShowInfo != 0,bDecode != 0,(IBppPrinter*) pErrPrinter,(IBppPrinter*)pDbgPrinter)) ? 1 : 0;
}
#ifndef NO_TCHAR_H
int BPPAPI Bpp_CompileTextW(void* pVM,
                            wchar_t* pcszSrc, 
                            int bShowInfo, 
                            int bDecode, 
                            void* pErrPrinter, 
                            void* pDbgPrinter)
{
   if (pVM == 0) return 0;
   BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
   BppString src = (const wchar_t*)pcszSrc;
   return (pVirt->compileText(src.c_str(),bShowInfo != 0,bDecode != 0,(IBppPrinter*) pErrPrinter,(IBppPrinter*)pDbgPrinter)) ? 1 : 0;
}
#endif
struct BppCharPtrArray
{
   BppChar** arrayPtr;
   int cnt;
   BppCharPtrArray(int argc, char** argv)
      :arrayPtr(0),cnt(argc)
   {
      if (argc > 0)
      {
         arrayPtr = (BppChar**) allocmem(argc * sizeof(BppChar*));
         for (int i=0;i<argc;i++)
         {
#ifndef _UNICODE
            arrayPtr[i] = strdup(argv[i]);
#else
            BppString s = (const char*)argv[i];
            arrayPtr[i] = _wcsdup(s.w_str());
#endif
         }
      }
   }

#ifndef NO_TCHAR_H
   BppCharPtrArray(int argc, wchar_t** argv)
      :arrayPtr(0),cnt(argc)
   {
      if (argc > 0)
      {
         arrayPtr = (BppChar**) allocmem(argc * sizeof(BppChar*));
         for (int i=0;i<argc;i++)
         {
#ifdef _UNICODE
            arrayPtr[i] = _wcsdup(argv[i]);
#else
            BppString s = (const wchar_t*)argv[i];
            arrayPtr[i] = strdup(s.a_str());
#endif
         }
      }
   }
#endif // NO_TCHAR_H

   ~BppCharPtrArray()
   {
      if (arrayPtr != 0)
      {
         for (int i=0;i<cnt;i++)
            freemem(arrayPtr[i]);
         freemem(arrayPtr);
      }
   }

   operator const BppChar**() const
   {
      return (const BppChar**) arrayPtr;
   }
};

int BPPAPI Bpp_ExecuteA(void* pVM,
                        char* pcszFuncName, 
                        int iArgc, 
                        char** ppcszArgv,
                        int bTrace,
                        void* pErrPrinter, 
                        void* pDbgPrinter)
{
   if (pVM == 0) return 0;
   BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
   BppString funcname = (const char*) pcszFuncName;
   BppCharPtrArray args(iArgc,ppcszArgv);
   return (pVirt->execute(funcname.c_str(),iArgc,args,bTrace != 0,(IBppPrinter*)pErrPrinter,(IBppPrinter*)pDbgPrinter)) ? 1 : 0;
}

#ifndef NO_TCHAR_H
int BPPAPI Bpp_ExecuteW(void* pVM,
                        wchar_t* pcszFuncName, 
                        int iArgc, 
                        wchar_t** ppcszArgv,
                        int bTrace,
                        void* pErrPrinter, 
                        void* pDbgPrinter)
{
   if (pVM == 0) return 0;
   BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
   BppString funcname = (const wchar_t*) pcszFuncName;
   BppCharPtrArray args(iArgc,ppcszArgv);
   return (pVirt->execute(funcname.c_str(),iArgc,args,bTrace != 0,(IBppPrinter*)pErrPrinter,(IBppPrinter*)pDbgPrinter)) ? 1 : 0;
}
#endif

int BPPAPI Bpp_WriteModuleA(void* pVM,
                            char* pcszFileName)
{
   if (pVM == 0) return 0;
   BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
   BppString fName = (const char*) pcszFileName;
   return (pVirt->writeModule(fName.c_str())) ? 1 : 0;
}

#ifndef NO_TCHAR_H
int BPPAPI Bpp_WriteModuleW(void* pVM,
                            wchar_t* pcszFileName)
{
   if (pVM == 0) return 0;
   BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
   BppString fName = (const wchar_t*) pcszFileName;
   return (pVirt->writeModule(fName.c_str())) ? 1 : 0;
}
#endif

int BPPAPI Bpp_ReadModuleA(void* pVM,
                           char* pcszFileName)
{
   if (pVM == 0) return 0;
   BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
   BppString fName = (const char*) pcszFileName;
   return (pVirt->readModule(fName.c_str())) ? 1 : 0;
}

#ifndef NO_TCHAR_H
int BPPAPI Bpp_ReadModuleW(void* pVM,
                           wchar_t* pcszFileName)
{
   if (pVM == 0) return 0;
   BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
   BppString fName = (const wchar_t*) pcszFileName;
   return (pVirt->readModule(fName.c_str())) ? 1 : 0;
}
#endif

char* BPPAPI Bpp_GetCurrentSourceA(void* pVM)
{
   if (pVM == 0) return 0;
   BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
   return (char*) pVirt->getCurrentSource().a_str();
}

#ifndef NO_TCHAR_H
wchar_t* BPPAPI Bpp_GetCurrentSourceW(void* pVM)
{
   if (pVM == 0) return 0;
   BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
   return (wchar_t*) pVirt->getCurrentSource().w_str();
}
#endif

long BPPAPI Bpp_GetCurrentLine(void* pVM)
{
   if (pVM == 0) return -1;
   BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
   return pVirt->getCurrentLine();
}

 int BPPAPI Bpp_GetIntResult(void* pVM)
 {
    if (pVM == 0) return INT_MIN;
   BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
   return (int) (BppInt) pVirt->getResult();
}

 void BPPAPI Bpp_Reinit(void* pVM)
 {
     if (pVM == 0) return;
   BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
   pVirt->reinit();
}

 void* BPPAPI Bpp_CreateConsolePrinter()
 {
    return new BppFileStreamPrinter(stdout);
 }

 void* BPPAPI Bpp_CreateCallbackPrinter(BppPrinterCallback pCallback)
 {
    if (pCallback == 0) return 0;
    return new BppCallbackPrinter(pCallback);
 }

int BPPAPI Bpp_ReleasePrinter(void* pPrinter)
{
   if (pPrinter == 0) return 0;
   delete (IBppPrinter*) pPrinter;
   return 1;
}

BppDbgWaitProc BPPAPI Bpp_SetDbgWaitProc(void* pVM, BppDbgWaitProc dbgWaitProc)
{
   if (pVM == 0) return 0;
   BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
   return pVirt->setDbgWaitProc(dbgWaitProc);
}

BppBreakRequestProc BPPAPI Bpp_SetBreakRequestProc( void* pVM, BppBreakRequestProc breakRequestProc )
{
   if (pVM == 0) return 0;
   BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
   return pVirt->setBreakRequestProc(breakRequestProc);
}

BppLineChangedProc BPPAPI Bpp_SetLineChangedProc( void* pVM, BppLineChangedProc lineChangedProc )
{
   if (pVM == 0) return 0;
   BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
   return pVirt->setLineChangedProc(lineChangedProc);
}

void* BPPAPI Bpp_CreateCallbackInput( BppCharInputCallback pcCallback, BppStringInputCallback psCallback )
{
    return new BppCallbackInput(pcCallback,psCallback);
}

int BPPAPI Bpp_ReleaseInput( void* pInput )
{
    if (pInput == 0) return 0;
    delete (IBppInput*) pInput;
    return 1;
}

void* BPPAPI Bpp_SetCurrentInput( void* pVM, void* pInput )
{
   if (pVM == 0) return 0;
   BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
   return pVirt->setCurrentInput((IBppInput*) pInput);
}

void* BPPAPI Bpp_GetCurrentInput( void* pVM )
{
   if (pVM == 0) return 0;
   BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
   return pVirt->getCurrentInput();
}

void BPPAPI Bpp_Cancel( void* pVM )
{
   BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
   if (pVirt !=0)
      pVirt->cancel();
}

#ifdef WIN32
void* BPPAPI Bpp_CreateWinMsgPrinter( HWND hwndDest )
{
   return new BppWinMsgPrinter(hwndDest);
}

UINT BPPAPI Bpp_GetPrintStringMsg()
{
   return BppWinMsgPrinter::getPrintStringMsg();
}

void* BPPAPI Bpp_CreateWinMsgInput( HWND hwndDest )
{
   return new BppWinMsgInput(hwndDest);
}

UINT BPPAPI Bpp_GetStringInputMsg()
{
   return BppWinMsgInput::getStringInputMsg();
}

UINT BPPAPI Bpp_GetCharInputMsg()
{
   return BppWinMsgInput::getCharInputMsg();
}

void BPPAPI Bpp_SetWinInputStringA( void* pMsgWinInput, char* pcszInputStr )
{
   if (pMsgWinInput == 0) return;
   BppWinMsgInput* p = (BppWinMsgInput*) pMsgWinInput;
   p->setString(pcszInputStr);
}

void BPPAPI Bpp_SetWinInputStringW( void* pMsgWinInput, wchar_t* pcszInputStr )
{
   if (pMsgWinInput == 0) return;
   BppWinMsgInput* p = (BppWinMsgInput*) pMsgWinInput;
   p->setString(pcszInputStr);
}

void BPPAPI Bpp_SetWinInputCharA( void* pMsgWinInput, char cInputChr )
{
   if (pMsgWinInput == 0) return;
   BppWinMsgInput* p = (BppWinMsgInput*) pMsgWinInput;
   p->setChar(cInputChr);
}

void BPPAPI Bpp_SetWinInputCharW( void* pMsgWinInput, wchar_t cInputChr )
{
   if (pMsgWinInput == 0) return;
   BppWinMsgInput* p = (BppWinMsgInput*) pMsgWinInput;
   p->setChar(cInputChr);
}

void BPPAPI Bpp_SetDbgWindow( void* pVM, HWND hwndDbg )
{
   if (pVM == 0) return ;
   BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
   pVirt->setDbgWindow(hwndDbg);
}

UINT BPPAPI Bpp_GetBreakRequestMsg()
{
   return BppVirtualMachine::getBreakRequestMsg();
}

UINT BPPAPI Bpp_GetDbgWaitMsg()
{
   return BppVirtualMachine::getDbgWaitMsg();
}

UINT BPPAPI Bpp_GetLineChangedMsg()
{
   return BppVirtualMachine::getLineChangedMsg();
}

char* BPPAPI Bpp_GetEnvA(char* pcszName)
{
   BppString nbuf = pcszName;
   static BppString buf = _tgetenv(nbuf.c_str());
   return (char*) buf.a_str();
}

void BPPAPI Bpp_SetEnvA(char* pcszName,char* pcszValue)
{
   BppString buf = pcszName;
   buf += _T('=');
   if (pcszValue != 0) buf += pcszValue;
   _tputenv(buf.c_str());
}

#ifndef NO_TCHAR_H
wchar_t* BPPAPI Bpp_GetEnvW(wchar_t* pcszName)
{
   BppString nbuf = pcszName;
   static BppString buf = _tgetenv(nbuf.c_str());
   return (wchar_t*) buf.w_str();
}

void BPPAPI Bpp_SetEnvW(wchar_t* pcszName,wchar_t* pcszValue)
{
   BppString buf = pcszName;
   buf += _T('=');
   if (pcszValue != 0) buf += pcszValue;
   _tputenv(buf.c_str());
}
#endif

#endif // WIN32
