#ifndef RATIONAL_H
#define RATIONAL_H

// Macro defines rint and ruint types (T may be any integer type, e.g. char, short, long or long long)
#define RINTTYPE(T) \
typedef T rint;\
typedef unsigned T ruint;


RINTTYPE(long long)

class rational
{
public:
	/// Initialization from integer type value(s).
	rational(rint p, ruint q=1);
	///Initialization from double value.
	rational(double d);
	/// Initialization from character string.
    rational(const char* str);
	/// Copy ctor
    rational(const rational& cpy);
	// dtor
	~rational();
	/// assignment
	rational& operator=(const rational& cpy);
    /// Conversion to integer type
	operator rint() const ;
	/// Conversion to double
	operator double() const;
	/// Conversion to character string.
	operator const char*() const;
	/// Conversion to string in floating point format.
    const char* fltStr(size_t maxdec = sizeof(rint)) const; 
	/// addition/assignment
    rational& operator+=(const rational& r);
	/// subtraction/assignment
	rational& operator-=(const rational& r);
	/// multiplication/assignment
	rational& operator*=(const rational& r);
	/// division/assignment
	rational& operator/=(const rational& r);
	/// comparison method
    rint compare(const rational& cmp) const;
	/// Computes own value power to n.
    rational pow_n(unsigned n) const;
	/// Computes square root.
    rational sqrt() const;
	/// Computes n-th root.
    rational root(unsigned n) const;
	/// Gets absolute value.
	rational abs() const;
	/// Reduced internal representation.
	rational reduce(size_t maxdec = sizeof(rint)) const;
 private:
	 /// greatest common divisor of two integers
    static ruint gcd(ruint a, ruint b);
	 /// lowest common multiple of two integers
    static ruint lcm(ruint a, ruint b);

	/// Normalizes internal representation.
	void norm();
	/// nominator
	rint _nom;
	/// denominator
	ruint _denom;
	/// Buffer for building string representations.
	char* _buf;
};

/// addition
rational operator+(const rational& l, const rational& r);
/// subtraction
rational operator-(const rational& l, const rational& r);
/// multiplication
rational operator*(const rational& l, const rational& r);
/// division
rational operator/(const rational& l, const rational& r);
/// equal
bool operator==(const rational& l, const rational& r);
/// not equal
bool operator!=(const rational& l, const rational& r);
/// lower or equal
bool operator<=(const rational& l, const rational& r);
/// greater or equal
bool operator>=(const rational& l, const rational& r);
/// lower than
bool operator<(const rational& l, const rational& r);
/// greater than
bool operator>(const rational& l, const rational& r);

#endif // RATIONAL.H