/*
 * Copyright (c) 2001-2002 Daniel van Vugt <danv@users.sf.net>
 * This file is part of the Rex library.
 * See LICENSE.txt for license details.
 */

#include "rexprivate.h"
#include "platform.h"
#include <stddef.h>


static blockid_t
GrowAutomaton(automaton_t **pa, int newblocks)
{
	automaton_t *a;
	int used;
	blockid_t id;

	a = *pa;
	used = a->used + newblocks;
	if (used > a->allocated)
	{
		int allocation;
		size_t size;

		allocation = a->allocated << 1;
		while ( allocation <= used )
		{
			allocation <<= 1;
		}

		size = sizeof(automaton_t) +
			(sizeof(block_t) * (allocation - DEFAULT_BLOCKS));
		if (size > rexmemlimit)
		{
			id = INVALID_BLOCKID;
		}
		else
		{
			void *newa;

			newa = REALLOC(a, size);

			if (newa == NULL)
			{
				id = INVALID_BLOCKID;
			}
			else
			{
				a = (automaton_t*)newa;
				*pa = a;
				a->allocated = allocation;
				id = a->used;
				a->used = used;
			}
		}
	}
	else
	{
		id = a->used;
		a->used = used;
	}

	return id;
}


static stateid_t
NewState(automaton_t **pa, int ntrans, size_t operandsize)
{
	blockid_t id;
	int opblocks;

	ASSERT(ntrans > 0);
	opblocks = (operandsize + sizeof(block_t) - 1) / sizeof(block_t);
	id = GrowAutomaton(pa, 1 + ntrans + opblocks);
	if (id != INVALID_BLOCKID)
	{
		state_t *state;
		transition_t *trans;
		block_t *block;
		int i;

		block = (*pa)->block;
		state = &block[id].state;
		(*pa)->maxstate = id;

		state->header.format = FORMAT_STATE;
		state->header.span = 1;
		trans = NULL; /* just to stop compiler warnings */
		state->lasttrans = id + ntrans;
		i = 1;
		while (i <= ntrans)
		{
			trans = &block[id+i].trans;
			trans->header.format = FORMAT_TRANS;
			trans->header.span = 1;
			trans->test = UNUSED;
			trans->dest = INVALID_BLOCKID;
			i++;
			trans->sibling = id + i;
		}
		trans->sibling = INVALID_BLOCKID;

		block[id+ntrans].header.span = opblocks + 1;
	}

	return id;
}


static transitionid_t
NewTransition(automaton_t **pa, stateid_t source, size_t operandsize)
{
	transitionid_t tid;
	int blocks;

	ASSERT(source != INVALID_BLOCKID);

	blocks = 1 + (operandsize + sizeof(block_t) - 1) / sizeof(block_t);
	tid = GrowAutomaton(pa, blocks);
	if (tid != INVALID_BLOCKID)
	{
		state_t *state;
		block_t *block;
		transition_t *trans;

		block = (*pa)->block;
		state = &block[source].state;
		ASSERT(state->lasttrans != INVALID_BLOCKID);
		block[state->lasttrans].trans.sibling = tid;
		state->lasttrans = tid;

		trans = &block[tid].trans;
		trans->header.format = FORMAT_TRANS;
		trans->header.span = blocks;
		trans->sibling = INVALID_BLOCKID;
		trans->test = UNUSED;
		trans->dest = INVALID_BLOCKID;
	}

	return tid;
}


static stateid_t
Duplicate(automaton_t **pa, blockid_t min, blockid_t max, stateid_t head,
		transitionid_t tail, int count, transitionid_t *newtail)
{
	blockid_t growth;
	blockid_t dest;
	int instance;
	int stride0;
	int headoffset, tailoffset;
	block_t *block;

	ASSERT( (head >= min) &&
		(head <= max) &&
		(tail >= min) &&
		(tail <= max) );

	stride0 = max - min + 1;
	growth = GrowAutomaton(pa, stride0 * count);
	if (growth == INVALID_BLOCKID)
	{
		return INVALID_BLOCKID;
	}
	block = (*pa)->block;

	headoffset = head - min;
	tailoffset = tail - min;
	dest = growth;

	for (instance=0; instance < count; instance++)
	{
		blockid_t src;
		int stride;

		src = min;
		stride = dest - src;

		while (src <= max)
		{
			int span;

			span = block[src].header.span;
			MEMCPY(block+dest, block+src, sizeof(block_t) * span);

			if (block[dest].header.format & FORMAT_STATE)
			{
				block[dest].state.lasttrans += stride;
				(*pa)->maxstate = dest;
			}

			if (block[dest].header.format & FORMAT_TRANS)
			{
				transition_t *trans;

				trans = &block[dest].trans;
				if ( (trans->sibling >= min) &&
				     (trans->sibling <= max) )
				{
					trans->sibling += stride;
				}
				if ( (trans->dest >= min) &&
				     (trans->dest <= max) )
				{
					trans->dest += stride;
				}
			}

			src += span;
			dest += span;
		}

		if (instance < (count-1))
		{
			transitionid_t tid;

			tid = dest - stride0 + tailoffset;
			ASSERT( block[tid].trans.dest == INVALID_BLOCKID );
			block[tid].trans.dest = dest + headoffset;
		}
	}

	*newtail = dest - stride0 + tailoffset;
	return growth + headoffset;
}


static 
parseresult_t
Compile(automaton_t **pa,
        const REXCHAR *pattern,
        int len,
        unsigned int options,
        int depth,
        stateid_t *head,
        transitionid_t *tail
        )
{
	int i;
	stateid_t chainhead, subhead;
	transitionid_t chaintail, subtail;
	blockid_t submin;
	int mincount, maxcount;

	if (len <= 0)
	{
		*head = INVALID_BLOCKID;
		*tail = INVALID_BLOCKID;
		return 0;
	}
	
	chainhead = INVALID_BLOCKID;
	subhead = INVALID_BLOCKID;
	chaintail = INVALID_BLOCKID;
	subtail = INVALID_BLOCKID;
	submin = INVALID_BLOCKID;

	i = 0;
	while (i < len)
	{
		REXCHAR c;
		int consumed;
		int nexti;
		bool_t isliteral;
		bool_t postfixed;
		blockid_t newsubmin;

		c = pattern[i];
		consumed = 0;
		isliteral = FALSE;
		newsubmin = INVALID_BLOCKID;
		     
		if (c == '.')
		{
			subhead = NewState(pa, 1, 0);
			if (subhead == INVALID_BLOCKID)
			{
				return ENCODE_PARSE_ERROR(REX_MEM, i);
			}
			subtail = FIRSTTRANSID(subhead);
			(*pa)->block[subtail].trans.test = DOT;
			consumed = 1;
		}
		else if (c == '*')
		{
			block_t *block;

			if (subhead == INVALID_BLOCKID)
			{
				return ENCODE_PARSE_ERROR(REX_POSTFIX, i);
			}

			consumed = 1;
			block = (*pa)->block;
			block[subtail].trans.dest = subhead;

			subtail = NewTransition(pa, subhead, 0);
			if (subtail == INVALID_BLOCKID)
			{
				return ENCODE_PARSE_ERROR(REX_MEM, i);
			}
			(*pa)->block[subtail].trans.test = NOP;
		}
		else if (c == '+')
		{
			stateid_t sid;
			block_t *block;
			transition_t *trans;
			transitionid_t tid;

			if (subhead == INVALID_BLOCKID)
			{
				return ENCODE_PARSE_ERROR(REX_POSTFIX, i);
			}
			consumed = 1;

			sid = NewState(pa, 2, 0);
			if (sid == INVALID_BLOCKID)
			{
				return ENCODE_PARSE_ERROR(REX_MEM, i);
			}

			block = (*pa)->block;
			block[subtail].trans.dest = sid;
			tid = FIRSTTRANSID(sid);
			trans = &block[tid].trans;
			trans->test = NOP;
			trans->dest = subhead;
			subtail = tid + 1;
			block[subtail].trans.test = NOP;
		}
		else if (c == '?')
		{
			stateid_t sid;
			transitionid_t tid;
			block_t *block;
			transition_t *trans;

			if (subhead == INVALID_BLOCKID)
			{
				return ENCODE_PARSE_ERROR(REX_POSTFIX, i);
			}

			consumed = 1;

			sid = NewState(pa, 1, 0);
			if (sid == INVALID_BLOCKID)
			{
				return ENCODE_PARSE_ERROR(REX_MEM, i);
			}

			block = (*pa)->block;
			block[subtail].trans.dest = sid;
			subtail = FIRSTTRANSID(sid);
			block[subtail].trans.test = NOP;
			
			tid = NewTransition(pa, subhead, 0);
			if (tid == INVALID_BLOCKID)
			{
				return ENCODE_PARSE_ERROR(REX_MEM, i);
			}
			trans = &(*pa)->block[tid].trans;
			trans->test = NOP;
			trans->dest = sid;
		}
		else if (c == '^')
		{
			subhead = NewState(pa, 1, 0);
			if (subhead == INVALID_BLOCKID)
			{
				return ENCODE_PARSE_ERROR(REX_MEM, i);
			}
			subtail = FIRSTTRANSID(subhead);
			(*pa)->block[subtail].trans.test = START_OF_LINE;
			consumed = 1;
		}
		else if (c == '$')
		{
			subhead = NewState(pa, 1, 0);
			if (subhead == INVALID_BLOCKID)
			{
				return ENCODE_PARSE_ERROR(REX_MEM, i);
			}
			subtail = FIRSTTRANSID(subhead);
			(*pa)->block[subtail].trans.test = END_OF_LINE;
			consumed = 1;
		}
		else if (c == '[')
		{
			parseresult_t sub;
			charset_t *set;
			transition_t *trans;
			int j;

			subhead = NewState(pa, 1, sizeof(charset_t));
			if (subhead == INVALID_BLOCKID)
			{
				return ENCODE_PARSE_ERROR(REX_MEM, i);
			}
			subtail = FIRSTTRANSID(subhead);

			trans = &(*pa)->block[subtail].trans;
			trans->test = CHARSET;
			set = (charset_t*)(&OPERANDOF(trans)->charset);
			sub = InitCharSet(set, pattern+i+1, len-i-1, options);
			if ( PARSE_FAILED(sub) )
			{
				return sub + i + 1;
			}
			consumed = 1 + sub;

			j = i + consumed;
			if ( (j < len) ? (pattern[j] != ']') : TRUE )
			{
				return ENCODE_PARSE_ERROR(REX_OPEN_CHARSET, j);
			}
			consumed++;
		}
		else if (c == '|')
		{
			parseresult_t sub;
			stateid_t split, merge;
			block_t *block;
			transition_t *left, *right;
			transitionid_t tid;

			if (chainhead == INVALID_BLOCKID)
			{
				return ENCODE_PARSE_ERROR(REX_EMPTY, i);
			}
			
			newsubmin = (*pa)->used;
			sub = Compile(pa, pattern+i+1, len-i-1,
				options, depth, &subhead, &subtail);
			if ( PARSE_FAILED(sub) )
			{
				return sub + i + 1;
			}
			consumed = 1 + sub;

			if (sub == 0)
			{
				return ENCODE_PARSE_ERROR(REX_EMPTY, i+1);
			}

			split = NewState(pa, 2, 0);
			merge = NewState(pa, 1, 0);
			if ( (split == INVALID_BLOCKID) ||
			     (merge == INVALID_BLOCKID) )
			{
				return ENCODE_PARSE_ERROR(REX_MEM, i);
			}
			block = (*pa)->block;
			tid = FIRSTTRANSID(split);
			left = &block[tid].trans;
			left->test = NOP;
			left->dest = chainhead;
			right = &block[tid+1].trans;
			right->test = NOP;
			right->dest = subhead;

			block[chaintail].trans.dest = merge;
			block[subtail].trans.dest = merge;
			chaintail = FIRSTTRANSID(merge);
			block[chaintail].trans.test = NOP;
			chainhead = split;

			i += consumed;
			continue;  /* No chaining to do! */
		}
		else if (c == '(')
		{
			parseresult_t sub;
			stateid_t newhead;
			transition_t *trans;
			int j;

			newsubmin = (*pa)->used;
			sub = Compile(pa, pattern+i+1, len-i-1,
				options, depth+1, &subhead, &subtail);
			if ( PARSE_FAILED(sub) )
			{
				return sub + i + 1;
			}

			consumed = 1 + sub;
			j = i + consumed;
			if ( (j < len) ? (pattern[j] != ')') : TRUE )
			{
				return ENCODE_PARSE_ERROR(REX_OPEN_PAREN, j);
			}
			consumed++;

			if (sub == 0)
			{
				return ENCODE_PARSE_ERROR(REX_EMPTY, i+1);
			}

			/*
			 * Add an extra null transition on the front to prevent
			 * loop bugs like:
			 *    (a*b)*  matching the whole string  abaabaa
			 *            while it should match only abaab
			 */
			newhead = NewState(pa, 1, 0);
			if (newhead == INVALID_BLOCKID)
			{
				return ENCODE_PARSE_ERROR(REX_MEM, i);
			}
			trans = &(*pa)->block[FIRSTTRANSID(newhead)].trans;
			trans->test = NOP;
			trans->dest = subhead;
			subhead = newhead;
		}
		else if (c == ')')
		{
			if (depth > 0)
			{
				consumed = 0;
				break;
			}
			else
			{
				return ENCODE_PARSE_ERROR(REX_UNMATCHED_CLOSE, i);
			}
		}
		else if (c == '{')
		{
			parseresult_t sub;
			blockid_t submax;
			int j;

			if (subhead == INVALID_BLOCKID)
			{
				return ENCODE_PARSE_ERROR(REX_POSTFIX, i);
			}

			sub = ParseBound(pattern+i+1, len-i-1,
				&mincount, &maxcount);
			if ( PARSE_FAILED(sub) )
			{
				return sub + i + 1;
			}

			consumed = 1 + sub;
			j = i + consumed;
			if ( (j < len) ? (pattern[j] != '}') : TRUE )
			{
				return ENCODE_PARSE_ERROR(REX_OPEN_BOUND, j);
			}
			consumed++;

			submax = (*pa)->used - 1;

			if (maxcount >= 0)
			{
				transitionid_t newtail;
				
				if (mincount > 1)
				{
					stateid_t newhead;

					newhead = Duplicate(pa, submin, submax,
							subhead, subtail,
							mincount-1, &newtail);
					if (newhead == INVALID_BLOCKID)
					{
						return ENCODE_PARSE_ERROR(REX_MEM, i);
					}
					(*pa)->block[subtail].trans.dest =
						newhead;
				}
				else
				{
					newtail = subtail;
				}

				if (maxcount > mincount)
				{
					transitionid_t exittid, tmptail;
					stateid_t sid, tmphead;
					blockid_t tmpmin, tmpmax;
					transition_t *trans;

					if (mincount == 0)
					{
						tmpmin = submin;
						tmphead = subhead;
						tmptail = subtail;
					}
					else
					{
						tmpmin = (*pa)->used;
						tmphead = Duplicate(pa, submin,
								submax, subhead,
								subtail, 1, &tmptail);
						if (tmphead == INVALID_BLOCKID)
						{
							return ENCODE_PARSE_ERROR(REX_MEM, i);
						}
						(*pa)->block[newtail].trans.dest = tmphead;
					}
					exittid = NewTransition(pa, tmphead, 0);
					if (exittid == INVALID_BLOCKID)
					{
						return ENCODE_PARSE_ERROR(REX_MEM, i);
					}
					tmpmax = (*pa)->used - 1;

					sid = NewState(pa, 1, 0);
					if (sid == INVALID_BLOCKID)
					{
						return ENCODE_PARSE_ERROR(REX_MEM, i);
					}
	
					trans = &(*pa)->block[exittid].trans;
					trans->dest = sid;
					trans->test = NOP;
					newtail = tmptail;

					if ((maxcount - mincount) > 1)
					{
						tmphead = Duplicate(pa, tmpmin,
								tmpmax, tmphead,
								tmptail,
								maxcount-mincount-1,
								&tmptail);
						if (tmphead == INVALID_BLOCKID)
						{
							return ENCODE_PARSE_ERROR(REX_MEM, i);
						}
						(*pa)->block[newtail].trans.dest = tmphead;
						newtail = tmptail;
					}

					(*pa)->block[newtail].trans.dest = sid;
					newtail = FIRSTTRANSID(sid);
					(*pa)->block[newtail].trans.test = NOP;
				}
				else if (mincount == 0)
				{
					ASSERT(maxcount == 0);
					subhead = INVALID_BLOCKID;
					subtail = INVALID_BLOCKID;
				}

				subtail = newtail;
			}
			else
			{
				transitionid_t newtail;
				stateid_t tmphead, sid;
				transition_t *loop;

				if (mincount > 2)
				{
					stateid_t newhead;
					newhead = Duplicate(pa,
							submin, submax,
							subhead, subtail,
							mincount-2, &newtail);
					if (newhead == INVALID_BLOCKID)
					{
						return ENCODE_PARSE_ERROR(REX_MEM, i);
					}
					(*pa)->block[subtail].trans.dest =
						newhead;
				}
				else
				{
					newtail = subtail;
				}

				if (mincount > 1)
				{
					transitionid_t tmptail;

					tmphead = Duplicate(pa,
						submin, submax,
						subhead, subtail,
						1, &tmptail);
					if (tmphead == INVALID_BLOCKID)
					{
						return ENCODE_PARSE_ERROR(REX_MEM, i);
					}
					(*pa)->block[newtail].trans.dest =
						tmphead;
					newtail = tmptail;
				}
				else
				{
					tmphead = subhead;
				}

				sid = NewState(pa, 2, 0);
				if (sid == INVALID_BLOCKID)
				{
					return ENCODE_PARSE_ERROR(REX_MEM, i);
				}
				(*pa)->block[newtail].trans.dest = sid;
				loop = &(*pa)->block[FIRSTTRANSID(sid)].trans;
				loop->test = NOP;
				loop->dest = tmphead;
				newtail = FIRSTTRANSID(sid) + 1;
				(*pa)->block[newtail].trans.test = NOP;

				if (mincount == 0)
				{
					subhead = sid;
				}

				subtail = newtail;
			}
		}
		else if (c == '\\')
		{
			parseresult_t sub;

			if ( i < (len-1) )
			{
				test_t test;

				switch (pattern[i+1])
				{
					case '<': test = START_OF_WORD; break;
					case '>': test = END_OF_WORD; break;
					case 'b': test = WORD_BOUNDARY; break;
					case 's': test = WHITESPACE; break;
					case 'S': test = NOT_WHITESPACE; break;
					case 'w': test = WORD_CHAR; break;
					case 'W': test = NOT_WORD_CHAR; break;
					case 'd': test = DIGIT; break;
					case 'D': test = NOT_DIGIT; break;
					default: test = UNUSED; break;
				}

				if (test != UNUSED)
				{
					subhead = NewState(pa, 1, 0);
					if (subhead == INVALID_BLOCKID)
					{
						return ENCODE_PARSE_ERROR(REX_MEM, i);
					}
					subtail = FIRSTTRANSID(subhead);
					(*pa)->block[subtail].trans.test = test;
					consumed = 2;
				}
			}

			if (consumed == 0)
			{
				sub = EscapeChar(pattern+i+1, len-i-1, &c);
				if ( PARSE_FAILED(sub) )
				{
					return sub + i + 1;
				}
				consumed = 1 + sub;
				isliteral = TRUE;
			}
		}
		else
		{
			isliteral = TRUE;
			consumed = 1;
		}

		nexti = i + consumed;
		postfixed = FALSE;
		if (nexti < len)
		{
			REXCHAR nextc;

			nextc = pattern[nexti];
			if ( (nextc == '*') ||
			     (nextc == '+') ||
			     (nextc == '?') ||
			     (nextc == '{') )
			{
				postfixed = TRUE;
			}
		}

		if (isliteral)
		{
			transition_t *trans;

			if (options & REX_COMPILE_ICASE)
			{
				c = TOUPPER(c);
			}

			subhead = NewState(pa, 1, sizeof(operand_t));
			if (subhead == INVALID_BLOCKID)
			{
				return ENCODE_PARSE_ERROR(REX_MEM, i);
			}
			subtail = FIRSTTRANSID(subhead);

			trans = &(*pa)->block[subtail].trans;
			trans->test = LITERAL;
			trans->dest = INVALID_BLOCKID;
			OPERANDOF(trans)->literal = c;
		}

		submin = (newsubmin == INVALID_BLOCKID) ? subhead : newsubmin;
		
		if ( (subhead != INVALID_BLOCKID) && !postfixed )
		{
			if (chainhead == INVALID_BLOCKID)
			{
				chainhead = subhead;
				chaintail = subtail;
			}
			else
			{
				block_t *block;

				block = (*pa)->block;
				ASSERT( block[chaintail].trans.dest ==
				        INVALID_BLOCKID );
				block[chaintail].trans.dest = subhead;
				chaintail = subtail;
			}
		}

		i = nexti;
	}

	*head = chainhead;
	*tail = chaintail;

	return i;
}


REX_API(rex_t)
rexCompile(const char *pattern, size_t len, unsigned int options,
		rex_error_t *error, size_t *erroroffset)
{
	automaton_t *a;

	if (sizeof(automaton_t) > rexmemlimit)
	{
		/* pointless? maybe. */
		return REX_INVALID;
	}

	a = (automaton_t*)MALLOC(sizeof(automaton_t));
	if (a != NULL)
	{
		stateid_t head;
		transitionid_t tail;
		parseresult_t result;
		rex_error_t e;
		size_t off;
		
		a->options = options;
		a->maxstate = INVALID_BLOCKID;
		ASSERT(ACCEPT_STATE == 0);
		a->used = 1;
		
		a->allocated = DEFAULT_BLOCKS;
		result = Compile(&a, pattern, len, options, 0, &head, &tail);
		if ( PARSE_FAILED(result) )
		{
			DECODE_PARSE_ERROR(result, e, off);
			FREE(a);
			a = REX_INVALID;
		}
		else
		{
#if 0
			/*
			 * Freeing unused memory typically isn't worth the
			 * slowdown incurred by calling REALLOC.
			 */
			automaton_t *b;

			b = REALLOC(a, sizeof(automaton_t) +
				(sizeof(block_t) * (a->used - DEFAULT_BLOCKS)));
			if (b != NULL)
			{
				a = b;
				a->allocated = a->used;
			}
#endif
			if (tail != INVALID_BLOCKID)
			{
				transition_t *trans;

				trans = &a->block[tail].trans;
				ASSERT(trans->dest == INVALID_BLOCKID);
				trans->dest = ACCEPT_STATE;
			}

			a->startstate = head;
			e = REX_SUCCESS;
			off = 0;
		}
		if (error != NULL)
		{
			*error = e;
		}
		if (erroroffset != NULL)
		{
			*erroroffset = off;
		}
	}

	return (rex_t)a;
}


REX_API(void)
rexFree(rex_t rex)
{
	ASSERT(rex != REX_INVALID);
	FREE((automaton_t*)rex);
}
