/*
 * Copyright (c) 2001-2002 Daniel van Vugt <danv@users.sf.net>
 * This file is part of the Rex library.
 * See LICENSE.txt for license details.
 */

#include "rexprivate.h"
#include "platform.h"


static
bool_t
Matches(const char *templ, const REXCHAR *input, size_t inputlen)
{
	bool_t result;

	ASSERT(inputlen >= 5);

	result = (templ[0] == input[0]) &&
		(templ[1] == input[1]) &&
		(templ[2] == input[2]) &&
		(templ[3] == input[3]) &&
		(templ[4] == input[4]);

	if (result && (templ[5] != '\0'))
	{
		result = (templ[5] == input[5]);
	}
	return result;
}


static
size_t
GetCharClass(const REXCHAR *str, size_t len, charset_t *set)
{
	size_t result;

	result = 0;
	if (len >= 5)
	{
		int c, maxchar;

		maxchar = 255;  /* TODO: introduce unicode */

		if (Matches("alnum", str, len))
		{
			for (c=0; c <= maxchar; c++)
			{
				if (ISALNUM(c))
				{
					CHARSET_ADD(*set, c);
				}
			}
			result = 5;
		}
		else if (Matches("alpha", str, len))
		{
			for (c=0; c <= maxchar; c++)
			{
				if (ISALPHA(c))
				{
					CHARSET_ADD(*set, c);
				}
			}
			result = 5;
		}
		else if (Matches("blank", str, len))
		{
			for (c=0; c <= maxchar; c++)
			{
				if (ISBLANK(c))
				{
					CHARSET_ADD(*set, c);
				}
			}
			result = 5;
		}
		else if (Matches("cntrl", str, len))
		{
			for (c=0; c <= maxchar; c++)
			{
				if (ISCNTRL(c))
				{
					CHARSET_ADD(*set, c);
				}
			}
			result = 5;
		}
		else if (Matches("digit", str, len))
		{
			for (c=0; c <= maxchar; c++)
			{
				if (ISDIGIT(c))
				{
					CHARSET_ADD(*set, c);
				}
			}
			result = 5;
		}
		else if (Matches("graph", str, len))
		{
			for (c=0; c <= maxchar; c++)
			{
				if (ISGRAPH(c))
				{
					CHARSET_ADD(*set, c);
				}
			}
			result = 5;
		}
		else if (Matches("lower", str, len))
		{
			for (c=0; c <= maxchar; c++)
			{
				if (ISLOWER(c))
				{
					CHARSET_ADD(*set, c);
				}
			}
			result = 5;
		}
		else if (Matches("print", str, len))
		{
			for (c=0; c <= maxchar; c++)
			{
				if (ISPRINT(c))
				{
					CHARSET_ADD(*set, c);
				}
			}
			result = 5;
		}
		else if (Matches("punct", str, len))
		{
			for (c=0; c <= maxchar; c++)
			{
				if (ISPUNCT(c))
				{
					CHARSET_ADD(*set, c);
				}
			}
			result = 5;
		}
		else if (Matches("space", str, len))
		{
			for (c=0; c <= maxchar; c++)
			{
				if (ISSPACE(c))
				{
					CHARSET_ADD(*set, c);
				}
			}
			result = 5;
		}
		else if (Matches("upper", str, len))
		{
			for (c=0; c <= maxchar; c++)
			{
				if (ISUPPER(c))
				{
					CHARSET_ADD(*set, c);
				}
			}
			result = 5;
		}
		else if (Matches("xdigit", str, len))
		{
			for (c=0; c <= maxchar; c++)
			{
				if (ISXDIGIT(c))
				{
					CHARSET_ADD(*set, c);
				}
			}
			result = 6;
		}
	}
	return result;
}


parseresult_t
InitCharSet(charset_t *set, const REXCHAR *chars, size_t len, unsigned int options)
{
	int i, invert;
	UREXCHAR endchar;

	MEMSET(set->chunk, 0,
		sizeof(chunk_t) * CHARSET_SIZE);

	invert = (chars[0] == '^') ? 1 : 0;

	i = invert;
	while (i < len)
	{
		UREXCHAR thischar, startchar;
		int notlast, c;

		notlast = (i < (len-1));

		thischar = chars[i];
		if ( (thischar == ']') && (i > invert) )
		{
			break;
		}
		else if ( (thischar == '-') && (i > invert) && notlast )
		{
			int oldi;

			oldi = i;
			startchar = endchar;
			endchar = chars[i+1];

			if (endchar == ']')
			{
				startchar = endchar = '-';
				i++;
			}
			else
			{
				if (endchar == '\\')
				{
					parseresult_t consumed;

					consumed = EscapeChar(chars+i+2, len-i-2, (REXCHAR*)&endchar);
					if ( PARSE_FAILED(consumed) )
					{
						return consumed + i + 2;
					}
					i += 2 + consumed;
				}
				else
				{
					i += 2;
				}

				if (endchar < startchar)
				{
					return ENCODE_PARSE_ERROR(REX_INVALID_RANGE, oldi+1);
				}
				startchar++;  /* because it's already in the set */
			}
		}
		else if ( (thischar == '\\') && notlast )
		{
			parseresult_t consumed;

			consumed = EscapeChar(chars+i+1, len-i-1, (REXCHAR*)&endchar);
			if ( PARSE_FAILED(consumed) )
			{
				return consumed + i + 1;
			}
			startchar = endchar;
			i += 1 + consumed;
		}
		else if ( ((thischar == '[') && notlast) ? (chars[i+1] == ':') : FALSE )
		{	/* POSIX-style character class: [:blah:] */
			size_t got;
			
			got = GetCharClass(chars+i+2, len-i-2, set);
			i += 2 + got;
			if ( ((i >= (len-1)) || (!got)) ? TRUE :
			     ((chars[i] != ':') || (chars[i+1] != ']')) )
			{
				return ENCODE_PARSE_ERROR(REX_INVALID_CHARCLASS, i);
			}
			i += 2;
            continue;
		}
		else
		{
			startchar = endchar = thischar;
			i++;
		}

		if (options & REX_COMPILE_ICASE)
		{
			startchar = TOUPPER(startchar);
			endchar = TOUPPER(endchar);
		}

		c = startchar;
		while (c <= endchar)
		{
			CHARSET_ADD(*set, c);
			c++;
			if (c == 0)
			{
				break;
			}
		}
	}

	if (invert)
	{
		int i;
		for (i=0; i < CHARSET_SIZE; i++)
		{
			set->chunk[i] = ~set->chunk[i];
		}
		if (options & REX_COMPILE_NEWLINE)
		{
			/* If the newline option is present, remove \n from the set. */
			set->chunk['\n' / BITSPER(chunk_t)] &=
				~( (chunk_t)1 << ('\n' % BITSPER(chunk_t)) );
		}
	}

	return i;
}

