/*
 * Copyright (c) 2001-2002 Daniel van Vugt <danv@users.sf.net>
 * This file is part of the Rex library.
 * See LICENSE.txt for license details.
 */

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include "rex.h"
#include "rexprivate.h"


static void
PrintCharSet(const charsethead_t *set)
{
	UREXCHAR c;

	c = 0;
	do
	{
		if ( CHARSET_CONTAINS(*set, c) )
		{
			if (isprint(c))
			{
				printf("%c", c);
			}
			else
			{
				printf("\33[32;1m\\x%02x\33[0m", (int)c);
			}
		}
		c++;
	} while (c != 0);
}


static void
PrintGraph(rex_t r)
{
	const automaton_t *a;
	stateid_t sid;

	a = (const automaton_t*)r;
	if (a->used <= 1)
	{
		printf("State machine is empty.\n");
		return;
	}

	printf("Start state is %d\n", a->startstate);

	sid = 1;
	do
	{
		const state_t *state;
		transitionid_t tid;

		state = &a->block[sid].state;

		printf("State %d:\n", sid);
		tid = FIRSTTRANSID(sid);
		while (tid != INVALID_BLOCKID)
		{
			const transition_t *trans;
			const operand_t *operand;

			trans = &a->block[tid].trans;
			operand = OPERANDOF(trans);

			printf("\t ");
			switch (trans->test)
			{
				case UNUSED:
					printf("UNUSED");
					break;
				case START_OF_LINE:
					printf("START_OF_LINE");
					break;
				case END_OF_LINE:
					printf("END_OF_LINE");
					break;
				case START_OF_WORD:
					printf("START_OF_WORD");
					break;
				case END_OF_WORD:
					printf("END_OF_WORD");
					break;
				case NOP:
					printf("NOP");
					break;
				case LITERAL:
					printf("LITERAL `%c'", operand->literal);
					break;
				case DOT:
					printf("DOT");
					break;
				case CHARSET:
					printf("CHARSET [");
					PrintCharSet(&operand->charset);
					printf("]");
					break;
				case WORD_BOUNDARY:
					printf("WORD_BOUNDARY");
					break;
				case WHITESPACE:
					printf("WHITESPACE");
					break;
				case NOT_WHITESPACE:
					printf("NOT_WHITESPACE");
					break;
				case WORD_CHAR:
					printf("WORD_CHAR");
					break;
				case NOT_WORD_CHAR:
					printf("NOT_WORD_CHAR");
					break;
				case DIGIT:
					printf("DIGIT");
					break;
				case NOT_DIGIT:
					printf("NOT_DIGIT");
					break;
				default:
					assert(FALSE);
					return;
			}

			printf(" --> ");
			switch (trans->dest)
			{
				case ACCEPT_STATE:
					printf("ACCEPT");
					break;
				case INVALID_BLOCKID:
					printf("INVALID_BLOCKID");
					break;
				default:
					printf("%d", trans->dest);
					break;
			}
			printf("\n");

			tid = trans->sibling;
		}

		while (1)
		{
			sid += a->block[sid].header.span;
			if (sid < a->used)
			{
				if ( a->block[sid].header.format & FORMAT_STATE )
				{
					break;
				}
			}
			else
			{
				break;
			}
		}
	} while (sid < a->used);

	printf( "%d x %d byte blocks used + %d byte header = %d bytes used\n"
		"%d x %d byte blocks = %d bytes unused\n",
		a->used,
		(int)sizeof(block_t),
		(int)(sizeof(automaton_t) - (sizeof(block_t) * DEFAULT_BLOCKS)),
		(int)(sizeof(automaton_t) + sizeof(block_t) * (a->used - DEFAULT_BLOCKS)),
		a->allocated - a->used,
		(int)sizeof(block_t),
		(int)(sizeof(block_t) * (a->allocated - a->used))
		);
}


int
main(int argc, char *argv[])
{
	rex_t r;
	rex_error_t error;
	size_t offset;
	const char *pattern;

	if (argc < 2)
	{
		printf("Usage: %s <regex> [s|m <searchstr1> [<searchstr2>...]]\n"
			"\ts: substring search\n"
			"\tm: match whole string\n"
			,argv[0]);
		return 1;
	}

	pattern = argv[1];

	r = rexCompile(pattern, strlen(pattern), REX_COMPILE_DV, &error, &offset);
	if (r == REX_INVALID)
	{
		int extra;
		static const char * const errorstr[REX_NERRORS] = REX_ERROR_STRINGS;

		printf("Error: %n%s\n", &extra, pattern);
		printf("%*c %s\n", extra + (int)offset + 1, '^', errorstr[error]);
		return 2;
	}

	printf("%s\n", pattern);

	PrintGraph(r);

	if (argc >= 4)
	{
		int i;
		rex_search_t s;
		size_t off;
		char op;

		op = argv[2][0];
		if (argc > 4)
		{
			s = rexPreallocSearch(r);
		}
		else
		{
			s = NULL;
		}

		for (i=3; i < argc; i++)
		{
			size_t e;
			unsigned int options;

			printf("----------- argv[%d]: `%s' ------------\n", i, argv[i]);
			options = (i >= (argc-1)) ? REX_MATCH_EOL : 0;
			options |= (op == 'm') ? REX_MATCH_WHOLE : REX_MATCH_SUBSTRING;
			e = rexMatch(r, argv[i], strlen(argv[i]), options, s, &off);
			printf("%s at offset %lu", e ? "Matched" : "No match", (unsigned long)off);
			if (e > 0)
			{
				int j;
				printf(", length %lu\n", (unsigned long)e - 1);
				for (j=3; j <= i; j++)
				{
					printf("%s", argv[j]);
				}
				printf("\n");
				while (off > 0)
				{
					off--;
					printf(" ");
				}
				e--;
				while (e > 0)
				{
					printf("^");
					e--;
				}
			}
			printf("\n");
		}

		if (s != NULL)
		{
			rexFreeSearch(s);
		}
	}

	rexFree(r);

	return 0;
}
