/*
 * Copyright (c) 2001-2002 Daniel van Vugt <danv@users.sf.net>
 * This file is part of the Rex library.
 * See LICENSE.txt for license details.
 */

#ifndef _REXPRIVATE_H
#define _REXPRIVATE_H

#include <stddef.h>
#include "rex.h"

typedef unsigned char byte_t;

#define ROUND_UP(n,f) ( (((n) + (f) - 1) / (f)) * (f) )

#define BITSPER(x) ( sizeof(x) * 8 )

#ifdef FALSE
#undef FALSE
#endif

#ifdef TRUE
#undef TRUE
#endif

#define FALSE 0
#define TRUE (!FALSE)
typedef int bool_t;

#define DEFAULT_BLOCKS 4
#define DEFAULT_MEM_LIMIT (256 * 1024)

#define ACCEPT_STATE 0
#define INVALID_BLOCKID (-1)
#define SEPARATOR (-2)

#define REXCHAR char
#define UREXCHAR unsigned char

typedef enum
{
	UNUSED,
	NOP,
	LITERAL,
	DOT,
	CHARSET,
	START_OF_LINE,
	END_OF_LINE,
	START_OF_WORD,
	END_OF_WORD,
	WORD_BOUNDARY,
	WHITESPACE,
	NOT_WHITESPACE,
	WORD_CHAR,
	NOT_WORD_CHAR,
	DIGIT,
	NOT_DIGIT
} test_t;

#define CHARSET_SIZE ( (1 << BITSPER(REXCHAR)) / BITSPER(chunk_t) )
#define CHARSET_CONTAINS(set,ch) \
	( (set).chunk[(UREXCHAR)(ch) / BITSPER(chunk_t)] & \
	  (1 << ((UREXCHAR)(ch) & (BITSPER(chunk_t)-1))) )

#define CHARSET_ADD(set,ch) \
	{ *( (set).chunk + (ch / BITSPER(chunk_t)) ) |= \
		( (chunk_t)1 << (ch % BITSPER(chunk_t)) ); }

typedef unsigned int chunk_t;
typedef struct
{
	chunk_t chunk[CHARSET_SIZE];
} charset_t;
typedef struct
{
	chunk_t chunk[1];
} charsethead_t;

typedef int       blockid_t;
typedef blockid_t stateid_t, transitionid_t;

#define FORMAT_STATE   0x1
#define FORMAT_TRANS   0x2

typedef struct
{
	int format;
	int span;
} header_t;

typedef struct
{
	header_t       header;
	test_t         test;
	stateid_t      dest;
	transitionid_t sibling;
} transition_t;

typedef union
{
	REXCHAR       literal;
	charsethead_t charset;
} operand_t;
#define OPERANDOF(transptr) ((operand_t*) ((block_t*)(transptr) + 1))
#define FIRSTTRANSID(stateid) ((transitionid_t)(stateid) + 1)

typedef struct
{
	header_t       header;
	transitionid_t lasttrans;
} state_t;

typedef union
{
	header_t     header;  /* only applies to state and trans */
	state_t      state;
	transition_t trans;
	operand_t    operand;
} block_t;

typedef struct
{
	int       used;
	int       allocated;
	stateid_t startstate;
	stateid_t maxstate;
	unsigned int options;
	block_t   block[DEFAULT_BLOCKS];
} automaton_t; /* A non-deterministic finite state automaton */

#if (REX_PATTERN_MAX <= 256)
typedef unsigned int parseresult_t;
#else
typedef unsigned long parseresult_t;
#endif
#define PARSE_FAILED(pr)  ( (pr) >= REX_PATTERN_MAX )
#define ENCODE_PARSE_ERROR(type,offset) ( ((type) * REX_PATTERN_MAX) | (offset) )
#define DECODE_PARSE_ERROR(error,typevar,offsetvar) \
	{ \
		(offsetvar) = (size_t)((error) & (REX_PATTERN_MAX - 1)); \
		(typevar) = (rex_error_t)((error) / REX_PATTERN_MAX);    \
	}

typedef struct
{
	size_t offset;  /* offset at which the respective state was last visited */
	size_t origin;  /* offset at which the start state was introduced to
                           reach the respective state */
} history_t;

/* search_t.flags values: */
#define SEARCH_MALLOCED    0x1 /* The structure has been MALLOC'd */
#define SEARCH_NO_MORE     0x2 /* Further REX_SEARCH_WHOLE matching is
                                  impossible */
#define SEARCH_NOT_ALLOWED 0x4 /* Further calls to rexMatch are not allowed
                                  because REX_MATCH_EOL has already been
                                  passed. */
typedef struct
{
	const automaton_t *automaton;
	int        size;
	int        head;
	int        tail;
	int        side;  /* history index: 0 or 1 */
	int        flags;
	size_t     base;
	size_t     bestorigin;
	size_t     bestlen;
	REXCHAR    prevchar;
	history_t *history[2];  /* history[side][stateid] */
	stateid_t *queue;
} search_t;

#define QUEUE_ALIGNMENT sizeof(stateid_t)
#define HIST_ALIGNMENT  sizeof(size_t)

#define MANGLE(name) _rex_private_##name

#define rexmemlimit MANGLE(rexmemlimit)
extern size_t rexmemlimit;

#ifdef __cplusplus
extern "C" {
#endif

#define EscapeChar MANGLE(EscapeChar)
extern parseresult_t EscapeChar(const REXCHAR *format, int len, REXCHAR *out);

#define InitCharSet MANGLE(InitCharSet)
extern parseresult_t InitCharSet(charset_t *set, const REXCHAR *chars, size_t len, unsigned int options);

#define ParseBound MANGLE(ParseBound)
extern parseresult_t ParseBound(const REXCHAR *chars, size_t len, int *min, int *max);

#ifdef __cplusplus
}
#endif

#endif
